/*

 Name  :        fgtcih.c
 Descr :        Input program for cps high kevek asynchronous mode test job 

 Author:	R. C. Yeager
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/



#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <acp_user.h>
#include "fgtched.h"

main() 

    {

    void gtimfunc();                    /* declare formatted time function   */

    int sub;                            /* general index                     */
    int ifile;                          /* input file id                     */
    int inbyte = 0;                     /* byte count from reads             */
    int byte_count;                     /* cps byte count                    */
    int block_no;                       /* cps block id (queue number)       */
    int queue_num;                      /* cps declare queue id              */
    int which_proc;                     /* cps process number                */
    int act_type;                       /* cps requested action type         */
    int msg_inc;                        /* block increment from getenv       */
    int num_of_msgs;                    /* number of blocks from getenv      */
    int max_imsg_size;                  /* input block size from getenv      */
    int nbuf;                           /* number of buffers from getenv     */
    int block_size;                     /* cps block size variable           */
    int num_elem;                       /* number of actual ints in block    */
    int bsize;                          /* stored block size (in block)      */
    int list;                           /* acp_get_next_tape list id         */ 
    int pos;                            /* acp_get_next_tape list position   */
    int status;                         /* acp_get_next_tape return status   */
    int *bp;                            /* buffer pointer subscript          */
    int eventno;                        /* record counter                    */ 
    int key[1];                         /* acp_buffer_index key              */
    int buf_index;                      /* acp_buffer_index offset           */

    HEADR *bufr;                        /* data buff ptrs cast w/HEADR struct*/
    char *getenv(char *);               /* declare char form of getenv       */
    char idev[8];                       /* input device type from getenv     */
    char fpin[80];                      /* input file name from getenv       */
    char cpsfpin[80];                   /* formatted cps tape driver name    */
    char errmsg[80];                    /* acp_get_next_tape error message   */
    char obuf[8];                       /* return variable from time function*/
    char device[80];                    /* acp_get_next_tape returned device */
    char vsn[12];                       /* acp_get_next_tape returned volser */
    char errormsg[80];                  /* formatted device error message    */
    char *sep = " ";                    /* strtok seperator character        */

/*  need some comments */

    acp_init();

    gtimfunc(obuf);
    printf("Starting  : Time %.8s\n",obuf);
    fflush(stdout);

/*  need some comments */

    strcpy(idev,IDEV);
    strcpy(fpin,IFILE);
    num_of_msgs = MSGNUM;
    msg_inc = MSGINC;
    max_imsg_size = MSGISIZE;
    nbuf = NBUF;

    if ((strncmp(idev,"tape",4) == 0) &&
        (strncmp(fpin,"cpsbatch",8) == 0)) 
        {
        list = 1;
        acp_get_next_tape(&list,device,vsn,&pos,&status,errmsg,80,12,80);
        if (status == ACP_OK) 
            {
            (strtok(device,sep));
            sprintf(cpsfpin,"/usr/cpsdev/%s_nrnsv",device);
            }
        else
            {
            printf("%s\n",errmsg);
            fflush(stdout);
            acp_stop_job();
            }
        }
     else
        {
        strcpy(cpsfpin,fpin);
        }

    if (strncmp(idev,"none",4) != 0) 
        {
#ifdef AIX
        sighold(SIGUSR1);
#endif
        if (0 > (ifile = open(cpsfpin,O_RDONLY,0))) 
            {
            perror(cpsfpin);
            fflush(stderr);
#ifdef AIX
            sigrelse(SIGUSR1);
#endif
            acp_stop_job();
            }
#ifdef AIX
        sigrelse(SIGUSR1);
#endif
        }

    sprintf(errormsg,"I/O error during read for %s",cpsfpin);
    block_size = (max_imsg_size-(MAX_HEADER_SIZE*NUM_OF_HEAD));
    num_elem = (block_size/sizeof(int));

    which_proc = ACP_ALL_PROCESSES;
    act_type = BEGIN_SYNC;
    acp_sync(&which_proc,&act_type);

    gtimfunc(obuf);
    printf("Synchd    : Time %.8s\n",obuf);
    fflush(stdout);

    eventno = 1;
    bp = 0;

    while (1)

        { 
            
        byte_count = max_imsg_size;

        buf_index = acp_buffer_index(key,&byte_count,&nbuf);
        bp = key;
        bp += buf_index - 1;
        bufr = (HEADR *) bp;

        if (strncmp(idev,"none",4) == 0) 
            {
            for (sub = 0; sub < num_elem; sub++) 
                {
                bufr->block[sub] = rand();
                }
            bufr->block[0] = max_imsg_size;
            }

        if (eventno <= num_of_msgs)
            {
            if (strncmp(idev,"none",4) != 0) 
                {
#ifdef AIX
                sighold(SIGUSR1);
#endif
                if (0 > (inbyte = read(ifile,bufr,byte_count))) 
                    {
                    perror(errormsg);
                    fflush(stderr);
#ifdef AIX
                    sigrelse(SIGUSR1);
#endif
                    acp_stop_job();
                    }
                else 
                    {
#ifdef AIX
                    sigrelse(SIGUSR1);
#endif
                    if (inbyte == 0) 
                        {
                        break;
                        }
                    }
                }
            gtimfunc(obuf);
            sprintf(bufr->header_ini, 
                "From Input: Time %.8s Block %d ", 
                obuf,eventno);
            }
        else
            {
            break;
            }

        bsize = bufr->block[0];

        gtimfunc(obuf);
        sprintf(bufr->header_ino, 
            "To Event  : Time %.8s a %d r %d b %d", 
            obuf,inbyte,byte_count,bsize);

        byte_count = bsize;
        block_no = IN_BLOCK;
        queue_num = READY_QUEUE;
        acp_start_send(bufr,&byte_count,&block_no,&queue_num);

        if (((eventno % msg_inc) == 1) || (msg_inc == 1)) 
            {
            gtimfunc(obuf);
            printf("Sent Block: Time %.8s Block %d \n",obuf,eventno);
            fflush(stdout);
            }

        eventno++;

        }

    if (strncmp(idev,"none",4) != 0) 
        {
#ifdef AIX
        sighold(SIGUSR1);
#endif
        close(ifile);
#ifdef AIX
        sigrelse(SIGUSR1);
#endif
        }

    queue_num = READY_QUEUE;
    act_type = ACP_FULL;
    acp_wait_queue(&queue_num,&act_type);

    act_type = ACP_END_OF_QUEUE;
    queue_num= DONE_QUEUE;
    acp_queue_process(&act_type,&queue_num);

    gtimfunc(obuf);
    printf("Stopping  : Time %.8s\n",obuf);
    fflush(stdout);
    acp_stop_process();

    }                                   /* end of main block                 */

    /*  called function to format and return the time in format HH:MM:SS     */

    void gtimfunc(obuf) 

        {
        
        long tloc;                      /* return variable from time function*/
        char buf[30];                   /* formatted time variable           */
 
        time(&tloc);                    /* get current time                  */
        strcpy(buf,ctime(&tloc));       /* format current time into buf      */
        strncpy(obuf,&buf[11],8);       /* move time value to obuf           */
        return;                         /* return to caller                  */

        }

