/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/



#include <stdio.h>
#include "rnd.h"

main() {

    int i,p1,p2,p3,p4;

    FILE *fp,*fopen();

    char devnam[80];

    float x[EVENT_SIZE];

    int p;

    int byte_count,blockno,offset;

    int eventno = 0;

    int listno,mstatus,dstatus,pos,rstatus;
    char tdrive[40],vsn[80],errmsg[80];

    char *c,*strchr();

    int one = 1;

    acp_init();

    listno = 1;
    acp_get_next_tape(&listno,tdrive,vsn,&pos,&rstatus,errmsg,40,80,80);
    c = strchr(tdrive,' ');
    *c = '\0';
    c = strchr(vsn,' ');
    *c = '\0';
    errmsg[50] = '\0';
    if (rstatus == ACP_OK) {
	sprintf(devnam,"/usr/cpsdev/%s_nsv",tdrive);
	printf("%s mounted on %s\n",vsn,tdrive);
    }
    else if (rstatus == ACP_DEVICE_UNAVAILABLE) {
	printf("No tape drives are available\n");
	acp_stop_job();
    }
    else if (rstatus == ACP_EOF) {
	printf("End of tape list reached\n");
    	acp_stop_job();
    }
    else if (rstatus == ACP_ERROR) {
	printf("Failed to mount tape (%s)\n",vsn);
	printf("Error message: %s\n",errmsg);
    	acp_stop_job();
    }
    else {
	printf("Unrecognizable mount reply status (%d)\n",rstatus);
    	acp_stop_job();
    }
    fflush(stdout);

    fp = fopen(devnam,"r");
    if (fp == NULL) {
	printf("failed to open %s\n",devnam);
    	fflush(stdout);
	acp_stop_job();
    }

    p1 = ACP_ALL_PROCESSES;
    p2 = BEGIN_SYNC;
    acp_sync(&p1,&p2);

    while (1) {
	for (i=0; i<EVENT_SIZE; i++) {
	    if ((eventno >= MAX_EVENTS) || 
		(fscanf(fp,"%f",&(x[i])) == EOF)) {   /* end of tape */

		printf("end of input tape\n");

		fclose(fp);

		dstatus = acpsta_done_with_tape(tdrive,40);
		if (dstatus != ACP_OK) {
		    printf("Failed to deallocate %s\n",tdrive);
		    fflush(stdout);
		}

    		p1 = READY_QUEUE;
    		p2 = ACP_FULL;
    		acp_wait_queue(&p1,&p2);

		p1 = acp_set_of_processes(&one,&one);
		p2 = END_SYNC;
		acp_sync(&p1,&p2);

    		p1 = ACP_END_OF_QUEUE;
    		p2 = DONE_QUEUE;
    		acp_queue_process(&p1,&p2);

    		acp_stop_process();
	    }
	}

	p1 = READY_QUEUE;
	acp_dequeue_process(&p,&p1);

	byte_count = EVENT_SIZE * sizeof(float);
	blockno = IN_BLOCK;
	offset = 0;
	acp_send(&p,x,&byte_count,&blockno,&offset);

	p1 = DONE_QUEUE;
	p2 = SORT_REMSUB;
	p3 = 0;
	acp_call(&p,&p1,&p2,&p3);

	eventno++;
	if ((eventno % 10) == 0) {
		printf("sent event number %d to be sorted\n",eventno);
		fflush(stdout);
	}
    }
}
