/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: db_param.h

*/

#define DB_NOPARAM -400	    /* No parameter should have this value */

/* Device Types (may be entered into the PM database) */

#define DB_NUM_DEVICE_TYPES 4
#define DB_DISK 0
#define DB_EXAB_8200 1		/* Exabyte 8200 tape drives */
#define DB_NTRK 2			/* 9-track reels */
#define DB_EXAB_8500 3		/* Exabyte 8500 tape drives */

/* Device Types (may be requested from the PM database) */

#define DBR_NUM_DEVICE_TYPES 5
#define DBR_DISK 0
#define DBR_EXAB 1			/* Exabyte 8200 or 8500 tape drives */
#define DBR_NTRK 2			/* 9-track reels */
#define DBR_EXAB_8500 3		/* Exabyte 8500 tape drives */
#define DBR_EXAB_8200 4		/* Exabyte 8200 tape drives */

/* Operating System Types */

#define DB_UNKNOWN 0
#define DB_VMS	1
#define DB_UNIX	2

/* CPU Types */

#define DB_NUM_CPU_TYPES 9
#define DB_ULTRIX 0
#define DB_VAXVMS 1
#define DB_MIPS	  2
#define DB_SUN	  3
#define DB_SGI    4
#define DB_AIX	  5
#define DB_R3000  6
#define DB_HPUX   7
#define DB_SONY   8

/* CPU Task Types */

#define DB_TASK_ANY 0
#define DB_TASK_IO 1
#define DB_TASK_COMPUTE 2

/* Hardware Status */

#define DB_WORKING 0
#define DB_BROKEN  1
#define DB_REDUCED 2			/* Working but not a full power */

/* Mount request types */

#define DB_LIST_MOUNT 0		    /* Mount next tape in a list on any drive */
#define DB_VSN_MOUNT 1		    /* Mount a specific tape on any drive*/
#define DB_USER_MOUNT 2
#define DB_SPEC_MOUNT 3		    /* Mount a specific tape on a specific 
				       drive */

/* Mount reply statuses */

#define DB_MOUNT_SUCCESS 0	    /* No problems at all */
#define DB_MOUNT_SKIPPED 1	    /* Success, but some tapes in list skipped */
#define DB_MOUNT_EOL 2		    /* End of tape list reached */
#define DB_MOUNT_NODEV 3	    /* No devices available */
#define DB_MOUNT_FAILED 4	    /* Mount failed for some reason */

/* Tape Drive Mount States */

#define DB_NOT_MOUNTED 0
#define DB_ACTIVE 1
#define DB_MOUNT_PENDING 2
#define DB_PRE_MOUNTED 3
#define DB_MOUNTED 4

/* Tape Allocation Parameters */

#define DB_ALLOC 0
#define DB_UNALLOC 1

/* Tape Reel Statuses */

#define DB_UNPROCESSED 0
#define DB_MOUNT_IN_PROG 1	/* Mount in progress */
#define DB_CURRENT 2		/* Currently being processed */
#define DB_PROCESSED 3
#define DB_MISSING 4
#define DB_DAMAGED 5

/* Tape Directions */

#define DB_INPUT 0
#define DB_OUTPUT 1

/* Job Pipeline Stages */

#define DB_PIPE_NEW 0
#define DB_PIPE_QUEUE 1
#define DB_PIPE_ON_DECK 2
#define DB_PIPE_CURRENT 3
#define DB_PIPE_RUNDOWN 4

/* CPS Job Statuses */

#define DB_SUCCESS 0
#define DB_FAILURE 1
#define DB_ABORT 2
#define DB_IN_PROGRESS 3
#define DB_PENDING 4

/* Database Update Message Types */

#define DB_NUM_UPDATE_MSGS 60
#define DB_ADD_CPU 1
#define DB_MODIFY_CPU 2
#define DB_DELETE_CPU 3
#define DB_ADD_TAPE 4
#define DB_MODIFY_TAPE 5
#define DB_DELETE_TAPE 6
#define DB_ADD_PRODSYS 7
#define DB_MODIFY_PRODSYS 8
#define DB_DELETE_PRODSYS 9
#define DB_ADD_CPU_TO_PROD 10
#define DB_MODIFY_CPU_IN_PROD 11
#define DB_DELETE_CPU_FROM_PROD 12
#define DB_ADD_TAPE_TO_PROD 13
#define DB_DELETE_TAPE_FROM_PROD 14
#define DB_ADD_FILESYSTEM 15
#define DB_REMOVE_FILESYSTEM 16
#define DB_ADD_JOB 17
#define DB_REMOVE_JOB 18
#define DB_ADVANCE_JOB 19
#define DB_UPDATE_JOB_STATUS 20
#define DB_SET_JOB_AREA 21
#define DB_FILL_SUBMIT_INFO 22
#define DB_ADD_JOB_CLASS 23
#define DB_ADD_TAPE_LIST 24
#define DB_ADD_TAPE_REEL 25
#define DB_REMOVE_TAPE_REEL 26
#define DB_MOVE_JOB_TO_QUEUE 27
#define DB_MOVE_JOB_TO_ON_DECK 28
#define DB_MOVE_JOB_TO_CURRENT 29
#define DB_MOVE_JOB_TO_RUNDOWN 30
#define DB_ADD_USER 31
#define DB_REMOVE_USER 32
#define DB_ALLOCATE_AREA 33
#define DB_FREE_AREA 34
#define DB_SET_ABORT_FLAG 35
#define DB_CLEAR_ABORT_FLAG 36
#define DB_SET_PHASE_TIMES 37
#define DB_MOUNT_INFO 38
#define DB_TAPE_REEL_STATUS 39
#define DB_INSERT_TAPE_REEL 40
#define DB_TAPE_TIME_REQ 41
#define DB_TAPE_TIME_MNT 42
#define DB_TAPE_TIME_DONE 43
#define DB_JCSS_PID 44
#define DB_SET_NEXT_ENTRY 45
#define DB_NO_MORE_UPDATES 60

/* Database Query Message Types (PM responds by sending update messages 
   ending with a DB_NO_MORE_UPDATES message) */

#define DB_QUERY_CPU -1		    /* Request single CPU */
#define DB_QUERY_TAPE -2	    /* Request single tape */
#define DB_QUERY_DB -3		    /* Request entire database */
#define DB_QUERY_CPU_LIST -4	    /* Request entire CPU list */
#define DB_QUERY_TAPE_LIST -5	    /* Request entire tape list */
#define DB_QUERY_PROD_CPU_LIST -6   /* Request production system's CPU list */
#define DB_QUERY_PROD_TAPE_LIST -7  /* Request production system's tape list */
#define DB_QUERY_PROD_DEF -8	    /* Request production system definition -
				       (CPU and tape lists are sent, but no
					job info)			     */
#define DB_QUERY_JOB -9		    /* Request job info */
#define DB_QUERY_PROD_RESOURCES -10 /* Request definitions of resources used
				       by a production system */
#define DB_QUERY_JOB_INFO -11	    /* Request information on all jobs in
				       a production system */
#define DB_QUERY_FS -12		    /* Request filesystem information */
#define DB_QUERY_FS_LIST -13	    /* Request filesystem information */
#define DB_QUERY_RESOURCES -14	    /* Request entire database except for
				       job information */
