/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				JM_MAIN.C												*/
/*																		*/
/*																		*/
/************************************************************************/



/*****************/
/* Include Files */
/*****************/

#include <stdio.h>
#include <signal.h>
#include "../inc/types.h"
#include "../inc/socket.h"
#include "../inc/types.h"
#include "../inc/time.h"
#include "../inc/perr.h"
#include "../inc/param.h"
#include "../inc/cps_q.h"
#include "jm_param.h"
#include "jm_structs.h"
#include "jm_globals.h"
#include "../inc/shm.h"
#include "../inc/xfunc.h"
#include "../inc/version.h"

/**************************/
/* Parameter Declarations */
/**************************/

#ifndef TEST_LEVEL
#define TEST_LEVEL 1000
#endif

/********************/
/* Global Variables */
/********************/

FILE *logfp = NULL;

/************************/
/* External Definitions */
/************************/

char *strchr();
FILE *fopen();
char *ttyname();

extern int lineno,numerr,errlin[32];

extern int hm_report;
extern int hm_op_interval;

extern FILE *yyin,*yyout;

void *malloc();

analyze_msg(s,c,seq,m)
int s,seq;
char *c;
cps_message *m;
{
	fprintf(logfp,"(%d/%s/%d) ",s,c,seq);
	log_cps_msg(logfp,m);
}

int getline( s, lim )
char s[];
int lim;
{
	int c, i;

	for( i = 0; i < lim - 1 && ( c = getchar() ) != EOF && c != '\n'; ++i )
		s[ i ] = c;
	if( c =='\n' )
		s[ i ] = '\0';
	return i;
}

void sigint_handler(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	abort_job(7);
}

display_error(s)
char *s;
{
	print_log(s);
}

print_set(fp,s,n)
FILE *fp;
int n;
unsigned long *s;
{
	int *x,i;

	x = set_query(s,n);
	i = 0;
	fprintf(fp,"(");
	while (x[i] != 0) {
		fprintf(fp,"%d",x[i++]);
		if (x[i] != 0) fprintf(fp,",");
	}
	fprintf(fp,")");
}
print_sdf_cpus() {

	int i,j;

	int *x;

	for (i=0; i<num_sdf_cpus; i++) {
		printf("(%d) %s: ",sdf_cpu[i].cpu_type,sdf_cpu[i].cpu_name);
		print_set(stdout,sdf_cpu[i].procs,MAX_PROCESSES);
		printf(" ");
		print_set(stdout,sdf_cpu[i].class,MAX_CLASSES);
		printf("\n");
	}

}
print_class_info() {

	int i,j;

	int *x;

	for (i=1; i<=MAX_CLASSES; i++) {
		if (class[i].valid) {
			printf("Class %d:\n",i);
			printf("  Share: ");
			print_set(stdout,class[i].share,MAX_CLASSES);
			printf("\n");
			printf("  Program: %s\n",class[i].program);
			printf("  CPU Type: %d\n",class[i].cpu_type);
			for (j=0; j<class[i].nspec; j++) {
				printf("  Spec CPU: %s\n",
				sdf_cpu[class[i].spec_cpu[j]].cpu_name);
			}
			printf("  Num Proc: %d\n",class[i].nproc);
			printf("  Per CPU: %d\n",class[i].per_cpu);
			printf("  Min Proc: %d\n",class[i].min_proc);
			printf("  Debug: %d\n",class[i].debug);
			printf("  Timeout: %d\n",class[i].timeout);
		}
	}
}
print_cpu_assignments() {

	int i,j;

	int *x;

	for (i=1; i<=MAX_CLASSES; i++) {
		if (class[i].valid) {
			printf("Class %d:\n",i);
			printf("  Num Proc: %d\n",class[i].nproc);
			printf("  Per CPU: %d\n",class[i].per_cpu);
			printf("  Processes: ");
			print_set(stdout,class[i].sop,MAX_PROCESSES);
			printf("\n");
			printf("  CPU Type: %d\n",class[i].cpu_type);
			for (j=0; j<class[i].nspec; j++) {
				printf("  Spec CPU: %s\n",
				sdf_cpu[class[i].spec_cpu[j]].cpu_name);
			}
			x = set_query(class[i].cpu,MAX_SDF_CPUS);
			j = 0;
			if ((class[i].nspec == 0) && (x[j] == 0)) 
				printf("  No possible CPUs\n");
			while (x[j] != 0) {
				printf("  Possible CPU: %s\n",sdf_cpu[x[j]-1].cpu_name);
				j++;
			}
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  contact_portmapper

  DESCRIPTION:

	  Attempts to contact portmapper

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/
int contact_portmapper()
{
	struct pmaplist *PmapList;
	struct sockaddr_in addr;
	struct hostent *hp;
	char errmsg[132];

	hp = gethostbyname( jm_hostname );
	if( hp == NULL )
	{
		sprintf(errmsg, "Failed to call gethostbyname for %s",jm_hostname);
		printerr(FACILITY,FATAL_MSG,errmsg,NULL,70);
		jm_exit();
	}
	strcpy( jm_hostname, hp->h_name );
	bzero( &addr, sizeof( addr ) );
	bcopy( hp->h_addr, &addr.sin_addr, hp->h_length );
	addr.sin_family = AF_INET;
	addr.sin_port = 0;

	PmapList = (struct pmaplist *)pmap_getmaps( &addr );

	if( PmapList == NULL )
	{
		sprintf(errmsg, "Could not contact portmapper");
		printerr(FACILITY,FATAL_MSG,errmsg,NULL,71);
		jm_exit();
	}
}
/******************************************************************/
/*
	FUNCTION NAME:

	  jm_exit

  DESCRIPTION:

	  Exit routine for JM child process.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_exit()
{
	time_t exit_time;

	if( upshot )
		upshot_stop();

	if( jm_prof )
		jm_profile_write();

	jm_rpc_exit();

	time( &exit_time );
	fprintf( jml, "jm[%s] exited gracefully at %s",
		jm_hostname, ctime( &exit_time ) );
	fflush( jml );

	exit(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  print_log

  DESCRIPTION:

	  Write a string to standard output as well as the .jml file.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

print2_usr_log(lpn,msg1,msg2)
int lpn;
char *msg1,*msg2;
{
	printf("[%d] %s%s\n",lpn,msg1,msg2);
	fprintf(jml,"[%d] %s%s\n",lpn,msg1,msg2);
	fflush(jml);
	fflush(stdout);
}

print_usr_log(lpn,msg)
int lpn;
char *msg;
{
	printf("[%d] %s\n",lpn,msg);
	fprintf(jml,"[%d] %s\n",lpn,msg);
	fflush(jml);
	fflush(stdout);
}

print_log(msg)
char *msg;
{
	printf("%s\n",msg);
	fprintf(jml,"%s\n",msg);
	fflush(jml);
	fflush(stdout);
}

print_log_no_cr(msg)
char *msg;
{
	printf("%s",msg);
	fprintf(jml,"%s",msg);
	fflush(jml);
	fflush(stdout);
}

print_shm_log(i,msg)
int i;
char *msg;
{
	printf("shm[%s] %s\n",sdf_cpu[shm_map[i].cpu].cpu_name,msg);
	fprintf(jml,"shm[%s] %s\n",sdf_cpu[shm_map[i].cpu].cpu_name,msg);
	fflush(jml);
	fflush(stdout);
}

print_lpn(lpn)
int lpn;
{
	printf("[%d]\n",lpn);
	fprintf(jml,"[%d]\n",lpn);
	fflush(jml);
	fflush(stdout);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  run_cpp

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

run_cpp()
{
	char cpp_file[80];
	char cpp_command[256];

	FILE *fp;

	if ((fp = fopen("/lib/cpp","r")) != NULL) {
		fclose(fp);
		strcpy(cpp_file,"/lib/cpp");
	}
	else if ((fp = fopen("/usr/lib/cpp","r")) != NULL) {
		fclose(fp);
		strcpy(cpp_file,"/usr/lib/cpp");
	}
	else {
		printf("Enter the pathname of the C preprocessor (/lib/cpp): ");
		scanf("%s",cpp_file);
		if ((fp = fopen(cpp_file,"r")) != NULL) {
			fclose(fp);
		}
		else {
			printf("Failed to locate the C preprocessor\n");
			exit(0);
		}
	}

	sprintf(cpp_command,"%s -P %s %s %s.cpp",
		cpp_file,cpp_options,jdf_file,job_name);
	if (TEST_LEVEL == 2)
		printf("%s\n",cpp_command);
	system(cpp_command);
	sprintf(jdf_file,"%s.cpp",job_name);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  parse_command_line

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

parse_command_line(argc,argv)
int argc;
char **argv;
{
	int index;
	char *arg,*c;
	int opt;

	int status;

	if (argc < 2)
	{
		printf("\n");
		printf("Usage: cps_jm [options] <jdf_filename>\n");
		printf("\n");
		printf("OPTION       DESCRIPTION\n");
		printf("-cpp         Run C pre-processor on JDF file\n");
		printf("-d           Start job in debug mode\n");
		printf("-f filename  Log file name\n");
		printf("-h           Activate hooks in CPS subroutines for use in debugging\n");
		printf("-pw          Prompt for password only once\n");
		printf("-s           Allows cps_mount_request calls\n");
		printf("-u           Log all messages to an `Upshot' log file\n");
		printf("\n");
		exit(0);
	}

/* Initialize cpp options */

	cpp_options[0] = '\0';

/* Parse command line options */

	status = 0;
	index = 1;

	while( index < argc )
	{
		arg = argv[index];

		if( !strcmp( arg, "-cc" ) )
		{
			carbon_copy = -1;
		}

		else if( !strcmp( arg, "-cpp" ) )
		{
			invoke_cpp = -1;
		}

		else if( !strcmp( arg, "-d" ) )
		{
			jm_pause = 1;
			jm_init  = 1;
		}

		else if( !strcmp( arg, "-f" ) )
		{
			index++;
			if( index < argc )
			{
				log_file = (char *)hmalloc( strlen( argv[index] ) + 1 );
				strcpy( log_file, argv[index] );
			}
			else
			{
				status = 1;
				break;
			}
		}

		else if( !strcmp( arg, "-h" ) )
		{
			jm_hooks = 1;
		}

		else if( !strcmp( arg, "-l" ) )
		{
			index++;
			if( index < argc )
				log_level = atoi( argv[index] );
			else
			{
				status = 1;
				break;
			}
		}

		else if( !strcmp( arg, "-m" ) )
		{
			index++;
			if( index < argc )
				dbg_level = atoi( argv[index] );
			else
			{
				status = 1;
				break;
			}
		}

		else if( !strcmp( arg, "-n" ) )
		{
			index++;
			if( index < argc )
				user_timeout = atoi( argv[index] );
			else
			{
				status = 1;
				break;
			}
		}

		else if( !strcmp( arg, "-p" ) )
		{
			jm_prof = 1;
		}

		else if( !strncmp( arg, "-ps", 3 ) )
		{
			if( strlen( arg ) > 3 )
			{
				batch_mode = -1;
				p_system = (char *)hmalloc( strlen( arg ) - 2 );
				strcpy( p_system, &arg[3] );
			}
			else
			{
				status = 1;
				break;
			}
		}

		else if( !strcmp( arg, "-pw" ) )
		{
			login_prompt = -1;
		}

		else if( !strcmp( arg, "-s" ) )
		{
			pm_connect_flag = -1;
		}

		else if( !strcmp( arg, "-t" ) )
		{
			index++;
			if( index < argc )
				term_level = atoi( argv[index] );
			else
			{
				status = 1;
				break;
			}
		}

		else if( !strcmp( arg, "-u" ) )
		{
			upshot = 1;
		}

		else if( !strncmp( arg, "-D", 2 ) || !strncmp( arg, "-U", 2 ) )
		{
			strcat( cpp_options, arg );
			strcat( cpp_options, " " );
		}

		else if( !strncmp( arg, "-", 1 ) )
		{
			printf("Unrecognized argument: %s\n",arg);
			status = 1;
			break;
		}

		else if( job_name == NULL )
		{
			c = strchr( arg, '.' );
			if( c != NULL )
				*c = '\0';
			job_name = (char *)hmalloc( strlen( arg ) + 1 );
			strcpy( job_name, arg );
			if( c != NULL )
				*c = '.';
			jdf_file = (char *)hmalloc( strlen( arg ) + 5 );
			if( c == NULL )
				sprintf( jdf_file, "%s.jdf", arg );
			else
				strcpy( jdf_file, arg );
		}

		else
		{
			printf("Error in command line: %s\n",arg);
			status = 1;
			break;
		}

		index++;
	}

	if (status) exit(0);

	if( job_name == NULL )
	{
		printf("job name not specified\n");
		exit(0);
	}

	if (log_file == NULL)
	{
		log_file = (char *) hmalloc(strlen(job_name)+5);
		sprintf(log_file,"%s.jml",job_name);
	}

	if (TEST_LEVEL == 1)
	{
		printf("job name: %s\n",job_name);
		printf("jdf_file: %s\n",jdf_file);
		printf("log_file: %s\n",log_file);
		printf("login_prompt: %d\n",login_prompt);
		printf("log_level: %d\n",log_level);
		printf("term_level: %d\n",term_level);
		printf("dbg_level: %d\n",dbg_level);
		printf("carbon_copy: %d\n",carbon_copy);
		printf("invoke_cpp: %d\n",invoke_cpp);
		printf("user_timeout: %d\n",user_timeout);
		printf("cpp_options: %s\n",cpp_options);
	}

	return;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  get_resources

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

get_resources()
{
	FILE *fp;

	char *home,*getenv();

	char fname[132];

	char buf[1024];
	int i,j,k;

/* Find the cps_sdf file */

	home = getenv("HOME");
	sprintf(fname,"%s/.cps_sdf",home);
	fp = fopen(fname,"r");
	if (fp != NULL)
	{
		fclose(fp);
		goto parse_sdf;
	}

	sprintf(fname,"%s/.acp_sdf",home);
	fp = fopen(fname,"r");
	if (fp != NULL)
	{
		fclose(fp);
		goto parse_sdf;
	}

	home = getenv("CPSBIN");
	sprintf(fname,"%s/cps_sdf",home);
	fp = fopen(fname,"r");
	if (fp != NULL)
	{
		fclose(fp);
		goto parse_sdf;
	}

	printf("Warning: failed to locate cps_sdf file\n");
	printf("         checked: ~/.cps_sdf, ~/.acp_sdf, $CPSBIN/cps_sdf\n");
	/* sdf lines can be in the JDF file */
	return;

	parse_sdf:
	if ((fp = freopen(fname,"r",stdin)) == NULL)
	{
		printf("Failed to redirect stdin to the cps_sdf file\n");
		exit(0);
	}
	yyin = fp;
	yyout = stdout;
	yyparse();

	k = numerr;
	if (numerr > 0)
	{
		printf("Warning: Errors found in the cps_sdf file\n");
		i = 0;
		j = 0;
		rewind(fp);
	}
	while (k > 0)
	{
		if (fgets(buf,1024,fp) == NULL) break;
		i++;
		if (i == errlin[j])
		{
			printf("    line %d: %s",i,buf);
			j++;
			k--;
		}
	}
	numerr = 0;
	lineno = 1;

	if (TEST_LEVEL == 3)
		print_sdf_cpus();

}

/******************************************************************/
/*
	FUNCTION NAME:

	  parse_jdf

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

parse_jdf()
{
	FILE *fp;

	char buf[1024];
	int i,j,k;

	if ((fp = freopen(jdf_file,"r",stdin)) == NULL) {
		fprintf(jml,"Failed to redirect stdin to the jdf file\n");
		fprintf(jml,"Check to make sure that the JDF (%s) exists\n",jdf_file);
		printf("Failed to redirect stdin to the jdf file\n");
		printf("Check to make sure that the JDF (%s) exists\n",jdf_file);
		exit(0);
	}
	yyin = fp;
	yyout = stdout;
	yyparse();

	k = numerr;
	if (numerr > 0) {
		fprintf(jml,"Errors found in the JDF file\n");
		printf("Errors found in the JDF file\n");
		i = 0;
		j = 0;
		rewind(fp);
	}
	while (k > 0) {
		if (fgets(buf,1024,fp) == NULL)
			break;
		i++;
		if (i == errlin[j]) {
			fprintf(jml,"    line %d: %s",i,buf);
			printf("    line %d: %s",i,buf);
			j++;
			k--;
		}
	}

	if (TEST_LEVEL == 4)
		print_class_info();

	if (numerr > 0)
		exit(0);

/* Redirect input back to the terminal (if not in batch mode) */

	if (!batch_mode && (tty_name[0] != '\0')) {
		if ((fp = freopen(tty_name,"r",stdin)) == NULL) {
			printf("Failed to redirect stdin back to the terminal\n");
			exit(0);
   		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  consistency_check

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

								  */
/******************************************************************/

consistency_check()
{
	int index;
	int num_classes = 0;
	char string[128];

	for( index = 1; index <= MAX_CLASSES; index++)
	{
		if ( class[ index ].valid )
		{
			num_classes++;
			if( ( class[ index ].program == NULL ) ||
				( strlen( class[ index ].program ) == 0 ) )
			{
				sprintf( string, "%s %d %s\n",
					"No Processes Specified for class", index, "in JDF file" ); 
				printf( string );
				fprintf( jml, string );
				exit(0);
			}
		}
	}
	if( num_classes == 0 )
	{
		sprintf( string, "%s\n",
			"No Classes Specified in JDF file" );
		printf( string );
		fprintf( jml, string );
		exit(0);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  initialize_queues

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

								  */
/******************************************************************/

initialize_queues()
{

	int i,j;

	zfree = cps_create_queue();
	for (i=0; i<MAX_QUEUES; i++) {
		jmq[i].active = 0;
		jmq[i].declared = create_empty_set(MAX_PROCESSES);
		jmq[i].wait_full = create_empty_set(MAX_PROCESSES);
		jmq[i].wait_empty = create_empty_set(MAX_PROCESSES);
		jmq[i].q_set = create_empty_set(MAX_PROCESSES);
		jmq[i].dq_set = create_empty_set(MAX_PROCESSES);
		jmq[i].declared = create_empty_set(MAX_PROCESSES);
		jmq[i].queued = cps_create_queue();
		jmq[i].free = cps_create_queue();
		jmq[i].dequeue = cps_create_queue();
		jmq[i].nargs = 0;
		jmq[i].bcount = 0;
		for (j=0; j<MAX_QUEUE_ARGS; j++)
			jmq[i].argcnt[j] = 0;
		jmq[i].eoq = 0;
		queue_stop[i] = 0;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  initialize_syncs

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

initialize_syncs()
{
	int i;

	for (i=0; i<MAX_SYNCS; i++) {
		need_sync[i] = NULL;
		at_sync[i] = NULL;
		sync_active[i] = 0;
		sync_stop[i] = 0;
	}

	for (i=0; i<num_usr_procs; i++)
		proc_sync[i] = -1;

}

initialize_hooks()
{
	int i;

	for( i = 0; i < MAX_USER_HOOKS; i++ )
		user_stop[i] = 0;
	for( i = 0; i < MAX_FUNC_HOOKS; i++ )
		proc_stop[i] = 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  main

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

main(argc,argv)
int argc;
char **argv;
{
	int i,j;

	char *c;

	char banner[1024];

	time_t t;

	FILE *fp,*fopen();
	char fname[80],*home,syscmd[80];
	char *getenv();

/* Turn on periodic memory allocation reporting */
/*
	hm_report = -1;
	hm_op_interval = 1000;
*/
/* Parse command line options */

	parse_command_line(argc,argv);
	if (TEST_LEVEL == 1)
		exit(0);

/* Get the hostname */

	jm_hostname = (char *)hmalloc( 80 );
	gethostname( jm_hostname, 80 );

/* Get the terminal name (if not in batch mode) */

	if (!batch_mode) {
		c = ttyname(0);
		if (c == NULL)
			tty_name[0] = '\0';
		else
			strcpy(tty_name,c);
	}

/* Invoke the pre-processor if necessary */

	if (invoke_cpp)
		run_cpp();
	if (TEST_LEVEL == 2)
		exit(0);

/* Open log files */

	jml = fopen(log_file,"w");
	if (jml == NULL)
	{
		printf("Failed to open log file %s for writing\n", log_file);
		exit(0);
	}
	logfp = jml;

/* Make sure portmapper is running */
	contact_portmapper();

/* Get resources from cps_sdf */

	initialize_logical_cpus();

/* Connect to a PM if we are in batch mode or if the -s option was used */

	if (pm_connect_flag || batch_mode)
		pm_init();

	if (!batch_mode)
		get_resources();

	if (TEST_LEVEL == 3)
		exit(0);

/* Print startup message */

	sprintf(banner,
"*******************************************************************\n\
*                                                                 *\n\
* CPS Job Manager Version %-39s *\n\
* Copyright (c) 1992,1993 Universities Research Association, Inc. *\n\
*                                                                 *\n\
* All rights reserved.                                            *\n\
* Fermi National Accelerator Laboratory                           *\n\
*                                                                 *\n\
* %-63s *\n\
*                                                                 *\n\
*******************************************************************\n",
 	cps_version, cps_build_date);

	printf("%s",banner);
	fprintf(jml,"%s",banner);

/* Initialize share classes */

	for (i=1; i<=MAX_CLASSES; i++) {
		class[i].share = create_empty_set(MAX_CLASSES);
		for (j=1; j<=MAX_CLASSES; j++)
			add_to_set(j,class[i].share);
	}

/* Parse JDF file */

	parse_jdf();
	if (TEST_LEVEL == 4)
		exit(0);

/* Do a consistency check on the job information */

	consistency_check();

/* Assign CPUs */

	assign_cpus();
	if (TEST_LEVEL == 5) {
		print_cpu_assignments();
		exit(0);
	}

/* Print process->class->cpu map */

	printf("Process Map:  Process   Class   CPU\n");
	printf("              -------   -----   ---\n");
	for (i=0; i<num_usr_procs; i++) {
		printf("           %7d      %3d     %s\n",i+FIRST_USER_LPN,
		usr_map[i].class,sdf_cpu[usr_map[i].cpu].cpu_name);
	}
	printf("\n");
	time(&t);
	printf("%s",ctime(&t));

	fprintf(jml,"Process Map:  Process   Class   CPU\n");
	fprintf(jml,"              -------   -----   ---\n");
	for (i=0; i<num_usr_procs; i++) {
		fprintf(jml,"           %7d      %3d     %s\n",i+FIRST_USER_LPN,
		usr_map[i].class,sdf_cpu[usr_map[i].cpu].cpu_name);
	}
	fprintf(jml,"\n");
	time(&t);
	fprintf(jml,"%s",ctime(&t));
	jm_starttime = (long)t;
	if (TEST_LEVEL == 6)
		exit(0);

/* Redirect output to a file (for debugging) */
/*
	freopen("jm_debug.log","w",stdout);
*/
/* Ignore interrupts until later */

	sig_establish_handler(SIGINT,SIG_IGN);

/* Ignore SIGPIPE */

	sig_establish_handler(SIGPIPE,SIG_IGN);

/* Check to see if a .netrc file exists */

	home = getenv("HOME");
	sprintf(fname,"%s/.netrc",home);
	fp = fopen(fname,"r");
	if (fp != NULL) {
		netrc_exists = -1;
		fclose(fp);
	}

/* Initialize the queues */

	initialize_queues();

/* Initialize syncs */

	initialize_syncs();

/* Initialize Hooks */

	initialize_hooks();

/* Initialize user statuses */

	bzero(user_status,MAX_PROCESSES*MAX_USER_STATUS*sizeof(int));

/* Initialize pointers to data ready blocks */

	for (i=0; i<MAX_PROCESSES; i++) {
		acp_dblocks[i] = NULL;
	}

/* Start the processes */
	start_proc();
	if (TEST_LEVEL == 6) {
		printf("Processes sucessfully started\n");
		sleep(5);
		exit(0);
	}

/* Initialize the monitor shared memory */

	init_mon_shm();

/* Create RPC service */
	jm_rpc_init();

/* Upshot Logging */
	if( upshot )
		upshot_start();

/* Initialize profiling */
	if( jm_prof )
		jm_profile_init();

/* Set up a CTRL-C handler */

	sig_establish_handler(SIGINT,sigint_handler);

	printf("All processes started, CTRL-C now enabled\n");
	fflush(stdout);
	fprintf(jml,"All processes started, CTRL-C now enabled\n");
	fflush(jml);

/* Write file to kill all processes & ipc objects */

	sprintf(fname,"%s.kill",job_name);
	abort_file = (char *) hmalloc(strlen(fname)+1);
	strcpy(abort_file,fname);
	fp = fopen(fname,"w");
	fprintf(fp,"#!/bin/csh -f\n");
	fprintf(fp,"ipcrm -m %d\n",mon_shmid);
	for (i=0; i<num_usr_procs; i++) {
#ifdef HPUX
		fprintf(fp,"remsh %s kill -9 %d\n",
		shm_map[usr_map[i].shm].hostname,usr_map[i].pid);
#else
		fprintf(fp,"rsh %s kill -9 %d\n",
		shm_map[usr_map[i].shm].hostname,usr_map[i].pid);
#endif
	}
	for (i=0; i<num_shm_procs; i++) {
		if (cpu_name_compare(shm_map[i].hostname,jm_hostname) == 0) {
			fprintf(fp,"ipcrm -m %d\n",shm_map[i].shmid);
			fprintf(fp,"ipcrm -s %d\n",shm_map[i].semid);
		}
		else {
#ifdef HPUX
			fprintf(fp,"remsh %s ipcrm -m %d\n",
			shm_map[i].hostname,shm_map[i].shmid);
#else
			fprintf(fp,"rsh %s ipcrm -m %d\n",
			shm_map[i].hostname,shm_map[i].shmid);
#endif
#ifdef HPUX
			fprintf(fp,"remsh %s ipcrm -s %d\n",
			shm_map[i].hostname,shm_map[i].semid);
#else
			fprintf(fp,"rsh %s ipcrm -s %d\n",
			shm_map[i].hostname,shm_map[i].semid);
#endif
		}
	}
	fclose(fp);
	sprintf(syscmd,"chmod u+x %s",fname);
	system(syscmd);

/* Manage the job */

	alarm(60);

	manage_job();
}
