/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

#include "all.h"
#include "../inc/types.h"
#include "../inc/time.h"

#define FORMATA "%5d %5d %5d %5d %5d %16d\n"
#define FORMATB "%5d %5d %5d %5d %5d %16d %-16s\n"
#define FORMATC "%5d %5d %5d %5d %5d %16d %s; %s"

#define START	98
#define END		99

static FILE *eventfp				= NULL;
static FILE *upshotfp				= NULL;
static int num_events;
static int num_classes				= 0;
static int num_event_types			= 9;
static int start_time				= 0;
static int end_time 				= 0;
static int num_timer_cycles			= 1;
static int rollover_point			= 0;
static long timestamp;
static short reset_timer			= 0;
static char upshot_on				= 0;

void
upshot_print_event( message )
cps_message *message;
{
	struct timeval timevalue;
	static int time_offset;
	int i;
	long *x;

	gettimeofday( &timevalue, 0 );

	if( num_events == 0 )
	{
		time_offset = timevalue.tv_sec;
		/*
		start_time = timevalue.tv_usec;
		*/
		start_time = timevalue.tv_usec - 10000;
		for( i = 1; i <= num_usr_procs; i++ )
		{
			num_events++;
			fprintf( eventfp, FORMATA, START, i, 0, 0, 0, start_time );
		}
	}
	else if( reset_timer )
	{
		time_offset = timevalue.tv_sec - ( timestamp / 1000000 + 1 );
		reset_timer = 0;
	}

	timestamp = 1000000 * ( timevalue.tv_sec - time_offset )
		+ timevalue.tv_usec;

	end_time = timestamp;
	num_events++;

	if( message->type == MSG_FUNC_HOOK )
	{
		x = (long *) message->msg;
		fprintf( eventfp, FORMATA, message->type,
			message->src_lpn, 0, *x, 0, timestamp );
	}
	else
	{
		fprintf( eventfp, FORMATA, message->type,
			message->src_lpn, 0, 0, 0, timestamp );
	}
}

upshot_stop()
{
	char *upshot_file = NULL;
	char creation_date[26];
	time_t today;
	int  field[ 6 ];
	static int reply = 1;
	int i;

	if( !upshot_on )
		return;

	upshot_file = (char *)hmalloc( strlen( job_name ) + 4 );
	sprintf( upshot_file, "%s.up", job_name );
	unlink( upshot_file );
	upshotfp = fopen( upshot_file, "w" );

	for( i = 0; i < MAX_CLASSES; i++ )
		if( class[ i ].valid )
			num_classes++;

	for( i = 1; i <= num_usr_procs; i++ )
		num_events++;

	end_time = end_time + 10000;

	today = time(0);
	strcpy( creation_date, ctime( &today ) );

	fprintf( upshotfp, FORMATC, -1, 0, 0, 0, 0, 0, "CPS", creation_date );
	fprintf( upshotfp, FORMATA, -2, 0, 0, num_events, 0, 0 );
	fprintf( upshotfp, FORMATA, -3, 0, 0, num_usr_procs, 0, 0 );
	fprintf( upshotfp, FORMATA, -4, 0, 0, num_classes, 0, 0 );
	fprintf( upshotfp, FORMATA, -5, 0, 0, num_event_types, 0, 0 );
	fprintf( upshotfp, FORMATA, -6, 0, 0, 0, 0, start_time );
	fprintf( upshotfp, FORMATA, -7, 0, 0, 0, 0, end_time );
	fprintf( upshotfp, FORMATA, -8, 0, 0, 1, 0, 0 );

/*
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_DECLARE_QUEUE, 0, 0, "Decl Queue" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_DECLARE_QUEUE, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_QUEUE_PROCESS, 0, 0, "Queue Proc" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_QUEUE_PROCESS, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_DEQUEUE_PROCESS, 0, 0, "Dequeue Proc" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_DEQUEUE_PROCESS, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_WAIT_SYNC, 0, 0, "Wait Sync" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_WAIT_SYNC, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_SUB_DECL, 0, 0, "Decl Subr" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_SUB_DECL, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_DECLARE_BLOCK, 0, 0, "Decl Block" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_DECLARE_BLOCK, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_USER_STATUS, 0, 0, "User Status" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_USER_STATUS, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_USER_HOOK, 0, 0, "User Hook" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_USER_HOOK, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, MSG_FUNC_HOOK, 0, 0, "Proc Hook" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, MSG_FUNC_HOOK, 0, 0, "" );
*/
	fprintf( upshotfp, FORMATB, -9, 0, 0, START, 0, 0, "Start" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, START, 0, 0, "" );
	fprintf( upshotfp, FORMATB, -9, 0, 0, END, 0, 0, "End" );
	fprintf( upshotfp, FORMATB, -10, 0, 0, END, 0, 0, "" );

	if( eventfp != NULL )
	{
		for( i = 1; i <= num_usr_procs; i++ )
			fprintf( eventfp, FORMATA, END, i, 0, 0, 0, end_time );
		fflush( eventfp );
		rewind( eventfp );
		while( fscanf( eventfp, FORMATA,
			&field[0], &field[1], &field[2], &field[3], &field[4], &field[5] )
			!= EOF )
		{
			fprintf( upshotfp, FORMATA,
				field[0], field[1], field[2], field[3], field[4], field[5] );
		}
	}

	fclose( upshotfp );
	reset_timer = 1;

	/*
	printf("jm[%s] Upshot logging turned OFF\n",jm_hostname);
	*/
	fprintf(jml, "jm[%s] Upshot logging turned OFF\n",jm_hostname);
	fflush(jml);

	upshot = 0;
	upshot_on = 0;
}

upshot_start()
{
	if( upshot_on )
		return;

	if( eventfp != NULL )
		fclose( eventfp );

	eventfp = tmpfile();

	num_events = 0;
	/*
	printf("jm[%s] Upshot logging turned ON\n",jm_hostname);
	*/
	fprintf(jml, "jm[%s] Upshot logging turned ON\n",jm_hostname);
	fflush(jml);

	upshot = 1;
	upshot_on = 1;
}
