/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
#include "jmdb.h"
#define MAIL_FORMAT "\"%s\\n\\n%s\\n\\n %-11s %s\\n %-11s %s %-11s %s\\n %-11s %s\\n\""

extern ClientInfo client;
extern JmInfo jm;

char monitor_mode = 0;
SVCXPRT *transp;
int sock;
int prognum;
int vers = 1;
unsigned alarm_time;
char callback_done = 0;

/******************************************************************/
/*
	FUNCTION NAME:

		sigalrm_handler

	DESCRIPTION:

		alarm handler for job monitor

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
sigalrm_handler( sig, code, sc )
int sig, code;
struct sigcontext *sc;
{
	if( !callback_done )
		monitor_exit( JOB_DISCONNECTED );
	callback_done = 0;
	alarm( alarm_time );
}

/******************************************************************/
/*
	FUNCTION NAME:

		callback

	DESCRIPTION:

		callback subroutine called by JM

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void
callback( rqstp, transp )
register struct svc_req *rqstp;
register SVCXPRT *transp;
{
	sig_establish_handler( SIGALRM, SIG_IGN );

	switch( rqstp->rq_proc )
	{
	case NULLPROC:
		break;
	case CALLBACK:
		callback_done = 1;
		break;
	case END_OF_JOB:
		break;
	default:
		break;
	}

	if( !svc_sendreply( transp, xdr_void, 0 ) )
		svcerr_systemerr(transp);

	if( rqstp->rq_proc == END_OF_JOB )
		monitor_exit( JOB_FINISHED );

	sig_establish_handler( SIGALRM, sigalrm_handler );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_monitor

	DESCRIPTION:

		initialize job monitor

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_monitor( interval )
int interval;
{
	client_info caller;

	if( !connected( client.cl ) ) return;

	if( sig_establish_handler( SIGALRM, sigalrm_handler ) )
	{
		fprintf( stdout, STR_SIGNAL_HANDLER_ERR, "SIGALRM" );
		return;
	}

	sock = RPC_ANYSOCK;

	prognum = gettransient( IPPROTO_TCP, vers, &sock );

	transp = svctcp_create( sock, 1024, 1024 );
	if( transp == NULL )
	{
		fprintf( stdout, STR_TCP_CREATE_ERR );
		pmap_unset( prognum, vers );
		close( sock );
		return;
	}
	if( !svc_register( transp, prognum, vers, callback, 0 ) )
	{
		fprintf( stdout, STR_TCP_REGISTER_ERR );
		pmap_unset( prognum, vers );
		close( sock );
		svc_destroy( transp );
		return;
	}

	caller.prognum = prognum;
	caller.vers = vers;
	caller.interval = ( interval * 60 ) / 2;
	strcpy( caller.hostname, client.hostname );
	if( !*rpc_initiate_callback_1( &caller, client.cl ) )
	{
		fprintf( stdout, STR_CALLBACK_ERR );
		pmap_unset( prognum, vers );
		close( sock );
		svc_destroy( transp );
		return;
	}

	monitor_mode = 1;

	fprintf( stdout, STR_JMDB_MONITOR, jm.username, client.hostname );

	alarm_time = (unsigned)( interval * 60 );
	alarm( alarm_time );

	svc_run();
}

/******************************************************************/
/*
	FUNCTION NAME:

		monitor_exit

	DESCRIPTION:

		terminate job monitor

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
monitor_exit( message )
char *message;
{
	char mail_message[1024];

	if( monitor_mode )
	{
		alarm( 0 );
		sig_establish_handler( SIGALRM, SIG_DFL );

		monitor_mode = 0;

		/*
		rpc_terminate_callback_1( NULL, client.cl );
		*/

		if( strcmp( message, MONITOR_INTERRUPTED ) )
		{
			fprintf( stdout, "%s\n", message );
			destroy_rpc_client( client.cl );
			client.prognum = 0;
			client.cl = NULL;
			fprintf( stdout, STR_CONNECTION_CLOSED );

			sprintf( mail_message, MAIL_FORMAT,
			    "JMDB Monitor Message",
			    message,
			    "[job]", jm.jobname,
			    "[start]", ctime( &jm.starttime ),
			    "[hostname]", jm.hostname,
			    "[version]", jm.version );

			mail_to_user( jm.username, mail_message );
		}
		else
		{
			fprintf( stdout, "%s\n", message );
			rpc_terminate_callback_1( NULL, client.cl );
		}


		pmap_unset( prognum, vers );
		close( sock );
		svc_destroy( transp );
		siglongjmp( env, 1 );
	}
}
