
%{

#include "y.tab.h"
#include "strings.h"

#define save savetok( yytext );

/*
#define parse( token );\
if( is_alias( yytext ) )\
	return( NULL_TOKEN );\
else\
{\
	savetok( yytext );\
	return( token );\
}
*/

#define parse( token );\
if( !is_alias( yytext ) )\
{\
	savetok( yytext );\
	return( token );\
}

extern int yylval;
extern int txt_ptr;
char word[128];
char *token[4] = { NULL, NULL, NULL, NULL };
int num_of_toks;
int n1, n2;
unsigned long *process_set = NULL;

%}

%e 700
%p 3000
%n 350
%k 60
%a 10000
%o 18000

ws			[ \t]
cr			[\n]
digit		[0-9]
letter		[a-zA-Z]
character	["#$%&'()*+,\-./:;<=>?@\[\\\]^_`\{|\}~a-zA-Z0-9]
range		{digit}+{ws}?-{ws}?{digit}+

%%

acp_mon			{ parse( CPS_MON ); }
alias			{ parse( ALIAS ); }
all				{ parse( ALL ); }
at				{ parse( AT ); }
cat				{ parse( CAT ); }
^cd				{ parse( CD ); }
commands		{ parse( COMMANDS ); }
connect			{ parse( CONNECT ); }
cont			{ parse( CONT ); }
coredir			{ parse( COREDIR ); }
cps_mon			{ parse( CPS_MON ); }
dbx				{ parse( DBX ); }
dbxcopy			{ parse( DBXCOPY ); }
dbxrun			{ parse( DBXRUN ); }
dbgdir			{ parse( DBGDIR ); }
debug			{ parse( DEBUG ); }
delete			{ parse( DELETE ); }
disconnect		{ parse( DISCONNECT ); }
display			{ parse( DISPLAY ); }
exit			{ parse( EXIT ); }
file			{ parse( FILENAME ); }
help			{ parse( HELP ); }
history			{ parse( HISTORY ); }
hook			{ parse( HOOK ); }
hooks			{ parse( HOOKS ); }
hosts			{ parse( HOSTS ); }
import			{ parse( IMPORT ); }
jm				{ parse( JM ); }
jml				{ parse( JML ); }
jminfo			{ parse( JMINFO ); }
job				{ parse( JOB ); }
kill			{ parse( KILL ); }
objdir			{ parse( OBJDIR ); }
on				{ parse( ON ); }
off				{ parse( OFF ); }
output			{ parse( OUTPUT ); }
hooks			{ parse( HOOKS ); }
monitor			{ parse( MONITOR ); }
pager			{ parse( PAGER ); }
process			{ parse( PROCESS ); }
prompt1			{ parse( PROMPT1 ); }
prompt2			{ parse( PROMPT2 ); }
^pwd			{ parse( PWD ); }
queue			{ parse( QUEUE ); }
queues			{ parse( QUEUES ); }
quit			{ parse( QUIT ); }
quit!			{ parse( QUITFORCE ); }
rexec			{ parse( REXEC ); }
set				{ parse( SET ); }
sh				{ parse( SH ); }
show			{ parse( SHOW ); }
start			{ parse( START ); }
state			{ parse( STATE ); }
status			{ parse( STATUS ); }
step			{ parse( STEP ); }
stop			{ parse( STOP ); }
sync			{ parse( SYNC ); }
syncs			{ parse( SYNCS ); }
tail			{ parse( TAIL ); }
trace			{ parse( TRACE ); }
unalias			{ parse( UNALIAS ); }
unset			{ parse( UNSET ); }
update			{ parse( UPDATE ); }
upshot			{ parse( UPSHOT ); }
view			{ parse( VIEW ); }
where			{ parse( WHERE ); }
refresh			{ parse( REFRESH ); }
{range}			{ save return( get_range( yytext ) ); }
{digit}+		{ save return( get_number( yytext ) ); }
{character}+	{ parse( get_word( yytext ) ); }
!!				{ return( BANGBANG ); }
!-?{digit}+		{ yylval = atoi( &yytext[1] ); return( BANGNUMBER ); }
!{letter}+		{ strcpy( word, &yytext[1] ); return( BANGSTRING ); }
{cr}			{ return(CR); }
{ws}+			;
.				{ save return(UNDEFINED); }

%%

get_word( string )
char *string;
{
	strcpy( word, string );
	return( WORD );
}

get_number( string )
char *string;
{
	yylval = atoi( string );
	return( NUMBER );
}

get_range( string )
char *string;
{
	n1 = atoi( string );
	n2 = atoi( strchr( string, '-' ) + 1 );
	return( RANGE );
}

int
gettok()
{
	int ch;
	int tok_len;
	char quotes = 0;

	tok_len = 0;

	while( ( ch = input() ) == ' ' );

	if( ch == '\n' )
	{
		unput('\n');
		return( 0 );
	}

	if( ch == '"' )
		quotes = 1;
	else
		yytext[ tok_len++ ] = ch;

	if( quotes )
		while( ( ( ch = input() ) != '"' ) && ( ch != '\n' ) )
			yytext[ tok_len++ ] = ch;
	else
		while( ( ( ch = input() ) != ' ' ) && ( ch != '\n' ) )
			yytext[ tok_len++ ] = ch;

	if( quotes && ( ch == '\n' ) )
	{
		num_of_toks = -1;
		fprintf( stdout, STR_NON_TERMINATED_STRING );
		unput('\n');
		return( 0 );
	}

	yytext[ tok_len++ ] = 0;

	if( token[ num_of_toks ] != NULL )
		free( token[ num_of_toks ] );
	token[ num_of_toks ] = (char *)malloc( tok_len );
	if( token[ num_of_toks ] == NULL )
	{
		fprintf( stdout, STR_MALLOC_ERR );
		abort();
	}

	strcpy( token[ num_of_toks ], yytext );

	if( quotes )
	{
		savetok( "\"" );
		txt_ptr--;
	}

	savetok( yytext );

	if( quotes )
	{
		txt_ptr--;
		savetok( "\"" );
	}

	num_of_toks++;

	if( ch == '\n' )
		unput('\n');

	return( 1 );
}

lex_unput( command )
char *command;
{
	int index;

	for( index = strlen( command ) - 1; index >= 0; index-- )
		unput( command[index] );
}
