
char xcps_mon = 1;

#include "xcps_mon.h"
#include <stdio.h>

#include "../inc/param.h"

#define MAX_PROCESSES 256
#define MAX_COLUMNS   16
#define MAX_STATES    16
#define TITLE         "xcps_mon 1.0"
#define ICON          "xcps_mon"

Widget TheTop;
Widget MainWindow;
Widget MainScroll;
Widget MainForm;
Widget WorkScroll;
Widget WorkForm;
Widget Header;
Widget Footer;
Widget MenuBar;

Widget ColumnHeader[ MAX_COLUMNS ];
Widget LabelWidget[ MAX_PROCESSES ];
Widget DataWidget[ MAX_COLUMNS ][ MAX_PROCESSES ];
Pixmap DataPixmap[ MAX_PROCESSES ];

Widget Legend;
Widget LegendWidget[ MAX_STATES ];
Widget LegendStringWidget[ MAX_STATES ];
int LegendWidgetColor[ MAX_STATES ];

Widget file_menu, monitor_menu, options_menu;

Widget PollDialog;
Widget PollTextField;

Widget CurrentMonitorWidget;

Dimension MaxHeaderWidth;

Dimension GraphWidgetHeight;
Dimension GraphWidgetWidth;

Pixel BackgroundColor;
Pixel ForegroundColor;

int NumProcs;
int NumColumnHeaders;
int NumLegendWidgets;

XtAppContext app;
Display      *display;
Screen       *screen;
GC           gc;

/*
void XcpsmonHeader();
void XcpsmonFooter();
void XcpsmonGraphLegend();
void XcpsmonPollDialog();
*/

typedef struct
{
	char *monitor;
	char *hostname;
	int  id;
	int  proc;
	int  class;
	int  interval;
} AppData;

AppData app_data;

XtResource resources[] =
{
	{
		"monitor",
		"Monitor",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,monitor),
		XmRImmediate,
		"calls"
	},
	{
		"hostname",
		"Hostname",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,hostname),
		XmRImmediate,
		"localhost"
	},
	{
		"id",
		"Id",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,id),
		XmRImmediate,
		(XtPointer) 0
	},
	{
		"proc",
		"Proc",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,proc),
		XmRImmediate,
		(XtPointer) -1
	},
	{
		"class",
		"Class",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,class),
		XmRImmediate,
		(XtPointer) -1
	},
	{
		"interval",
		"interval",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,interval),
		XmRImmediate,
		(XtPointer) 5
	}
};

XrmOptionDescRec options[] =
{
	{"-monitor",   "*monitor",  XrmoptionSepArg, NULL },
	{"-m",         "*monitor",  XrmoptionSepArg, NULL },
	{"-d",         "*monitor",  XrmoptionNoArg,  "deltas" },
	{"-u",         "*monitor",  XrmoptionNoArg,  "user" },
	{"-f",         "*monitor",  XrmoptionNoArg,  "calls" },
	{"-s",         "*monitor",  XrmoptionNoArg,  "states" },
	{"-q",         "*monitor",  XrmoptionNoArg,  "queues" },
	{"-t",         "*monitor",  XrmoptionNoArg,  "rates" },
	{"-g",         "*monitor",  XrmoptionNoArg,  "graph" },
	{"-hostname",  "*hostname", XrmoptionSepArg, NULL },
	{"-h",         "*hostname", XrmoptionSepArg, NULL },
	{"-a",         "*hostname", XrmoptionSepArg, NULL },
	{"-id",        "*id",       XrmoptionSepArg, NULL },
	{"-i",         "*id",       XrmoptionSepArg, NULL },
	{"-fn",        "*fontList", XrmoptionSepArg, NULL },
	{"-font",      "*fontList", XrmoptionSepArg, NULL },
	{"-p",         "*proc",     XrmoptionSepArg, NULL },
	{"-c",         "*class",    XrmoptionSepArg, NULL },
	{"-interval",  "*interval", XrmoptionSepArg, NULL },
};

String colors[] =
{
	"Black", "Blue", "", "Green", "Red", "Cyan", "Yellow",
	"", "", "", "Purple", "", "", ""
};

String names[] =
{
	"dead", "call", "", "send", "get", "queue", "dequeue",
	"", "", "", "service_calls", "", ""
};

String fallbacks[] =
{
	"xcps_mon*fontList: -adobe-helvetica-medium-r-normal--14-140-75-75-p-77-iso8859-1",
	NULL
};

int Timeout;
int OldTimeout;

/* externs */
extern int mon_type;
extern void handle_alarm();

Dimension GraphFormOffset()
{
	Dimension width, total_width = 0;
	Position position;

	XtVaGetValues( LegendStringWidget[ NumLegendWidgets - 1 ],
		XmNx,     &position,
		XmNwidth, &width,
		NULL );

	total_width = position + width + 10;

	return( total_width );
}

Dimension StateFormOffset()
{
	Dimension width, total_width = 0;
	Position position;

	XtVaGetValues( ColumnHeader[ 1 ],
		XmNx,     &position,
		XmNwidth, &width,
		NULL );

	total_width = position + 3 * width + 10;

	return( total_width );
}

Dimension WindowWidthOffset()
{
	Dimension width, total_width = 0;

	total_width += 30;

	return( total_width );
}

Dimension WindowHeightOffset()
{
	Dimension height, total_height = 0;

/* height of Footer */
	XtVaGetValues( Footer, XmNheight, &height, NULL );
	total_height += height;

/* height of MenuBar */
	XtVaGetValues( MenuBar, XmNheight, &height, NULL );
	total_height += height;

/* Offset between MenuBar and MainScroll */
	total_height += 10;

/* Offset between Footer and MainScroll */
	total_height += 5;

/* Height of scrollbar and margin */
	total_height += 20;

/* Height of Window Border? */
	total_height += 20;

	return( total_height );
}

TheTopResizeCallback( widget, data, event )
Widget widget;
XtPointer data;
XConfigureEvent *event;
{
	Dimension width, height;

	if( event->type != ConfigureNotify )
		return;
	height = event->height - WindowHeightOffset();

	XtVaSetValues( MainForm, XmNheight, height, NULL );

	if( ( mon_type == MON_DELTA_GRAPH ) || ( mon_type == MON_STATES ) )
	{
		int index;
		Position position;

		width = event->width - WindowWidthOffset();
		if( ( mon_type == MON_DELTA_GRAPH ) && ( width < GraphFormOffset() ) )
			width = GraphFormOffset();
		if( ( mon_type == MON_STATES ) && ( width < StateFormOffset() ) )
			width = StateFormOffset();

		XtVaSetValues( MainForm, XmNwidth, width, NULL );

		XtVaGetValues( ColumnHeader[ 1 ], XmNx, &position, NULL );
		GraphWidgetWidth = width - position - 40;
		for( index = 0; index < NumProcs; index++ )
			XtVaSetValues( DataWidget[ 0 ][ index ],
				XmNwidth, GraphWidgetWidth, NULL );
	}
}

WorkFormResizeCallback( widget, data, event )
Widget widget;
XtPointer data;
XConfigureEvent *event;
{
	Dimension width, height;

	if( event->type != ConfigureNotify )
		return;

	XtVaGetValues( TheTop, XmNheight, &height, NULL );
	height = height - WindowHeightOffset();

	XtVaSetValues( MainForm, XmNheight, height, NULL );

	if( ( mon_type == MON_DELTA_GRAPH ) || ( mon_type == MON_STATES ) )
	{
		int index;
		Position position;

		XtVaGetValues( TheTop, XmNwidth, &width, NULL );

		width = width - WindowWidthOffset();
		if( ( mon_type == MON_DELTA_GRAPH ) && ( width < GraphFormOffset() ) )
			width = GraphFormOffset();
		if( ( mon_type == MON_STATES ) && ( width < StateFormOffset() ) )
			width = StateFormOffset();

		XtVaSetValues( MainForm, XmNwidth, width, NULL );

		XtVaGetValues( ColumnHeader[ 1 ], XmNx, &position, NULL );
		GraphWidgetWidth = width - position - 40;
		for( index = 0; index < NumProcs; index++ )
			XtVaSetValues( DataWidget[ 0 ][ index ],
				XmNwidth, GraphWidgetWidth, NULL );
	}
}

FillGraphWidget( widget, data, cbs )
Widget widget;
XtPointer data;
XmDrawingAreaCallbackStruct *cbs;
{
	int index;
	for( index = 0; index < NumProcs; index++ )
		if( widget == DataWidget[ 0 ][ index ] )
		{
			XCopyArea( display, DataPixmap[ index ],
				cbs->window, gc, 0, 0,
				GraphWidgetWidth, GraphWidgetHeight, 0, 0 );
			break;
		}
}

FillLegendWidget( widget, data, cbs )
Widget widget;
XtPointer data;
XmDrawingAreaCallbackStruct *cbs;
{
	int index;
	for( index = 0; index < NumLegendWidgets; index++ )
		if( widget == LegendWidget[ index ] )
		{
			SetColor( display, DefaultScreen( display ),
				colors[ LegendWidgetColor[ index ] ], gc );
			XFillRectangle( display, cbs->window,
				gc, 0, 0, GraphWidgetHeight, GraphWidgetHeight );
			break;
		}
}

FileCallback( w, data )
Widget w;
int data;
{
	switch( data )
	{
		case 0:
			exit( 0 );
			break;
		default:
			break;
	}
}

MonitorCallback( widget, data )
Widget widget;
int data;
{
	if( mon_type != data )
	{
		Dimension height;
		XtVaSetValues( CurrentMonitorWidget, XmNset, False ,NULL );
		CurrentMonitorWidget = widget;
		mon_type = data;
		mon_init();
		mon_refresh();
		force_updates();
	}
	else
		XtVaSetValues( widget, XmNset, True ,NULL );
}

OptionsCallback( w, data )
Widget w;
int data;
{
	switch( data )
	{
		case 0:
			XcpsmonPollDialog( w );
			break;
		default:
			break;
	}
}

XcpsmonDeactivateMenus()
{
	int index;

	XtSetSensitive( monitor_menu, False );
	XtSetSensitive( options_menu, False );
	if( mon_type == MON_DELTA_GRAPH )
		for( index = 0; index < NumProcs; index++ )
		{
			XcpsmonUpdateProcessGraph( 0, index, 0, -1, 0 );
			XcpsmonUpdateProcessGraph( 0, index, 0, 999, 0 );
		}
}

XcpsmonMenus()
{
	Widget widget;
	XmString xm_string;
	int type;

/* Create MenuBar */
	MenuBar = XmCreateMenuBar( MainWindow, "MenuBar", NULL, 0 );
	XtVaSetValues( MenuBar,
		XmNleftAttachment,  XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		NULL );

/* Create Menu Headings */
	file_menu = XmCreatePulldownMenu( MenuBar, "FileMenu", NULL, 0 );
	monitor_menu = XmCreatePulldownMenu( MenuBar, "MonitorMenu", NULL, 0 );
	options_menu = XmCreatePulldownMenu( MenuBar, "OptionsMenu", NULL, 0 );

/* File Menu */
	xm_string = XmStringCreateSimple( "File" );
	XtVaCreateManagedWidget( "File",
		xmCascadeButtonWidgetClass, MenuBar,
		XmNlabelString,             xm_string,
		XmNmnemonic,                'F',
		XmNsubMenuId,               file_menu,
		NULL );
	XmStringFree( xm_string );

/* Exit */
	xm_string  = XmStringCreateSimple( "Ctrl-X" );
	widget = XtVaCreateManagedWidget( "Exit",
		xmPushButtonWidgetClass, file_menu,
		XmNmnemonic,             'E',
		XmNaccelerator,          "Ctrl<Key>X",
		XmNacceleratorText,      xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNactivateCallback,
		FileCallback, 0 );

/* Monitor Menu */
	xm_string = XmStringCreateSimple( "Monitor" );
	XtVaCreateManagedWidget( "Monitor",
		xmCascadeButtonWidgetClass, MenuBar,
		XmNlabelString,             xm_string,
		XmNmnemonic,                'M',
		XmNsubMenuId,               monitor_menu,
		NULL );
	XmStringFree( xm_string );

/* Monitor Calls */
	type = MON_CALLS;
	xm_string  = XmStringCreateSimple( "Ctrl-C" );
	widget = XtVaCreateManagedWidget( "Calls",
		xmToggleButtonWidgetClass, monitor_menu,
		XmNindicatorType,          XmONE_OF_MANY,
		XmNmnemonic,               'C',
		XmNaccelerator,            "Ctrl<Key>C",
		XmNacceleratorText,        xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNvalueChangedCallback,
		MonitorCallback, type );
	if( mon_type == type )
	{
		XtVaSetValues( widget, XmNset, True, NULL );
		CurrentMonitorWidget = widget;
	}

/* Monitor States */
	type = MON_STATES;
	xm_string  = XmStringCreateSimple( "Ctrl-S" );
	widget = XtVaCreateManagedWidget( "States",
		xmToggleButtonWidgetClass, monitor_menu,
		XmNindicatorType,          XmONE_OF_MANY,
		XmNmnemonic,               'S',
		XmNaccelerator,            "Ctrl<Key>S",
		XmNacceleratorText,        xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNvalueChangedCallback,
		MonitorCallback, type );
	if( mon_type == type )
	{
		XtVaSetValues( widget, XmNset, True, NULL );
		CurrentMonitorWidget = widget;
	}

/* Monitor User Status */
	type = MON_USER_STATUS;
	xm_string  = XmStringCreateSimple( "Ctrl-U" );
	widget = XtVaCreateManagedWidget( "User Status",
		xmToggleButtonWidgetClass, monitor_menu,
		XmNindicatorType,          XmONE_OF_MANY,
		XmNmnemonic,               'U',
		XmNaccelerator,            "Ctrl<Key>U",
		XmNacceleratorText,        xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNvalueChangedCallback,
		MonitorCallback, type );
	if( mon_type == type )
	{
		XtVaSetValues( widget, XmNset, True, NULL );
		CurrentMonitorWidget = widget;
	}

/* Monitor Queues */
	type = MON_QUEUES;
	xm_string  = XmStringCreateSimple( "Ctrl-Q" );
	widget = XtVaCreateManagedWidget( "Queues",
		xmToggleButtonWidgetClass, monitor_menu,
		XmNindicatorType,          XmONE_OF_MANY,
		XmNmnemonic,               'Q',
		XmNaccelerator,            "Ctrl<Key>Q",
		XmNacceleratorText,        xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNvalueChangedCallback,
		MonitorCallback, type );
	if( mon_type == type )
	{
		XtVaSetValues( widget, XmNset, True, NULL );
		CurrentMonitorWidget = widget;
	}

/* Monitor Transfer Rates */
	type = MON_TRATES;
	xm_string  = XmStringCreateSimple( "Ctrl-T" );
	widget = XtVaCreateManagedWidget( "Transfer Rates",
		xmToggleButtonWidgetClass, monitor_menu,
		XmNindicatorType,          XmONE_OF_MANY,
		XmNmnemonic,               'T',
		XmNaccelerator,            "Ctrl<Key>T",
		XmNacceleratorText,        xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNvalueChangedCallback,
		MonitorCallback, type );
	if( mon_type == type )
	{
		XtVaSetValues( widget, XmNset, True, NULL );
		CurrentMonitorWidget = widget;
	}

/* Monitor Delta Calls */
	type = MON_DELTA_CALLS;
	xm_string  = XmStringCreateSimple( "Ctrl-D" );
	widget = XtVaCreateManagedWidget( "Delta Calls",
		xmToggleButtonWidgetClass, monitor_menu,
		XmNindicatorType,          XmONE_OF_MANY,
		XmNmnemonic,               'D',
		XmNaccelerator,            "Ctrl<Key>D",
		XmNacceleratorText,        xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNvalueChangedCallback,
		MonitorCallback, type );
	if( mon_type == type )
	{
		XtVaSetValues( widget, XmNset, True, NULL );
		CurrentMonitorWidget = widget;
	}

/* Monitor Activity */
	type = MON_DELTA_GRAPH;
	xm_string  = XmStringCreateSimple( "Ctrl-A" );
	widget = XtVaCreateManagedWidget( "Activity Graph",
		xmToggleButtonWidgetClass, monitor_menu,
		XmNindicatorType,          XmONE_OF_MANY,
		XmNmnemonic,               'A',
		XmNaccelerator,            "Ctrl<Key>A",
		XmNacceleratorText,        xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNvalueChangedCallback,
		MonitorCallback, type );
	if( mon_type == type )
	{
		XtVaSetValues( widget, XmNset, True, NULL );
		CurrentMonitorWidget = widget;
	}

/* Options Menu */
	xm_string = XmStringCreateSimple( "Options" );
	XtVaCreateManagedWidget( "Options",
		xmCascadeButtonWidgetClass, MenuBar,
		XmNlabelString,             xm_string,
		XmNmnemonic,                'O',
		XmNsubMenuId,               options_menu,
		NULL );
	XmStringFree( xm_string );

/* Polling Interval */
	xm_string  = XmStringCreateSimple( "Ctrl-P" );
	widget = XtVaCreateManagedWidget( "Polling Interval",
		xmPushButtonWidgetClass, options_menu,
		XmNmnemonic,             'P',
		XmNaccelerator,          "Ctrl<Key>P",
		XmNacceleratorText,      xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget, XmNactivateCallback,
		OptionsCallback, NULL );

/* Bring all to bear */		
	XtManageChild( MenuBar );
}

XcpsmonCommandLine( hostname, id, mon_type, mon_proc, mon_class, rpcvers )
char *hostname;
int  *id;
int  *mon_type;
int  *mon_proc;
int  *mon_class;
char *rpcvers;
{
	if( !strcmp( app_data.monitor, "calls" ) )
		*mon_type = MON_CALLS;
	else if( !strcmp( app_data.monitor, "states" ) )
		*mon_type = MON_STATES;
	else if( !strcmp( app_data.monitor, "rates" ) )
		*mon_type = MON_TRATES;
	else if( !strcmp( app_data.monitor, "queues" ) )
		*mon_type = MON_QUEUES;
	else if( !strcmp( app_data.monitor, "user" ) )
		*mon_type = MON_USER_STATUS;
	else if( !strcmp( app_data.monitor, "graph" ) )
		*mon_type = MON_DELTA_GRAPH;
	else if( !strcmp( app_data.monitor, "deltas" ) )
		*mon_type = MON_DELTA_CALLS;

	*id        = app_data.id;
	*mon_proc  = app_data.proc;
	*mon_class = app_data.class;

	strcpy( hostname, app_data.hostname );

	if( !strcmp( app_data.hostname, "localhost" ) )
		*rpcvers = False;
	else
		*rpcvers = True;

	if( app_data.interval < 1 )
		app_data.interval = 1;
	Timeout = app_data.interval;
	OldTimeout = app_data.interval;
}

XcpsmonInit( argc, argv )
int *argc;
char **argv;
{
	char *monstr;

	TheTop = XtVaAppInitialize( &app, "XCps_mon",
		options, XtNumber( options ),
		argc, argv, fallbacks, NULL );

	XtGetApplicationResources( TheTop,
		&app_data, resources, XtNumber( resources ), NULL, 0 );

	XtAddEventHandler( TheTop,
		StructureNotifyMask, False,
		TheTopResizeCallback, NULL );
}

XcpsmonWindow()
{
	Widget widget;

	MainWindow = XtVaCreateManagedWidget( "MainWindow",
		xmFormWidgetClass, TheTop,
		NULL );

	XcpsmonMenus();

	XcpsmonFooter("Footer");

	MainScroll = XtVaCreateManagedWidget( "MainScroll",
		xmScrolledWindowWidgetClass, MainWindow,
		XmNtopAttachment,            XmATTACH_WIDGET,
		XmNtopWidget,                MenuBar,
		XmNtopOffset,                10,
		XmNleftAttachment,           XmATTACH_FORM,
		XmNleftOffset,               10,
		XmNrightAttachment,          XmATTACH_FORM,
		XmNrightOffset,              10,
		XmNbottomAttachment,         XmATTACH_WIDGET,
		XmNbottomOffset,             5,
		XmNbottomWidget,             Footer,
		XmNscrollingPolicy,          XmAUTOMATIC,
		NULL );

	XtVaGetValues( MainScroll, XmNverticalScrollBar, &widget, NULL );
	XtUnmanageChild( widget );

	MainForm = XtVaCreateManagedWidget( "MainForm",
		xmFormWidgetClass, MainScroll,
		NULL );

	XcpsmonHeader();

	XcpsmonGraphLegend();

	display = XtDisplay( MainWindow );
	screen = XtScreen( MainWindow );
	gc = XCreateGC( display, RootWindowOfScreen( screen ), NULL, NULL );

	XtVaGetValues( MainScroll, XmNbackground, &BackgroundColor, NULL );
	XtVaGetValues( MainScroll, XmNforeground, &ForegroundColor, NULL );

	WorkScroll = XtVaCreateManagedWidget( "WorkScroll",
		xmScrolledWindowWidgetClass, MainForm,
		XmNtopAttachment,            XmATTACH_WIDGET,
		XmNtopWidget,                Header,
		XmNtopOffset,                0,
		XmNleftAttachment,           XmATTACH_FORM,
		XmNleftOffset,               10,
		XmNrightAttachment,          XmATTACH_FORM,
		XmNrightOffset,              10,
		XmNbottomAttachment,         XmATTACH_WIDGET,
		XmNbottomOffset,             10,
		XmNbottomWidget,             Legend,
		XmNscrollBarPlacement,       XmTOP_LEFT,
		XmNscrollBarDisplayPolicy,   XmSTATIC,
		XmNscrollingPolicy,          XmAUTOMATIC,
		NULL );

	XtVaGetValues( WorkScroll, XmNhorizontalScrollBar, &widget, NULL );
	XtUnmanageChild( widget );
}

XcpsmonEventLoop()
{
	XtRealizeWidget( TheTop );
	XtAppMainLoop( app );
}

XcpsmonTitle( string )
char *string;
{
	char title[128];

	sprintf( title, "%s - %s", TITLE, string );
	XtVaSetValues( TheTop,
		XtNtitle,    title,
		XtNiconName, ICON,
		NULL );
}

XcpsmonHeader()
{
	Header = XtVaCreateManagedWidget( "Header",
		xmBulletinBoardWidgetClass,  MainForm,
		XmNtopAttachment,   XmATTACH_FORM,
		XmNtopOffset,       0,
		XmNleftAttachment,  XmATTACH_FORM,
		XmNleftOffset,      10,
		XmNrightAttachment, XmATTACH_FORM,
		XmNrightOffset,     10,
		XmNmarginHeight,    5,
		NULL );
}

XcpsmonFooter( string )
char *string;
{
	XmString xm_string = XmStringCreateSimple( string );
	if( !Footer )
		Footer = XtVaCreateManagedWidget( "Footer",
			xmLabelWidgetClass,     MainWindow,
			XmNlabelString,         xm_string,
			XmNleftAttachment,      XmATTACH_FORM,
			XmNleftOffset,          1,
			XmNrightAttachment,     XmATTACH_FORM,
			XmNrightOffset,         1,
			XmNbottomAttachment,    XmATTACH_FORM,
			XmNbottomOffset,        1,
			NULL );
	else
		XtVaSetValues( Footer, XmNlabelString, xm_string, NULL );
	XmStringFree( xm_string );
}

Dimension HeaderWidth( string )
char *string;
{
	Dimension width;
	XmFontList fontlist;
	XmString xm_string = XmStringCreateLtoR( string, XmSTRING_DEFAULT_CHARSET );
	Widget label = XtVaCreateWidget( "label",
		xmLabelWidgetClass, Header,
		XmNlabelString,     xm_string,
		NULL );
	XtVaGetValues( label, XmNfontList, &fontlist, NULL );

	width = XmStringWidth( fontlist, xm_string ) + 10;

	XmStringFree( xm_string );
	XtDestroyWidget( label );

	return( width );
}

XcpsmonWorkForm( columns, type )
int columns;
int type;
{
	int index, column;
	Dimension height;

	for( index = 0; index < NumLegendWidgets; index++ )
	{
		/*
		XtDestroyWidget( LegendWidget[ index ] );
		*/
		XtDestroyWidget( LegendStringWidget[ index ] );
	}

	for( index = 0; index < NumColumnHeaders; index++ )
		XtDestroyWidget( ColumnHeader[ index ] );

	for( index = 0; index < NumProcs; index++ )
	{
		for( column = 0; column < NumColumnHeaders - 1; column++ )
			XtDestroyWidget( DataWidget[ column ][ index ] );
		XtDestroyWidget( LabelWidget[ index ] );
	}

	NumProcs = 0;
	NumLegendWidgets = 0;
	NumColumnHeaders = 0;
	MaxHeaderWidth = 0;

	if( WorkForm )
		XtDestroyWidget( WorkForm );

	WorkForm = XtVaCreateManagedWidget( "WorkForm",
		xmBulletinBoardWidgetClass, WorkScroll,
		NULL );

	XtAddEventHandler( WorkForm,
		StructureNotifyMask, False, WorkFormResizeCallback, NULL );
}

XcpsmonColumnHeader( string, align )
char *string;
int align;
{
	XmString xm_string;

	if( align == 0 )
		align = XmALIGNMENT_BEGINNING;
	else if( align == 1 )
		align = XmALIGNMENT_CENTER;
	else if( align == 2 )
		align = XmALIGNMENT_END;

	if( string == NULL )
	{
		int index;
		for( index = 0; index < NumColumnHeaders; index++ )
			XtVaSetValues( ColumnHeader[ index ],
				XmNx,     index * MaxHeaderWidth + 35,
				XmNwidth, MaxHeaderWidth,
				NULL );
		return;
	}

	xm_string = XmStringCreateLtoR( string, XmSTRING_DEFAULT_CHARSET );
	ColumnHeader[ NumColumnHeaders ] = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass, Header,
		XmNlabelString,     xm_string,
		XmNalignment,       align,
		NULL );
	XmStringFree( xm_string );

	if( HeaderWidth( string ) > MaxHeaderWidth )
		MaxHeaderWidth = HeaderWidth( string );

	NumColumnHeaders++;
}

XcpsmonGraphColumnHeader( string )
char *string;
{
	XmString xm_string;

	if( string == NULL )
	{
		int index;
		for( index = 0; index < NumColumnHeaders; index++ )
			XtVaSetValues( ColumnHeader[ index ],
				XmNx,     index * MaxHeaderWidth + 35,
				XmNwidth, MaxHeaderWidth,
				XmNalignment, XmALIGNMENT_BEGINNING,
				NULL );
		return;
	}

	xm_string = XmStringCreateLtoR( string, XmSTRING_DEFAULT_CHARSET );
	ColumnHeader[ NumColumnHeaders ] = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass, Header,
		XmNlabelString,     xm_string,
		NULL );
	XmStringFree( xm_string );

	if( HeaderWidth( string ) > MaxHeaderWidth )
		MaxHeaderWidth = HeaderWidth( string );

	NumColumnHeaders++;
}

XcpsmonGraphLegend()
{
	Legend = XtVaCreateManagedWidget( "Legend",
		xmFormWidgetClass,            MainForm,
		XmNleftAttachment,            XmATTACH_FORM,
		XmNleftOffset,                0,
		XmNrightAttachment,           XmATTACH_FORM,
		XmNrightOffset,               0,
		XmNbottomAttachment,          XmATTACH_FORM,
		XmNbottomOffset,              0,
		XmNheight,                    GraphWidgetHeight,
		NULL );
}

XcpsmonGraphLegendElement( type )
int type;
{
	Window win;
	XmString xm_string;

	/*
	LegendWidget[ NumLegendWidgets ] = XtVaCreateManagedWidget( "box",
		xmDrawingAreaWidgetClass, Legend,
		XmNwidth,                 GraphWidgetHeight,
		XmNheight,                GraphWidgetHeight,
		XmNbottomAttachment,      XmATTACH_FORM,
		XmNbottomOffset,          0,
		XmNleftAttachment,        NumLegendWidgets?
			XmATTACH_WIDGET: XmATTACH_FORM,
		XmNleftWidget,            NumLegendWidgets?
			LegendStringWidget[ NumLegendWidgets - 1 ]: NULL,
		XmNleftOffset,            10,
		NULL );

	XtAddCallback( LegendWidget[ NumLegendWidgets ],
		XmNexposeCallback, FillLegendWidget, NULL );
	*/
	LegendWidgetColor[ NumLegendWidgets ] = type;

	xm_string = XmStringCreateSimple( names[ type ] );
	LegendStringWidget[ NumLegendWidgets ] = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass,       Legend,
		XmNlabelString,           xm_string,
		/*
		XmNleftAttachment,        XmATTACH_WIDGET,
		XmNleftWidget,            LegendWidget[ NumLegendWidgets ],
		*/
		XmNleftOffset,            5,
		XmNtopAttachment,         XmATTACH_OPPOSITE_WIDGET,
		XmNtopWidget,             LegendWidget[ NumLegendWidgets ],

		XmNleftAttachment,        NumLegendWidgets?
			XmATTACH_WIDGET: XmATTACH_FORM,
		XmNleftWidget,            NumLegendWidgets?
			LegendStringWidget[ NumLegendWidgets - 1 ]: NULL,
		XmNforeground,
			(Pixel) SetColor( display, DefaultScreen( display ),
				colors[ type ], gc ),

		NULL );
	XmStringFree( xm_string );

	NumLegendWidgets++;
}

XcpsmonProcessLine( string, align )
char *string;
int align;
{
	XmString xm_string;
	int column;
	Dimension width, height;
	Position pos, pos2;
	char var_size;

	if( align == 0 )
	{
		align = XmALIGNMENT_BEGINNING;
		var_size = True;
	}
	else if( align == 1 )
	{
		align = XmALIGNMENT_CENTER;
		var_size = False;
	}
	else if( align == 2 )
	{
		align = XmALIGNMENT_END;
		var_size = True;
	}

		XtVaGetValues( ColumnHeader[ 0 ], XmNx, &pos, NULL );
		XtVaGetValues( ColumnHeader[ 0 ], XmNheight, &height, NULL );
		pos2 = pos - 30;

	xm_string = XmStringCreateSimple( string );
	LabelWidget[ NumProcs ] = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass, WorkForm,
		XmNlabelString,     xm_string,
		XmNx,               pos2,
		XmNy,               NumProcs * height + 10,
		XmNwidth,           MaxHeaderWidth,
		XmNalignment,       align,
		XmNrecomputeSize,   var_size,
		NULL );
	XmStringFree( xm_string );

	for( column = 0; column < NumColumnHeaders - 1; column++ )
	{
		XtVaGetValues( ColumnHeader[ column + 1 ], XmNx, &pos, NULL );
		pos2 = pos - 25;

		xm_string = XmStringCreateSimple( " " );
		DataWidget[ column ][ NumProcs ] = XtVaCreateManagedWidget( "label",
			xmLabelWidgetClass, WorkForm,
			XmNlabelString,     xm_string,
			XmNx,               pos2,
			XmNy,               NumProcs * height + 10,
			XmNwidth,           MaxHeaderWidth,
			XmNalignment,       align,
			XmNrecomputeSize,   var_size,
			NULL );
		XmStringFree( xm_string );
	}

	NumProcs++;
}

XcpsmonProcessGraph( string )
char *string;
{
	XmString xm_string;
	Widget DrawingFrame;
	Position position;
	Dimension width;

	XtVaGetValues( ColumnHeader[ 0 ], XmNx, &position, NULL );

	XtVaGetValues( ColumnHeader[ 0 ], XmNheight, &GraphWidgetHeight, NULL );

	xm_string = XmStringCreateSimple( string );
	LabelWidget[ NumProcs ] = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass, WorkForm,
		XmNlabelString,     xm_string,
		XmNx,               position - 30,
		XmNy,               NumProcs * ( GraphWidgetHeight + 10 ) + 10,
		XmNalignment,       XmALIGNMENT_BEGINNING,
		NULL );
	XmStringFree( xm_string );

	XtVaGetValues( ColumnHeader[ 1 ], XmNx, &position, NULL );

	XtVaGetValues( MainForm, XmNwidth, &width, NULL );
	GraphWidgetWidth = width - position - 40;

	DrawingFrame = XtVaCreateManagedWidget( "DrawingFrame",
		xmFrameWidgetClass, WorkForm,
		XmNshadowType,      XmSHADOW_ETCHED_IN,
		XmNx,               position - 25,
		XmNy,               NumProcs * ( GraphWidgetHeight + 10 ) + 10,
		NULL );

	DataWidget[ 0 ][ NumProcs ] = XtVaCreateManagedWidget( "graph",
		xmDrawingAreaWidgetClass, DrawingFrame,
		XmNwidth,                 GraphWidgetWidth,
		XmNheight,                GraphWidgetHeight,
		NULL );

	XtAddCallback( DataWidget[ 0 ][ NumProcs ],
		XmNexposeCallback, FillGraphWidget, NULL );

	if( !DataPixmap[ NumProcs ] )
		DataPixmap[ NumProcs ] = XCreatePixmap( display,
			RootWindowOfScreen( screen ),
			1152, GraphWidgetHeight,
			DefaultDepthOfScreen( screen ) );

	XSetForeground( display, gc, BackgroundColor );
	XFillRectangle( display, DataPixmap[ NumProcs ], gc, 0, 0,
		1152, GraphWidgetHeight );

	NumProcs++;
}

XcpsmonUpdateProcessLine( column, row, string )
int column, row;
char *string;
{
	XmString xm_string = XmStringCreateSimple( string );
	XtVaSetValues( DataWidget[ column ][ row ],
		XmNlabelString, xm_string,
		NULL );
	XmStringFree( xm_string );
}

XcpsmonUpdateDeadProcess( column, row )
int column, row;
{
	SetColor( display, DefaultScreen( display ), "Black", gc );
	XFillRectangle( display, DataPixmap[ row ], gc, 0, 0,
		1152, GraphWidgetHeight );
}

XcpsmonUpdateProcessGraph( column, row, width, type, offset )
int column, row;
int width;
int type;
int offset;
{
	Window win = XtWindow( DataWidget[ column ][ row ] );
	int index;
	static float ratio;

/* Clear pixmap */
	if( type == -1 )
	{
		/*
		ratio = ( 10.0 / Timeout ) * ( GraphWidgetHeight / 20.0 );
		*/
		ratio = ( 20.0 / Timeout ) * ( GraphWidgetHeight / 20.0 );
		XSetForeground( display, gc, BackgroundColor );
		XFillRectangle( display, DataPixmap[ row ], gc, 0, 0,
			1152, GraphWidgetHeight );
	}
/* Copy pixmap to screen */
	else if( type == 999 )
	{
		XSetForeground( display, gc, ForegroundColor );
		for( index = 20; index < GraphWidgetWidth / ratio; index += 20 )
		{
			XDrawLine( display, DataPixmap[ row ], gc,
				(int)( ratio * index ), GraphWidgetHeight - 3,
				(int)( ratio * index ), GraphWidgetHeight );
		}

		XCopyArea( display, DataPixmap[ row ], win, gc, 0, 0,
			GraphWidgetWidth, GraphWidgetHeight, 0, 0 );
	}
/* Build pixmap from multi-colored graphs */
	else
	{
		SetColor( display, DefaultScreen( display ), colors[ type ], gc );
		XFillRectangle( display, DataPixmap[ row ], gc,
			(int)( ratio * offset ), 0,
			(int)( ratio * width ), GraphWidgetHeight );
	}
}

XcpsmonUpdateFooter( string )
char *string;
{
	XmString xm_string = XmStringCreateSimple( string );
	XtVaSetValues( Footer,
		XmNlabelString, xm_string,
		NULL );
	XmStringFree( xm_string );
	ForceUpdate( Footer );
}

XcpsmonForceUpdates()
{
	ForceUpdate( WorkForm );
}

XcpsmonTimeout()
{
	XtAppAddTimeOut( app, Timeout * 1000, handle_alarm, NULL );
}

PollDialogCallback( w, data )
Widget w;
int data;
{
	char *string;
	switch( data )
	{
		case 0:
			string = XmTextFieldGetString( PollTextField );
			Timeout = atoi( string );
			if( Timeout < 1 )
				Timeout = 1;
			OldTimeout = Timeout;
			XtFree( string );
			XtUnmanageChild( PollDialog );
			break;
		case 1:
			Timeout = OldTimeout;
			XtUnmanageChild( PollDialog );
			break;
		default:
			break;
	}
}

PollDialogMask( widget, data, cbs )
Widget widget;
XtPointer data;
XmTextVerifyCallbackStruct *cbs;
{
	int len;

	if( cbs->startPos < cbs->currInsert ) /* backspace */
		return;

	for( len = 0; len < cbs->text->length; len++ )
	{
		if( !isdigit( cbs->text->ptr[ len ] ) )
		{
			cbs->doit = False;
			return;
		}
	}
}

XcpsmonPollDialog( w )
Widget w;
{
	Widget dialog, action_area, control_area, separator;
	Widget label, text_field;
	XmString xm_string;
	char string[16];
	static ActionAreaItem action_items[] =
	{
		{ "OK",     (XtCallbackProc)PollDialogCallback, NULL },
		{ "Cancel", (XtCallbackProc)PollDialogCallback, NULL }
	};

	if( PollDialog )
	{
		sprintf( string, "%d", Timeout );
		XmTextFieldSetString( PollTextField, string );
		XtManageChild( PollDialog );
		return;
	}
	dialog = XmCreateFormDialog( w, "dialog", NULL, 0 );
	PollDialog = dialog;

	action_items[ 0 ].data = (caddr_t)0;
	action_items[ 1 ].data = (caddr_t)1;
	
	action_area = CreateActionArea( dialog, action_items,
		XtNumber( action_items ), 0 );

	separator = CreateActionSeparator( dialog, action_area );

	control_area = CreateControlArea( dialog, separator );

	xm_string = XmStringCreateSimple( "Polling Interval:" );
	label = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass,     control_area,
		XmNlabelString,         xm_string,
		XmNtopAttachment,       XmATTACH_FORM,
		XmNleftAttachment,      XmATTACH_FORM,
		NULL );
	XmStringFree( xm_string );

	text_field = XtVaCreateManagedWidget( "text-field",
		xmTextFieldWidgetClass, control_area,
		XmNtopAttachment,       XmATTACH_WIDGET,
		XmNtopWidget,           label,
		XmNleftAttachment,      XmATTACH_FORM,
		XmNwidth,               50,
		NULL );
	sprintf( string, "%d", Timeout );
	XmTextFieldSetString( text_field, string );
	XtAddCallback( text_field, XmNactivateCallback,
		PollDialogCallback, NULL );
	XtAddCallback( text_field, XmNmodifyVerifyCallback,
		 PollDialogMask, NULL );
	PollTextField = text_field;

	xm_string = XmStringCreateSimple( "seconds" );
	label = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass,     control_area,
		XmNlabelString,         xm_string,
		XmNbottomAttachment,    XmATTACH_OPPOSITE_WIDGET,
		XmNbottomWidget,        text_field,
		XmNbottomOffset,        5,
		XmNleftAttachment,      XmATTACH_WIDGET,
		XmNleftWidget,          text_field,
		XmNleftOffset,          5,
		NULL );
	XmStringFree( xm_string );

	XtManageChild( dialog );
}
