/*

 Author:	Roberto Ullfig
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL:ULLFIG
		Internet: ullfig@fnrobo.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.	    
	   All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or	     
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
/*
	Include Files
*/
#include "all.h"
#include "../inc/types.h"
#include "../inc/time.h"
#include <stdio.h>
#include <sys/socket.h>
#include <rpc/rpc.h>
#include "jm_rpc.h"
#include "../inc/version.h"

short message_ready = FALSE;
rpc_message *jm_message = NULL;

/******************************************************************/
/*
	REMOTE PROCEDURE: 1

		rpc_get_user_status_1 (cps_mon)

	DESCRIPTION:

		return a value from the user_status array

	FORMAL PARAMETERS:

		x and y index into array
*/
/******************************************************************/
int *rpc_get_user_status_1( pair )
struct int_pair *pair;
{
	return( &user_status[ pair->x ][ pair->y ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 2

		rpc_get_call_cnt_1 (cps_mon)

	DESCRIPTION:

		return a value from the call_cnt array in shared memory

	FORMAL PARAMETERS:

		x and y index into array
*/
/******************************************************************/
int *rpc_get_call_count_1( pair )
struct int_pair *pair;
{
	return( &mon_shm->call_cnt[ pair->x ][ pair->y ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 3

		rpc_get_process_state_1 (cps_mon & jmdb)

	DESCRIPTION:

		return a value from the p_state array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_process_state_1( process )
int *process;
{
	return( &mon_shm->p_state[ *process ].state );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 4

		rpc_get_process_attributes_1 (cps_mon & jmdb)

	DESCRIPTION:

		return a value from the p_state array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_process_attributes_1( process )
int *process;
{
	return( &mon_shm->p_state[ *process ].attrib );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 5

		rpc_get_nsends_1 (cps_mon)

	DESCRIPTION:

		return a value from the nsends array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_nsends_1( process )
int *process;
{
	return( &mon_shm->nsends[ *process ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 6

		rpc_get_sndcur_1 (cps_mon)

	DESCRIPTION:

		return a value from the sndcur array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_sndcur_1( process )
int *process;
{
	return( &mon_shm->sndcur[ *process ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 7

		rpc_get_sndavg_1 (cps_mon)

	DESCRIPTION:

		return a value from the snd_avg array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_sndavg_1( process )
int *process;
{
	return( &mon_shm->sndavg[ *process ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 8

		rpc_get_ngets_1 (cps_mon)

	DESCRIPTION:

		return a value from the ngets array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_ngets_1( process )
int *process;
{
	return( &mon_shm->ngets[ *process ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 9

		rpc_get_getcur_1 (cps_mon)

	DESCRIPTION:

		return a value from the getcur array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_getcur_1( process )
int *process;
{
	return( &mon_shm->getcur[ *process ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 10

		rpc_get_getavg_1 (cps_mon)

	DESCRIPTION:

		return a value from the getavg array in shared memory

	FORMAL PARAMETERS:

		index into array
*/
/******************************************************************/
int *rpc_get_getavg_1( process )
int *process;
{
	return( &mon_shm->getavg[ *process ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 11

		rpc_get_q_declared_1 (cps_mon & jmdb)

	DESCRIPTION:

		returns 1 if queue is declared
		returns 0 otherwise

	FORMAL PARAMETERS:

		queue number
*/
/******************************************************************/
int *rpc_get_q_declared_1( queue )
int *queue;
{
	return( &mon_shm->q_declared[ *queue ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 12

		rpc_get_q_set_1 (cps_mon & jmdb)

	DESCRIPTION:

		returns set of processes on queue

	FORMAL PARAMETERS:

		queue number
*/
/******************************************************************/
int *rpc_get_q_set_1( queue )
int *queue;
{
	return( &mon_shm->qset[ *queue ] );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 13

		rpc_get_job_done_1 (cps_mon)

	DESCRIPTION:

		returns value of job_done from shared memory

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
int *rpc_get_job_done_1()
{
	return( &mon_shm->job_done );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 14

		rpc_get_num_proc_1 (cps_mon & jmdb)

	DESCRIPTION:

		returns number of process in job from shared memory

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
int *rpc_get_num_proc_1()
{
	return( &mon_shm->num_proc );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 15

		rpc_get_process_class_1 (cps_mon & jmdb)

	DESCRIPTION:

		returns class that process is is

	FORMAL PARAMETERS:

		process number
*/
/******************************************************************/
int *rpc_get_process_class_1( process )
int *process;
{
	return( &usr_map[ *process ].class );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 16

		rpc_get_ndq_set_1 (jmdb)

	DESCRIPTION:

		returns number of pending dequeues

	FORMAL PARAMETERS:

		queue number
*/
/******************************************************************/
int *rpc_get_ndq_set_1( queue )
int *queue;
{
	int ndq;
	ndq = set_size( jmq[*queue].dq_set, MAX_PROCESSES );
	return( &ndq );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 20

		rpc_kill_jm_1 (jmdb)

	DESCRIPTION:

		aborts a CPS job

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char *rpc_kill_jm_1()
{
	static char reply = 1;

	abort_job(8);

	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 21

		rpc_step_jm_1 (jmdb)

	DESCRIPTION:

		allows one message to be processed by JM before stopping it

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char *rpc_step_jm_1()
{
	static char reply;

	if( jm_pause )
	{
		if( sync_point != -1 )
			sync_point = -1;
		else if( queue_point != -1 )
			queue_point = -1;
		else if( user_point != -1 )
			user_point = -1;
		else if( proc_point != -1 )
			proc_point = -1;
		jm_pause = FALSE;
		jm_step  = TRUE;
		if( jm_init )
			jm_init = FALSE;
		reply = 1;
	}
	else
	{
		reply = 0;
	}
	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 22

		rpc_stop_jm_1 (jmdb)

	DESCRIPTION:

		stop the JM

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char *rpc_stop_jm_1()
{
	static char reply;


	if( !jm_pause )
	{
		jm_pause = TRUE;
		reply = 1;
	}
	else
	{
		reply = 0;
	}
	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 23

		rpc_continue_jm_1 (jmdb)

	DESCRIPTION:

		starts the JM

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char *rpc_continue_jm_1()
{
	static char reply;

	if( jm_pause )
	{
		if( sync_point != -1 )
			sync_point = -1;
		else if( queue_point != -1 )
			queue_point = -1;
		else if( user_point != -1 )
			user_point = -1;
		else if( proc_point != -1 )
			proc_point = -1;
		jm_pause = FALSE;
		if( jm_init )
			jm_init = FALSE;
		reply = 1;
	}
	else
	{
		reply = 0;
	}
	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 30

		rpc_get_jm_status_1 (jmdb)

	DESCRIPTION:

		returns the status of the JM (stopped or running)
		used to test RPC connection

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char *rpc_get_jm_status_1()
{
	return( &jm_pause );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 31

		rpc_get_jm_pid_1 (jmdb)

	DESCRIPTION:

		get PID of JM

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
int *rpc_get_jm_pid_1()
{
	static int reply;
	reply = getpid();
	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 32

		rpc_get_jm_directory_1 (jmdb)

	DESCRIPTION:

		return JM working directory

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char **rpc_get_jm_directory_1()
{
	static char *directory;
	directory = (char *)getcwd((char *)NULL, 64);
	return( &directory );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 33

		rpc_get_jm_version_1 (jmdb)

	DESCRIPTION:

		return version of JM

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char **rpc_get_jm_version_1()
{
	return( &cps_version );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 34

		rpc_get_jm_username_1 (jmdb)

	DESCRIPTION:

		return owner of JM

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char **rpc_get_jm_username_1()
{
	static char *logname;
	logname = cuserid( NULL );
	return( &logname );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 35

		rpc_get_jm_message_1 (jmdb)

	DESCRIPTION:

		return last message received by JM

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
rpc_message *rpc_get_jm_message_1()
{
	if( jm_message == NULL )
		jm_message = (rpc_message *)hmalloc( sizeof( cps_message ) );

	if( message_ready )
	{
		message_ready = FALSE;
	}
	else
	{
		jm_message->type = -1;
	}
	return( (rpc_message *)jm_message );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 36

		rpc_get_job_name_1 (jmdb)

	DESCRIPTION:

		return name of job from JDF file

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char **rpc_get_job_name_1()
{
	return( &job_name );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 37

		rpc_get_starttime_1 (jmdb)

	DESCRIPTION:

		return time job started

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
long *rpc_get_jm_starttime_1()
{
	static long reply;
	reply = (long)jm_starttime;
	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 40

		rpc_get_process_id_1 (jmdb)

	DESCRIPTION:

		get PID of user-process

	FORMAL PARAMETERS:

		process number
*/
/******************************************************************/
int *rpc_get_process_id_1( process )
int *process;
{
	return( &usr_map[ *process ].pid );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 41

		rpc_get_process_host_1 (jmdb)

	DESCRIPTION:

		return host that process is running on

	FORMAL PARAMETERS:

		process number
*/
/******************************************************************/
char **rpc_get_process_host_1( process )
int *process;
{
	return( &sdf_cpu[ usr_map[ *process ].cpu ].cpu_name );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 42

		rpc_get_process_name_1 (jmdb)

	DESCRIPTION:

		return name of process

	FORMAL PARAMETERS:

		process number
*/
/******************************************************************/
char **rpc_get_process_name_1( process )
int *process;
{
	return( &class[ usr_map[ *process ].class ].program );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 43

		rpc_get_process_dir_1 (jmdb)

	DESCRIPTION:

		return process working directory

	FORMAL PARAMETERS:

		process number
*/
/******************************************************************/
char **rpc_get_process_dir_1( process )
int *process;
{
	return( &usr_map[ *process ].directory );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 44

		rpc_get_log_file_1 (jmdb)

	DESCRIPTION:

		return name of JML

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
char **rpc_get_log_file_1()
{
	return( &log_file );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 50

		rpc_upshot_switch_1 (jmdb)

	DESCRIPTION:

		turns upshot logging on/off

	FORMAL PARAMETERS:

		flag specifying on or off
*/
/******************************************************************/
char *rpc_upshot_switch_1( flag )
int *flag;
{
	static char status;

	if( *flag == 0 )
		if( upshot )
		{
			upshot_stop();
			status = 0;
		}
		else
		{
			status = 1;
		}
	else if( *flag == 1 )
		if( upshot )
		{
			status = 1;
		}
		else
		{
			upshot_start();
			status = 0;
		}
	else if( *flag == 2 )
		status = upshot;

	return( &status );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 51

		rpc_hook_switch_1 (jmdb)

	DESCRIPTION:

		turns hooks on/off

	FORMAL PARAMETERS:

		flag specifying on or off
*/
/******************************************************************/
char *rpc_hook_switch_1( flag )
int *flag;
{
	static char status;
	cps_message m;

	if( *flag == 0 )
		if( jm_hooks )
		{
			jm_hooks = 0;
			cps_bld_hook_switch( &m );
			shm_usr_broadcast( &m );
			status = 0;
			/*
			printf("jm[%s] Hooks turned OFF\n",jm_hostname);
			fflush(stdout);
			*/
			fprintf(jml, "jm[%s] Hooks turned OFF\n",jm_hostname);
			fflush(jml);
		}
		else
		{
			status = 1;
		}
	else if( *flag == 1 )
		if( jm_hooks )
		{
			status = 1;
		}
		else
		{
			jm_hooks = 1;
			cps_bld_hook_switch( &m );
			shm_usr_broadcast( &m );
			status = 0;
			/*
			printf("jm[%s] Hooks turned ON\n",jm_hostname);
			fflush(stdout);
			*/
			fprintf(jml, "jm[%s] Hooks turned ON\n",jm_hostname);
			fflush(jml);
		}
	else if( *flag == 2 )
		status = jm_hooks;

	return( &status );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 60

		rpc_get_process_queue_1 (jmdb)

	DESCRIPTION:

		returns a linked list of processes on queue in order

	FORMAL PARAMETERS:

		queue number
*/
/******************************************************************/
int_array_res *rpc_get_process_queue_1( queue )
int *queue;
{
	static int_array_res result;
	static int firsttime = 1;
	int_array newp, *pp;
	struct cps_queue_element *node;

	if( firsttime )
		firsttime = 0;
	else
		xdr_free( xdr_int_array_res, &result );

	pp = &result.int_array_res_u.plist;

	result.errno = 1;
	for( node = jmq[*queue].queued->front; node != NULL; node = node->next )
	{
		newp = *pp = (element *)malloc( sizeof( element ) );
		newp->lpn = node->lpn;
		pp = &newp->next;
		result.errno = 0;
	}
	*pp = NULL;

	return( &result );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 61

		rpc_get_at_sync_1 (jmdb)

	DESCRIPTION:

		returns linked list of processes at sync point

	FORMAL PARAMETERS:

		sync point number
*/
/******************************************************************/
int_array_res *rpc_get_at_sync_1( sync )
int *sync;
{
	static int_array_res result;
	static int firsttime = 1;
	int_array newp, *pp;
	int lpn;

	if( firsttime )
		firsttime = 0;
	else
		xdr_free( xdr_int_array_res, &result );

	pp = &result.int_array_res_u.plist;

	result.errno = 1;
	if( sync_active[ *sync ] )
	{
		for( lpn = 1; lpn <= MAX_PROCESSES; lpn++ )
		{
			if( set_member( lpn, at_sync[ *sync ] ) )
			{
				newp = *pp = (element *)malloc( sizeof( element ) );
				newp->lpn = lpn;
				pp = &newp->next;
				result.errno = 0;
			}
		}
	}

	*pp = NULL;

	return( &result );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 62

		rpc_get_need_sync_1 (jmdb)

	DESCRIPTION:

		returns linked list of processes required to be at sync point

	FORMAL PARAMETERS:

		sync point number
*/
/******************************************************************/
int_array_res *rpc_get_need_sync_1( sync )
int *sync;
{
	static int_array_res result;
	static int firsttime = 1;
	int_array newp, *pp;
	int lpn;

	if( firsttime )
		firsttime = 0;
	else
		xdr_free( xdr_int_array_res, &result );

	pp = &result.int_array_res_u.plist;

	result.errno = 1;
	if( sync_active[ *sync ] )
	{
		for( lpn = 1; lpn <= MAX_PROCESSES; lpn++ )
		{
			if( set_member( lpn, need_sync[ *sync ] ) )
			{
				newp = *pp = (element *)malloc( sizeof( element ) );
				newp->lpn = lpn;
				pp = &newp->next;
				result.errno = 0;
			}
		}
	}

	*pp = NULL;

	return( &result );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 70

		rpc_stop_at_brkpt_1 (jmdb)

	DESCRIPTION:

		set a breakpoint

	FORMAL PARAMETERS:

		pair[0] - type of breakpoint
		pair[1] - breakpoint number
*/
/******************************************************************/
char *rpc_stop_at_brkpt_1( pair )
struct int_pair *pair;
{
	static char reply = 0;

	switch( pair->x )
	{
		/* sync */
		case 1:
			if( !sync_stop[ pair->y ] )
			{
				sync_stop[ pair->y ] = 1;
				reply = 1;
			}
			break;
		/* queue */
		case 2:
			if( !queue_stop[ pair->y ] )
			{
				queue_stop[ pair->y ] = 1;
				reply = 1;
			}
			break;
		/* user hook */
		case 3:
			if( !user_stop[ pair->y ] )
			{
				user_stop[ pair->y ] = 1;
				reply = 1;
			}
			break;
		/* proc hook */
		case 4:
			if( !proc_stop[ pair->y ] )
			{
				proc_stop[ pair->y ] = 1;
				reply = 1;
			}
			break;
	}
	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 71

		rpc_cont_at_brkpt_1 (jmdb)

	DESCRIPTION:

		unset a breakpoint

	FORMAL PARAMETERS:

		pair[0] - type of breakpoint
		pair[1] - breakpoint number
*/
/******************************************************************/
char *rpc_cont_at_brkpt_1( pair )
struct int_pair *pair;
{
	static char reply = 0;

	switch( pair->x )
	{
		/* sync */
		case 1:
			if( sync_stop[ pair->y ] )
			{
				sync_stop[ pair->y ] = 0;
				reply = 1;
			}
			break;
		/* queue */
		case 2:
			if( queue_stop[ pair->y ] )
			{
				queue_stop[ pair->y ] = 0;
				reply = 1;
			}
			break;
		/* user hook */
		case 3:
			if( user_stop[ pair->y ] )
			{
				user_stop[ pair->y ] = 0;
				reply = 1;
			}
			break;
		/* proc hook */
		case 4:
			if( proc_stop[ pair->y ] )
			{
				proc_stop[ pair->y ] = 0;
				reply = 1;
			}
			break;
	}

	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 72

		rpc_get_location_1 (jmdb)

	DESCRIPTION:

		return location of job if at breakpoint

	FORMAL PARAMETERS:

		none
*/
/******************************************************************/
int_pair *rpc_get_location_1()
{
	static int_pair reply;

	if( sync_point >= 0 )
	{
		reply.x = 1;
		reply.y = sync_point;
	}
	else if( queue_point >= 0 )
	{
		reply.x = 2;
		reply.y = queue_point;
	}
	else if( user_point >= 0 )
	{
		reply.x = 3;
		reply.y = user_point;
	}
	else if( proc_point >= 0 )
	{
		reply.x = 4;
		reply.y = proc_point;
	}
	else if( jm_init == 1 )
	{
		reply.x = 5;
		reply.y = 0;
	}
	else
	{
		reply.x = 0;
		reply.y = 0;
	}

	return( &reply );
}

/******************************************************************/
/*
	REMOTE PROCEDURE: 73

		rpc_get_breakpoints_1 (jmdb)

	DESCRIPTION:

		return linked list of all set breakpoints of certain type

	FORMAL PARAMETERS:

		type of breakpoint
*/
/******************************************************************/
int_array_res *rpc_get_breakpoints_1( type )
int *type;
{
	static int_array_res result;
	static int firsttime = 1;
	int_array newp, *pp;
	int i;

	if( firsttime )
		firsttime = 0;
	else
		xdr_free( xdr_int_array_res, &result );

	pp = &result.int_array_res_u.plist;

	result.errno = 1;
	switch( *type )
	{
		/* sync */
		case 1:
			for( i = 0; i < MAX_SYNCS; i++ )
			{
				newp = *pp = (element *)malloc( sizeof( element ) );
				newp->lpn = sync_stop[ i ];
				pp = &newp->next;
			}
			result.errno = 0;
			break;
		/* queue */
		case 2:
			for( i = 0; i < MAX_QUEUES; i++ )
			{
				newp = *pp = (element *)malloc( sizeof( element ) );
				newp->lpn = queue_stop[ i ];
				pp = &newp->next;
			}
			result.errno = 0;
			break;
		/* user hook */
		case 3:
			for( i = 0; i < MAX_USER_HOOKS; i++ )
			{
				newp = *pp = (element *)malloc( sizeof( element ) );
				newp->lpn = user_stop[ i ];
				pp = &newp->next;
			}
			result.errno = 0;
			break;
		/* proc hook */
		case 4:
			for( i = 0; i < MAX_FUNC_HOOKS; i++ )
			{
				newp = *pp = (element *)malloc( sizeof( element ) );
				newp->lpn = proc_stop[ i ];
				pp = &newp->next;
			}
			result.errno = 0;
			break;
	}
	*pp = NULL;

	return( &result );
}

/******************************************************************/
/*
	FUNCTION

		jmdb_store_message

	DESCRIPTION:

		stores messages in area so that rpc_get_jm_message can get the info
		Called by cps_jm

	FORMAL PARAMETERS:

		cps_message
*/
/******************************************************************/
void jmdb_store_message( m )
cps_message *m;
{
	long *data;
	int i;

	if( jm_message == NULL )
		jm_message = (rpc_message *)hmalloc( sizeof( cps_message ) );

	jm_message->type = m->type;
	jm_message->src_lpn = m->src_lpn;
	jm_message->dst_lpn = m->dst_lpn;
	jm_message->size = m->size;

	if( jm_message->size > 16 )
		jm_message->size = 16;

	for( i = 0; i < jm_message->size; i++ )
		jm_message->msg[ i ] = m->msg[ i ];

	message_ready = TRUE;
}
