/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/*****************/
/* Include Files */
/*****************/

#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include "../inc/perr.h"
#include "../inc/param.h"
#include "../inc/shm.h"
#include "../inc/xfunc.h"


#define FACILITY "SHM_LOOK"

int num_user_proc;
int num_shm_proc;
int num_local_proc;
int num_semaphores;

/* Shared memory globals */

int *acp_nprocs;
int *acp_semids;
struct procmap *acp_pmap;
struct shmmap *acp_smap;
cps_message *acp_shmem;
cps_msg *acp_sshmem;
extern struct qelem *qelem;
extern int *q_head;
extern int *q_tail;
extern int *q_block;

/* Semaphore ids */

extern int q_semid;

display_error(str)
char *str;
{
	printf("%s\n",str);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  dump_shm

  DESCRIPTION:

	  [@tbs@]

								  */
/******************************************************************/

dump_shm()
{
	int i,j;

	struct qelem *q;

	cps_message *m;

	unsigned short array[32];

	int s_used[100],l_used[25];

	for (i=0; i<100; i++) s_used[i] = 0;
	for (i=0; i<25; i++) l_used[i] = 0;

	printf("Number of user processes = %d\n",acp_nprocs[0]);
	printf("Number of shm processes = %d\n",acp_nprocs[1]);
	printf("Number of local processes = %d\n",acp_nprocs[2]);
	printf("Queue semaphore id = %d\n",acp_semids[0]);

	printf("\nShared Memory Processes\n");
	printf(  "-----------------------\n");
	for (i=0; i<=num_shm_proc; i++) {
		printf("(%s,%s,%d) pid=%d semnum=%d sock=%d status=%d\n",
			acp_smap[i].hostname,acp_smap[i].hostaddr,acp_smap[i].portno,
			acp_smap[i].pid,acp_smap[i].semnum,acp_smap[i].sock,
			acp_smap[i].status);
	}

	printf("\nUser Processes\n");
	printf(  "---------------\n");
	for (i=0; i<num_user_proc; i++) {
		printf("local=%d pid=%d shmid=%d semnum=%d\n",
			acp_pmap[i].local,acp_pmap[i].pid,acp_pmap[i].shmid,
			acp_pmap[i].semnum);
	}

	printf("\nQueues (%d)\n",num_semaphores);
	printf(  "------\n");
	for (i=0; i<num_semaphores; i++) {
		if (q_head[i] == -1)
			q = NULL;
		else
			q = &qelem[q_head[i]];
		printf("%d: ",i);
		while (q != NULL) {
			printf("%d",q->buf);
			if (q->size == 0)
				printf("s ");
			else
				printf("l ");
			if (q->next == q->index) {
				printf(" (circular)");
			 	break;
			}
			if (q->next == -1)
				q = NULL;
			else
				q = &(qelem[q->next]);
		}
		printf("\n");
	}


	for (i=0; i<num_semaphores; i++) {
		if (q_head[i] == -1) {
			q = NULL;
		}
		else {
			q = &qelem[q_head[i]];
			printf("Queue %d:\n",i);
		}
		while (q != NULL) {
			if (q->size == 0) {
				s_used[q->buf] = -1;
				m = (cps_message *) &(acp_sshmem[q->buf]);
			}
			else {
				l_used[q->buf] = -1;
				m = &(acp_shmem[q->buf]);
			}
			log_cps_msg(stdout,m);
			if (q->next == -1)
				q = NULL;
			else
			q = &(qelem[q->next]);
		}
		printf("\n");
	}

	printf("Stashed small messages:\n");
	for (i=0; i<100; i++) {
		if (!s_used[i]) {
			m = (cps_message *) &(acp_sshmem[i]);
			log_cps_msg(stdout,m);
		}
	}
	printf("\n");

	printf("Stashed large messages:\n");
	for (i=0; i<25; i++) {
		if (!l_used[i]) {
			m = &(acp_shmem[i]);
			log_cps_msg(stdout,m);
		}
	}
	printf("\n");

	printf("\n");
	semctl(q_semid,0,GETALL,array);
	printf("Semaphore values: (");
	for (i=0; i<2*num_semaphores; i++) {
		j = (int) array[i];
		printf("%d ",j);
	}
	printf(")\n");

}

main(argc,argv) 
int argc;
char **argv;
{
	char *base_addr,*addr;

	int i,j;

	char fname[80];

	FILE *fopen(),*fp;

	int shmid;

	if (argc != 2) {
		printf("Usage: shm_look <job_name>\n");
		exit(0);
	}

	sprintf(fname,"%s.shm",argv[1]);
	fp = fopen(fname,"r");

	get_shmid:
	if (fscanf(fp,"%d",&shmid) == EOF) {
		printerr(FACILITY,ERROR_MSG,"Failed to attach to shared memory"
			,NULL,1);
		exit(0);
	}

/* Attach to the shared memory */

	if ((int)(base_addr = shm_attach(shmid)) == -1) goto get_shmid;

/* Initialize pointers into shared memory */

	addr = base_addr;
	acp_nprocs = (int *) addr;
	addr += 3 * sizeof(int);
	num_user_proc = acp_nprocs[0];
	num_shm_proc = acp_nprocs[1];
	num_local_proc = acp_nprocs[2];
	num_semaphores = num_local_proc + 3;

	acp_semids = (int *) addr;
	addr += 2 * sizeof(int);
	q_semid = acp_semids[0];

	acp_pmap = (struct procmap *) addr;
	addr += num_user_proc * sizeof(struct procmap);
	acp_smap = (struct shmmap *) addr;
	addr += (num_shm_proc + 1) * sizeof(struct shmmap);  /* One for the JM */

	acp_shmem = (struct shmem *) addr;
	addr += SHM_FUNC(num_semaphores) * sizeof(cps_message);
	acp_sshmem = (struct shmem_s *) addr;
	addr += S_SHM_FUNC(num_semaphores) * sizeof(cps_msg);

	qelem = (struct qelem *) addr;
	addr += (SHM_FUNC(num_semaphores) + S_SHM_FUNC(num_semaphores))
		* sizeof(struct qelem);
	q_head = (int *) addr;
	addr += num_semaphores * sizeof(int *);
	q_tail = (int *) addr;
	addr += num_semaphores * sizeof(int *);
	q_block = (int *) addr;

	dump_shm();
}
