/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				QUEUE.C													*/
/*																		*/
/*																		*/
/************************************************************************/



/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include <signal.h>


int acp_check_qnum(q)
int q;
{
	if ((q < 0) || (q > MAX_QUEUE_NUMBER)) {
		sprintf(acp_error_info,
			"Queue number (%d) out of range; must be between 0 and %d",
			q,MAX_QUEUE_NUMBER);
		acp_return_status = CPS_BAD_QUEUE_NUMBER;
		return(CPS_BAD_QUEUE_NUMBER);
	}
	return(CPS_OK);
}

int acp_check_narg(q,n)
int q,n;
{
	if (acp_queue_declared[q] != -1) {
		if (n != acp_queue_declared[q]) {
			sprintf(acp_error_info,
"Number of queue arguments does not agree with number declared (queue=%d)",
			q);
			acp_return_status = CPS_BAD_NUM_OF_ARGS;
			return(CPS_BAD_NUM_OF_ARGS);
		}
	}
	else {
		if ((n < 0) || (n > MAX_QUEUE_ARGS)) {
			sprintf(acp_error_info,
"Number of queue arguments out of range (queue=%d); must be between 0 and %d"
			,q,MAX_QUEUE_ARGS);
			acp_return_status = CPS_BAD_NUM_OF_ARGS;
			return(CPS_BAD_NUM_OF_ARGS);
		}
	}
	return(CPS_OK);
}

int acp_check_qarg(q,argnum,n)
int q;
int argnum;
int n;
{
	if (n % 4) {
		sprintf(acp_error_info,
"Byte count must be a multiple of four (queue=%d, arg#=%d, bcount=%d)",
		q,argnum,n);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}
	return(CPS_OK);
}

int acp_check_qargs(q,n,a)
int q;
int n;
int *a;
{
	int sum = 0;
	int i;

	int s1;

	for (i=1; i<=n; i++) {
		if ((s1 = acp_check_qarg(q,i,*a)) != CPS_OK) return(s1);
		sum += *a;
		a++;
	}

	if (sum > MAX_QUEUE_BYTES) {
		sprintf(acp_error_info,
"Byte count exceeded remote subroutine limit (queue=%d, count=%d, limit=%d)"
		,q,sum,MAX_QUEUE_BYTES);
		acp_return_status = CPS_BYTE_COUNT_EXCEEDED;
		return(CPS_BYTE_COUNT_EXCEEDED);
	}
	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_request_queue_info

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_request_queue_info(qn)
int qn;
{
	cps_msg *m;
	struct qelem *q;

	acp_request_complete = 0;

/* Request declaration from the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_req_q_info(m,acp_local_lpn,JM_LPN,qn);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait until the declaration has arrived */

	acp_wait_condition(&acp_request_complete,1);

	if (acp_queue_declared[qn] == -1) {
		sprintf(acp_error_info,
			"Failed to obtain queue declaration from JM");
		acp_return_status = CPS_QUEUE_UNDECLARED;
		return(CPS_QUEUE_UNDECLARED);
	}

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_make_array

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int acp_make_array(a,n
	,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *a,n;
int b1,b2,b3,b4,b5,b6,b7,b8,b9,b10;
int b11,b12,b13,b14,b15,b16,b17,b18,b19,b20;
int b21,b22,b23,b24,b25,b26,b27,b28,b29,b30;
int b31,b32;
{
	if (n == 0) return;
	*a = b1;
	a++;
	if (n == 1) return;
	*a = b2;
	a++;
	if (n == 2) return;
	*a = b3;
	a++;
	if (n == 3) return;
	*a = b4;
	a++;
	if (n == 4) return;
	*a = b5;
	a++;
	if (n == 5) return;
	*a = b6;
	a++;
	if (n == 6) return;
	*a = b7;
	a++;
	if (n == 7) return;
	*a = b8;
	a++;
	if (n == 8) return;
	*a = b9;
	a++;
	if (n == 9) return;
	*a = b10;
	a++;
	if (n == 10) return;
	*a = b11;
	a++;
	if (n == 11) return;
	*a = b12;
	a++;
	if (n == 12) return;
	*a = b13;
	a++;
	if (n == 13) return;
	*a = b14;
	a++;
	if (n == 14) return;
	*a = b15;
	a++;
	if (n == 15) return;
	*a = b16;
	a++;
	if (n == 16) return;
	*a = b17;
	a++;
	if (n == 17) return;
	*a = b18;
	a++;
	if (n == 18) return;
	*a = b19;
	a++;
	if (n == 19) return;
	*a = b20;
	a++;
	if (n == 20) return;
	*a = b21;
	a++;
	if (n == 21) return;
	*a = b22;
	a++;
	if (n == 22) return;
	*a = b23;
	a++;
	if (n == 23) return;
	*a = b24;
	a++;
	if (n == 24) return;
	*a = b25;
	a++;
	if (n == 25) return;
	*a = b26;
	a++;
	if (n == 26) return;
	*a = b27;
	a++;
	if (n == 27) return;
	*a = b28;
	a++;
	if (n == 28) return;
	*a = b29;
	a++;
	if (n == 29) return;
	*a = b30;
	a++;
	if (n == 30) return;
	*a = b31;
	a++;
	if (n == 31) return;
	*a = b32;
}

int acp_make_array1(a,n
	,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *a,n;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	if (n == 0) return;
	*a = *b1;
	a++;
	if (n == 1) return;
	*a = *b2;
	a++;
	if (n == 2) return;
	*a = *b3;
	a++;
	if (n == 3) return;
	*a = *b4;
	a++;
	if (n == 4) return;
	*a = *b5;
	a++;
	if (n == 5) return;
	*a = *b6;
	a++;
	if (n == 6) return;
	*a = *b7;
	a++;
	if (n == 7) return;
	*a = *b8;
	a++;
	if (n == 8) return;
	*a = *b9;
	a++;
	if (n == 9) return;
	*a = *b10;
	a++;
	if (n == 10) return;
	*a = *b11;
	a++;
	if (n == 11) return;
	*a = *b12;
	a++;
	if (n == 12) return;
	*a = *b13;
	a++;
	if (n == 13) return;
	*a = *b14;
	a++;
	if (n == 14) return;
	*a = *b15;
	a++;
	if (n == 15) return;
	*a = *b16;
	a++;
	if (n == 16) return;
	*a = *b17;
	a++;
	if (n == 17) return;
	*a = *b18;
	a++;
	if (n == 18) return;
	*a = *b19;
	a++;
	if (n == 19) return;
	*a = *b20;
	a++;
	if (n == 20) return;
	*a = *b21;
	a++;
	if (n == 21) return;
	*a = *b22;
	a++;
	if (n == 22) return;
	*a = *b23;
	a++;
	if (n == 23) return;
	*a = *b24;
	a++;
	if (n == 24) return;
	*a = *b25;
	a++;
	if (n == 25) return;
	*a = *b26;
	a++;
	if (n == 26) return;
	*a = *b27;
	a++;
	if (n == 27) return;
	*a = *b28;
	a++;
	if (n == 28) return;
	*a = *b29;
	a++;
	if (n == 29) return;
	*a = *b30;
	a++;
	if (n == 30) return;
	*a = *b31;
	a++;
	if (n == 31) return;
	*a = *b32;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_declare_queue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_declare_queue(qnum,nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	acp_fntype = "cps_";
	cpssta_declare_queue(qnum,nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";

}

cps_declare_queue_(qnum,nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	cps_declare_queue(qnum,nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
}

int cpssta_declare_queue_(qnum,nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	return(cpssta_declare_queue(qnum,nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32));
}

int cpssta_declare_queue(qnum,nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	int s;

	int bcount[MAX_QUEUE_ARGS];
	int i;

	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_func = "declare_queue";
	acp_set_of_procs = CPS_THIS_PROCESS;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DECLARE_QUEUE )

	s = acp_check_qnum(*qnum);
	if (s != CPS_OK) return(s);

	if (acp_queue_declared[*qnum] != -1) {
		sprintf(acp_error_info,"Queue number %d already declared",*qnum);
		acp_return_status = CPS_ALREADY_DECLARED;
		return(CPS_ALREADY_DECLARED);
	}

	s = acp_check_narg(*qnum,*nargs);
	if (s != CPS_OK) return(s);
	acp_make_array1(bcount,*nargs,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
	,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
	,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
	,b31,b32);
	s = acp_check_qargs(*qnum,*nargs,bcount);
	if (s != CPS_OK) return(s);

/* Send declaration to JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_declare_queue(m,acp_local_lpn,JM_LPN,*qnum,*nargs,bcount);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Store the queue information */

	acp_queue_declared[*qnum] = *nargs;
	if (*nargs > 0) {
		acp_queue_bcount[*qnum] = (int *) hmalloc((*nargs)*sizeof(int));
		bcopy(bcount,acp_queue_bcount[*qnum],(*nargs)*sizeof(int));
	}
	for (i=0; i<*nargs; i++) acp_queue_bsum[*qnum] += bcount[i];

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_queue_process

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_queue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	acp_fntype = "cps_";
	cpssta_queue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";

}

cps_queue_process_(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	cps_queue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
}

int cpssta_queue_process_(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	return(cpssta_queue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32));
}

int cpssta_queue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	static unsigned long *ssop = NULL;
	int *plist;

	int s;

	int baddr[MAX_QUEUE_ARGS];
	int i,n;

	struct qelem *q;
	cps_message *m;

	int size;

	acp_clear_error_info();
	acp_func = "queue_process";
	acp_set_of_procs = *sop;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_QUEUE_PROCESS )

	s = acp_check_qnum(*qnum);
	if (s != CPS_OK) return(s);

	if (acp_queue_declared[*qnum] == -1) {
		s = acp_request_queue_info(*qnum);
		if (s != CPS_OK) return(s);
	}

	else if (acp_queue_declared[*qnum] == -2) {
		if (*sop == CPS_END_OF_QUEUE) return(CPS_OK);
		sprintf(acp_error_info,
			"END_OF_QUEUE already placed on this queue");
		acp_return_status = CPS_QUEUE_ENDED;
		return(CPS_QUEUE_ENDED);
	}

	if (*sop == CPS_END_OF_QUEUE) {

/* Send END-OF-QUEUE message to JM */

		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_end_of_queue(m,acp_local_lpn,JM_LPN,*qnum);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();

		acp_queue_declared[*qnum] = -2;
		return(CPS_OK);
	}

/* Make an array of pointers */

	acp_make_array(baddr,acp_queue_declared[*qnum]
	,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
	,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
	,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
	,b31,b32);

/* Translate CPS set of processes into standard set of processes */

	if (ssop == NULL) ssop = create_empty_set(MAX_PROCESSES);
	acp_translate_sop(*sop,ssop);

/* Get an array of lpns terminated by zero */

	plist = set_query(ssop,MAX_PROCESSES);

/* Send a queue process message (to the JM) for each process */

	n = 0;
	while (plist[n] != 0) {

		size = 3*sizeof(long) + acp_queue_bsum[*qnum];
		if (size <= SMALL_MSG_SIZE) {
			q = acp_q_pop2(S_FREE_Q);
			m = (cps_message *) SMALL_MSG(q);
		}
		else {
			q = acp_q_pop2(L_FREE_Q);
			m = LARGE_MSG(q);
		}
		cps_bld_queue_process(m,acp_local_lpn,JM_LPN,plist[n],*qnum,
			acp_queue_declared[*qnum],acp_queue_bcount[*qnum],baddr);

		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();

		n++;
	}

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_dequeue_process

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_dequeue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	acp_fntype = "cps_";
	cpssta_dequeue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";

}

cps_dequeue_process_(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	cps_dequeue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
}

int cpssta_dequeue_process_(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	return(cpssta_dequeue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32));
}

int cpssta_dequeue_process(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	int s,i;
	char *x;

	int baddr[MAX_QUEUE_ARGS];

	acp_clear_error_info();
	acp_func = "dequeue_process";
	acp_set_of_procs = *sop;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DEQUEUE_PROCESS )

/* Check arguments & queue */

	s = acp_check_qnum(*qnum);
	if (s != CPS_OK) return(s);

	if (acp_queue_declared[*qnum] == -1) {
		s = acp_request_queue_info(*qnum);
		if (s != CPS_OK) return(s);
	}

	else if (acp_queue_declared[*qnum] == -2) {
		*sop = CPS_END_OF_QUEUE;
		return(CPS_OK);
	}

/* Dequeue a process and put the result in the globals acp_process_dequeued
   and acp_queue_data */

	acp_dequeue_p(sop,qnum,1);

/* Make an array of pointers */

	acp_make_array(baddr,acp_queue_declared[*qnum]
	,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
	,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
	,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
	,b31,b32);

/* Copy the outputs to the arguments */

	*sop = acp_process_dequeued;
	if (*sop != CPS_END_OF_QUEUE) {
		x = acp_queue_data;
		for (i=0; i<acp_queue_declared[*qnum]; i++) {
			bcopy(x,baddr[i],acp_queue_bcount[*qnum][i]);
			x += acp_queue_bcount[*qnum][i];
		}
	}
	return(CPS_OK);
}

int acp_dequeue_p(sop,qnum,wait)
int *sop,*qnum,wait;
{
	struct qelem *q;
	cps_msg *m;

	acp_queue_number = *qnum;
	acp_qreq_complete = 0;

/* Send a dequeue process message to the Job Manager */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_dequeue_process(m,acp_local_lpn,JM_LPN,*qnum,wait);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for a response from the JM */

	acp_wait_condition(&acp_qreq_complete,1);

	return;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_dequeue_if_possible

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_dequeue_if_possible(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	acp_fntype = "cps_";
	cpssta_dequeue_if_possible(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";

}

cps_dequeue_if_possible_(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	cps_dequeue_if_possible(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32);
}

int cpssta_dequeue_if_possible_(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	return(cpssta_dequeue_if_possible(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32));
}

int cpssta_dequeue_if_possible(sop,qnum,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,b11,b12
	,b13,b14,b15,b16,b17,b18,b19,b20,b21,b22,b23,b24,b25,b26,b27,b28
	,b29,b30,b31,b32)
int *qnum,*sop;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	int s,i;
	char *x;
	int baddr[MAX_QUEUE_ARGS];

	acp_clear_error_info();
	acp_func = "dequeue_if_possible";
	acp_set_of_procs = *sop;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DEQUEUE_IF_POSSIBLE )

/* Check arguments & queue */

	s = acp_check_qnum(*qnum);
	if (s != CPS_OK) return(s);

	if (acp_queue_declared[*qnum] == -1) {
		s = acp_request_queue_info(*qnum);
		if (s != CPS_OK) return(s);
	}

	else if (acp_queue_declared[*qnum] == -2) {
		*sop = CPS_END_OF_QUEUE;
		return(CPS_OK);
	}

/* Dequeue a process and put the result in the globals acp_process_dequeued
   and acp_queue_data */

	acp_dequeue_p(sop,qnum,0);

/* Make an array of pointers */

	acp_make_array(baddr,acp_queue_declared[*qnum]
	,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
	,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
	,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
	,b31,b32);

/* Copy the outputs to the arguments */

	*sop = acp_process_dequeued;
	if (*sop == CPS_EMPTY) return(CPS_OK);
	if (*sop == CPS_END_OF_QUEUE) return(CPS_OK);
	x = acp_queue_data;
	for (i=0; i<acp_queue_declared[*qnum]; i++) {
		bcopy(x,baddr[i],acp_queue_bcount[*qnum][i]);
		x += acp_queue_bcount[*qnum][i];
	}
	return(CPS_OK);
}
