/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				REMSUB.C												*/
/*																		*/
/*																		*/
/************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include "../inc/cps_msgs.h"
#include <signal.h>
#include <errno.h>

/************************/
/* External Definitions */
/************************/

static int dump_rsub()
{
	struct acp_remote_sub *r;

	printf("Remote Subroutines\n");
	printf("------------------\n");
	r = acp_subs;
	while (r != NULL) {
		printf("id: %d,  #args: %d\n",r->id,r->nargs);
		r = r->next;
	}
	fflush(stdout);
}

static int get_inactive_inst(rsub)
struct acp_remote_sub *rsub;
{
	int i;

/* Get an instance structure (we already know there is at least one) */

	i = 0;
	while ((rsub->inst[i] != NULL) && (rsub->inst[i]->active)) i++;

	if (rsub->inst[i] == NULL) {
		rsub->inst[i] = (struct acp_sub_instance *) 
		hmalloc(sizeof(struct acp_sub_instance));
		rsub->inst[i]->active = -1;
		rsub->inst[i]->complete = 0;
		rsub->inst[i]->baddr = (char **)
		hmalloc(rsub->nargs*sizeof(char *));
	}
	else {
		rsub->inst[i]->active = -1;
		rsub->inst[i]->complete = 0;
	}
	rsub->num_active++;

	return(i);
}

struct acp_remote_sub *acp_find_remsub(id)
int id;
{
	struct acp_remote_sub *r;

	r = acp_subs;
	while (r != NULL) {
		if (r->id == id) return(r);
		r = r->next;
	}
	return(NULL);
}

int acp_check_sub_narg(n)
int n;
{
	if ((n < 0) || (n > MAX_QUEUE_ARGS)) {
		sprintf(acp_error_info,
"Number of arguments out of range; must be between 0 and %d"
		,MAX_QUEUE_ARGS);
		acp_return_status = CPS_BAD_NUM_OF_ARGS;
		return(CPS_BAD_NUM_OF_ARGS);
	}
	return(CPS_OK);
}

int acp_check_sub_arg(argnum,n)
int argnum;
int n;
{
	if (n % 4) {
		sprintf(acp_error_info,
		"Byte count must be a multiple of four (arg#=%d, bcount=%d)",
			argnum,n);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}
	return(CPS_OK);
}

int acp_check_sargs(n,a)
int n;
int *a;
{
	int sum = 0;
	int i;

	int s1;

	for (i=1; i<=n; i++) {
		if ((s1 = acp_check_sub_arg(i,*a)) != CPS_OK) return(s1);
		sum += *a;
		a++;
	}

	if (sum > MAX_QUEUE_BYTES) {
		sprintf(acp_error_info,
			"Byte count exceeded remote subroutine limit (count=%d, limit=%d)"
			,sum,MAX_QUEUE_BYTES);
		acp_return_status = CPS_BYTE_COUNT_EXCEEDED;
		return(CPS_BYTE_COUNT_EXCEEDED);
	}
	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_request_rsub_decl

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_request_rsub_decl(id)
int id;
{
	cps_msg *m;
	struct qelem *q;

/* Request declaration from the JM */

	acp_rsub_complete = 0;

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_req_rsub_decl(m,acp_local_lpn,JM_LPN,id);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait until the declaration has arrived */

	acp_wait_condition(&acp_rsub_complete,1);

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_declare_subroutine

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_declare_subroutine(func,id,nargs,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int (*func)();
int *id,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	acp_fntype = "cps_";
	cpssta_declare_subroutine(func,id,nargs,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_declare_subroutine_(func,id,nargs,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int (*func)();
int *id,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	cps_declare_subroutine(func,id,nargs,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32);
}

int cpssta_declare_subroutine_(func,id,nargs,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int (*func)();
int *id,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	return(cpssta_declare_subroutine(func,id,nargs,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32));
}

int cpssta_declare_subroutine(func,id,nargs,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int (*func)();
int *id,*nargs;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	struct acp_remote_sub *rsub;

	char warning[132];

	struct qelem *q;
	cps_msg *m;

	int i;

	int s;

	acp_clear_error_info();
	acp_func = "declare_subroutine";
	acp_set_of_procs = CPS_THIS_PROCESS;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DECLARE_SUBROUTINE )

/* Check the number of arguments */

	s = acp_check_sub_narg(*nargs);
	if (s != CPS_OK) return(s);

/* Check to see if the remote subroutine was already declared */

	if ((rsub = acp_find_remsub(*id)) != NULL) {
		sprintf(warning,
			"%s%s: Warning - remote subroutine %d declared more than once",
			acp_fntype,acp_func,*id);
		acp_log_msg(warning);
	}
	else
	rsub = (struct acp_remote_sub *) hmalloc(sizeof(struct acp_remote_sub));

/* Check the byte count for each argument & the total byte count */

	rsub->id = *id;
	rsub->func = func;
	rsub->nargs = *nargs;
	acp_make_array1(rsub->bcount,*nargs
	,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
	,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
	,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
	,b31,b32);

	rsub->bsum = 0;
	for (i=0; i<(*nargs); i++) {
		rsub->argtyp[i] = CPS_READ_WRITE;
		rsub->bsum += rsub->bcount[i];
	}

	s = acp_check_sargs(*nargs,rsub->bcount);
	if (s != CPS_OK) {
		hfree(rsub);
		return(s);
	}

/* Add the remote subroutine to the global list */

	rsub->num_active = 0;
	rsub->inst = (struct acp_sub_instance **) 
	hmalloc(MAX_SUB_INSTANCES*sizeof(struct acp_sub_instance *));
	for (i=0; i<MAX_SUB_INSTANCES; i++) rsub->inst[i] = NULL;

	rsub->next = acp_subs;
	acp_subs = rsub;

/* Send the declaration to the Job Manager */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_sub_decl(m,acp_local_lpn,JM_LPN,rsub);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_declare_arguments

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_declare_arguments(id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	acp_fntype = "cps_";
	cpssta_declare_arguments(id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_declare_arguments_(id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	cps_declare_arguments(id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32);
}

int cpssta_declare_arguments_(id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	return(cpssta_declare_arguments(id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32));
}

int cpssta_declare_arguments(id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	struct acp_remote_sub *rsub;

	struct qelem *q;
	cps_msg *m;

	int i;

	int argtyp[MAX_QUEUE_ARGS];

	acp_clear_error_info();
	acp_func = "declare_arguments";
	acp_set_of_procs = CPS_THIS_PROCESS;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DECLARE_ARGUMENTS )

/* Check to see if the remote subroutine was already declared */

	if ((rsub = acp_find_remsub(*id)) == NULL) {
		sprintf(acp_error_info,"Remote subroutine %d undeclared",*id);
		acp_return_status = CPS_SUB_UNDECLARED;
		return(CPS_SUB_UNDECLARED);
	}

	acp_make_array1(argtyp,rsub->nargs
	,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
	,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
	,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
	,b31,b32);
	for (i=0; i<rsub->nargs; i++) rsub->argtyp[i] = argtyp[i];

/* Send the new declaration to the Job Manager */
/*
	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_sub_decl(m,acp_local_lpn,JM_LPN,rsub);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();
*/

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_remote_call

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_remote_call(m,q)
cps_message *m;
struct qelem *q;
{
	int (*func)();

	int id,inst;
	char *args;

	int reply_lpn;

	int i;
	char *b[32],*a;

	struct acp_remote_sub *rsub;

	cps_msg *m1;
	struct qelem *q1;

/* Increment the number of subroutines serviced */

	q1 = acp_q_pop2(S_FREE_Q);
	m1 = SMALL_MSG(q1);
	cps_bld_mon_update(m1,acp_local_lpn,JM_LPN,MOP_INC_CALL,acp_local_lpn,
	INC_CALLS_SERV,0);
	acp_q_push2(acp_shm_qnum,q1);
	acp_signal_shm();

/* Extract the data */

	cps_get_rem_call(m,&id,&inst,&acp_return_queue,&args);

/* Find the subroutine declaration */

	rsub = acp_find_remsub(id);
	if (rsub == NULL) {
		acp_log_msg("Failed to obtain remote subroutine declaration from JM");
		return;
	}

/* Create pointers to arguments */

	a = args;
	for (i=0; i<rsub->nargs; i++) {
		b[i] = a;
		a += rsub->bcount[i];
	}

/* Call the subroutine */

	func = rsub->func;
	switch (rsub->nargs) {
	case 0 :
		(*func)();
		break;
	case 1 :
		(*func)(b[0]);
		break;
	case 2 :
		(*func)(b[0],b[1]);
		break;
	case 3 :
		(*func)(b[0],b[1],b[2]);
		break;
	case 4 :
		(*func)(b[0],b[1],b[2],b[3]);
		break;
	case 5 :
		(*func)(b[0],b[1],b[2],b[3],b[4]);
		break;
	case 6 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5]);
		break;
	case 7 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6]);
		break;
	case 8 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7]);
		break;
	case 9 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8]);
		break;
	case 10 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9]);
		break;
	case 11 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10]);
		break;
	case 12 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11]);
		break;
	case 13 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11],b[12]);
		break;
	case 14 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11],b[12],b[13]);
		break;
	case 15 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11],b[12],b[13],b[14]);
		break;
	case 16 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11],b[12],b[13],b[14],b[15]);
		break;
	case 17 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11],b[12],b[13],b[14],b[15],b[16]);
		break;
	case 18 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11],b[12],b[13],b[14],b[15],b[16],b[17]);
		break;
	case 19 :
		(*func)(b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],
		b[10],b[11],b[12],b[13],b[14],b[15],b[16],b[17],b[18]);
		break;
	case 20 :
		break;
	case 21 :
		break;
	case 22 :
		break;
	case 23 :
		break;
	case 24 :
		break;
	case 25 :
		break;
	case 26 :
		break;
	case 27 :
		break;
	case 28 :
		break;
	case 29 :
		break;
	case 30 :
		break;
	case 31 :
		break;
	case 32 :
		break;
	default:
		break;
	}

/* If return_queue is CPS_WAIT or CPS_NO_WAIT, 
   send REMOTE_RETURN message to user */

	if ((acp_return_queue == CPS_WAIT) || (acp_return_queue == CPS_NO_WAIT)) {
		m->type = (long) MSG_REMOTE_RETURN;
		m->dst_lpn = m->src_lpn;
		m->src_lpn = (long) acp_local_lpn;
		reply_lpn = (int) m->dst_lpn;
		if (acp_send_to_user(q,reply_lpn,2) != CPS_OK) {
			acp_log_msg(acp_error_info);
		}
	}
	else {

/* Otherwise queue the process */

		long *x;

		m->type = (long) MSG_QUEUE_PROCESS;
		m->dst_lpn = JM_LPN;
		m->src_lpn = (long) acp_local_lpn;
		x = (long *) m->msg;
		*x = (long) acp_local_lpn;
		x++;
		*x = (long) acp_return_queue;

		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();

	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_service_calls

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_service_calls()
{
	struct qelem *q;
	cps_message *m;

	int state = PSTATE_RUNNING;

#if defined(MIPS) || defined(R3000)
	int omask;
#else
	sigset_t set,oset;
#endif

	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_SERVICE_CALLS )

#if defined(MIPS) || defined(R3000)
/*
	omask = sigblock(sigmask(SIGUSR1));
*/
	sighold(SIGUSR1);
#else

/* Create a signal mask with one bit set */

	sigemptyset(&set);
	sigaddset(&set,SIGUSR1);

/* Add SIGUSR1 to the current set of signals */

	sigprocmask(SIG_BLOCK,&set,&oset);

#endif

/* Check to see if the condition is satisfied */

	check_cond:

	if (acp_rms_tail == NULL) goto need_to_wait;

/* Change state */

	if (state != PSTATE_RUNNING) {
		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_RUNNING,0);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();
		state = PSTATE_RUNNING;
	}

/* Pop a message */

	q = acp_rms_head;
	if (q == NULL) {
		printerr(FACILITY,ERROR_MSG,"RMS queue empty when it shouldn't be",
			NULL,17);
		goto check_cond;
	}
	if (acp_rms_head->next == -1) {	/* List is now empty */
		acp_rms_head = NULL;
		acp_rms_tail = NULL;
	}
	else {
		acp_rms_head = &qelem[q->next];
	}

/* Restore signal mask */

#if defined(MIPS) || defined(R3000)
/*
	sigsetmask(omask);
*/
	sigrelse(SIGUSR1);
#else
	sigprocmask(SIG_SETMASK,&oset,NULL);
#endif

/* Call the remote subroutine */

	if (q->size)
		m = LARGE_MSG(q);
	else
		m = (cps_message *) SMALL_MSG(q);

	handle_remote_call(m,q);

/* Check to see if another is pending */

	goto check_cond;

/* Need to wait for a signal */

	need_to_wait:

	if (state != PSTATE_WAIT_SERVICE) {
		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_WAIT_SERVICE,0);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();
		state = PSTATE_WAIT_SERVICE;
		goto check_cond;
	}	

/* Wait for a SIGUSR1 signal */

#if defined(MIPS) || defined(R3000)
/*
	sigpause(omask);
*/
	sigpause(SIGUSR1);
#else
	sigsuspend(&oset);
#endif

	goto check_cond;

}

cps_service_calls_()
{
	cps_service_calls();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_call

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_call(sop,action,id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *sop,*action,*id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	acp_fntype = "cps_";
	cpssta_call(sop,action,id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_call_(sop,action,id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *sop,*action,*id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	cps_call(sop,action,id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32);
}

int cpssta_call_(sop,action,id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *sop,*action,*id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	return(cpssta_call(sop,action,id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32));
}

int cpssta_call(sop,action,id,
	b1,b2,b3,b4,b5,b6,b7,b8,b9,b10,
	b11,b12,b13,b14,b15,b16,b17,b18,b19,b20,
	b21,b22,b23,b24,b25,b26,b27,b28,b29,b30,
	b31,b32)
int *sop,*action,*id;
int *b1,*b2,*b3,*b4,*b5,*b6,*b7,*b8,*b9,*b10;
int *b11,*b12,*b13,*b14,*b15,*b16,*b17,*b18,*b19,*b20;
int *b21,*b22,*b23,*b24,*b25,*b26,*b27,*b28,*b29,*b30;
int *b31,*b32;
{
	static unsigned long *ssop = NULL;
	int *plist;

	static unsigned long *inst_set = NULL;
	int *ilist;

	int i,j,n;

	int s;

	int some_died = 0;
	int lpn;

	struct qelem *q;
	cps_message *m;

	char *baddr[32];

	int np;

	int size;

	struct acp_remote_sub *rsub;
	struct acp_sub_instance *inst;

	int call_and_queue;

#if defined(MIPS) || defined(R3000)
	int omask;
#else
	sigset_t set,oset;
#endif

	acp_clear_error_info();
	acp_func = "call";
	acp_set_of_procs = *sop;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_CALL )

/* Make sure we have the subroutine declaration */

	if ((rsub = acp_find_remsub(*id)) == NULL) {
		s = acp_request_rsub_decl(*id);
		if (s != CPS_OK) return(s);
		if ((rsub = acp_find_remsub(*id)) == NULL) {
			sprintf(acp_error_info,
"Failed to obtain remote subroutine declaration from JM (sub# = %d)",*id);
			acp_return_status = CPS_SUB_UNDECLARED;
			return(CPS_SUB_UNDECLARED);
		}
	}

/* Check to see if we are using the call-and-queue mechanism */

	if ((*action == CPS_WAIT) || (*action == CPS_NO_WAIT)) {
		if (inst_set == NULL) 
			inst_set = create_empty_set(MAX_SUB_INSTANCES);
		else
			set_clear(inst_set,MAX_SUB_INSTANCES);
		call_and_queue = 0;
	}
	else {
		s = acp_check_qnum(*action);	/* Check the queue number */
		if (s != CPS_OK) return(s);
		call_and_queue = -1;
	}

/* No instances are needed for the call-and-queue mechanism */

/* Translate CPS set of processes into standard set of processes */

	if (ssop == NULL) ssop = create_empty_set(MAX_PROCESSES);
	acp_translate_sop(*sop,ssop);
	np = set_size(ssop,MAX_PROCESSES);

/* Check to see if we've got enough inactive instance structures for the call */

	if (!call_and_queue) {
		if ((rsub->num_active + np) > MAX_SUB_INSTANCES) {
			sprintf(acp_error_info,
				"Maximum subroutine instances exceeded (sub# = %d)",*id);
			acp_return_status = CPS_MAX_INSTANCE_EXCEEDED;
			return(CPS_MAX_INSTANCE_EXCEEDED);
		}
	}

/* Get an array of lpns terminated by zero */

	plist = set_query(ssop,MAX_PROCESSES);

/* Call each remote subroutine */

	n = 0;
	while (plist[n] != 0) {

/* Get an instance structure (if necessary) */

		if (!call_and_queue) {
			i = get_inactive_inst(rsub);
			add_to_set(i+1,inst_set);
			rsub->inst[i]->lpn = plist[n];
			rsub->inst[i]->action = *action;
		}

/* Make array of argument addresses */

		if (!call_and_queue) {
			acp_make_array(rsub->inst[i]->baddr,rsub->nargs
			,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
			,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
			,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
			,b31,b32);
		}
		else {
			acp_make_array(baddr,rsub->nargs
			,b1,b2,b3,b4,b5,b6,b7,b8,b9,b10
			,b11,b12,b13,b14,b15,b16,b17,b18,b19,b20
			,b21,b22,b23,b24,b25,b26,b27,b28,b29,b30
			,b31,b32);
		}

		size = 3*sizeof(long) + rsub->bsum;
		if (size <= SMALL_MSG_SIZE) {
			q = acp_q_pop2(S_FREE_Q);
			m = (cps_message *) SMALL_MSG(q);
		}
		else {
			q = acp_q_pop2(L_FREE_Q);
			m = LARGE_MSG(q);
		}

		if (call_and_queue)
			cps_bld_rem_call(m,acp_local_lpn,plist[n],
				rsub->id,i,*action,rsub->nargs,baddr,rsub->bcount,
				rsub->argtyp);
		else
			cps_bld_rem_call(m,acp_local_lpn,plist[n],
				rsub->id,i,*action,rsub->nargs,rsub->inst[i]->baddr,
				rsub->bcount,rsub->argtyp);

		if (acp_send_to_user(q,plist[n],2) != CPS_OK) {
			acp_log_msg(acp_error_info);
		}

		n++;

	}

/* Increment number of calls */

	q = acp_q_pop2(S_FREE_Q);
	m = (cps_message *) SMALL_MSG(q);
	cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_INC_CALL,
		acp_local_lpn,INC_CALL,0);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Possibly wait for return (arguments copied back in background, unless
   call-and-queue mechanism is used, in which case no completion message
   is sent) */

	if (*action == CPS_WAIT) {

/* Tell the JM we are waiting for a subroutine call */

		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,
			acp_local_lpn,PSTATE_WAIT_CALL,*id);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();

		ilist = set_query(inst_set,MAX_SUB_INSTANCES);


#if defined(MIPS) || defined(R3000)
/*
		omask = sigblock(sigmask(SIGUSR1));
*/
		sighold(SIGUSR1);
#else

/* Create a signal mask with one bit set */

		sigemptyset(&set);
		sigaddset(&set,SIGUSR1);

/* Add SIGUSR1 to the current set of signals */

		sigprocmask(SIG_BLOCK,&set,&oset);

#endif

/* Check to see if the condition is satisfied */

		check_cond:

/* Check to see if any processes died */

		i = 0;
		while (ilist[i] != 0) {
			lpn = rsub->inst[ilist[i]-1]->lpn;
			if (acp_pmap[lpn-FIRST_USER_LPN].status == 0) { /* Process is dead */
				rsub->inst[ilist[i]-1]->complete = -1;
				some_died = -1;
			}
			i++;
		}

/* Check to see if all subroutines completed */

		i = 0;
		while (ilist[i] != 0) {
			if (!rsub->inst[ilist[i]-1]->complete) {
				goto need_to_wait;
			}
			i++;
		}

/* Restore signal mask */

#if defined(MIPS) || defined(R3000)
/*
		sigsetmask(omask);
*/
		sigrelse(SIGUSR1);
#else
		sigprocmask(SIG_SETMASK,&oset,NULL);
#endif

/* Deactivate all instances */

		i = 0;
		while (ilist[i] != 0) {
			rsub->inst[ilist[i]-1]->active = 0;
			rsub->num_active--;
			i++;
		}

/* Tell the JM we got the return from the subroutine call */

		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,
			acp_local_lpn,PSTATE_RUNNING,0);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();

		if (some_died)
			return(CPS_PARTNER_DEAD);
		else
			return(CPS_OK);

/* Need to wait for a signal */

		need_to_wait:

#if defined(MIPS) || defined(R3000)
/*
		sigpause(omask);
*/
		sigpause(SIGUSR1);
#else
		sigsuspend(&oset);
#endif

		goto check_cond;

	}

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_change_return_queue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_change_return_queue(newq)
int *newq;
{
	acp_fntype = "cps_";
	cpssta_change_return_queue(newq);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_change_return_queue_(newq)
int *newq;
{
	cps_change_return_queue(newq);
}

cpssta_change_return_queue_(newq)
int *newq;
{
	return(cpssta_change_return_queue(newq));
}

cpssta_change_return_queue(newq)
int *newq;
{
	int s;

	acp_clear_error_info();
	acp_func = "change_return_queue";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_CHANGE_RETURN_QUEUE )

	s = acp_check_qnum(*newq);
	if (s != CPS_OK) return(s);
	acp_return_queue = *newq;

	return(CPS_OK);
}
