/*****************************************************************
 *                                                               *
 * acp_vbbc.c - Module with routines that do data transfer using *
 *             the VBBC device driver.                           *
 *                                                               *
 * obs : This file should be linked with /usr/lib/libm.a         *
 *                                                               *
 * Author: Carmen Silva                                          *
 *                                                               *
 *****************************************************************/

#include <stdio.h>

#ifdef R3000

#include <sys/types.h>
#include <fcntl.h>
#include <math.h>
#include <fcntl.h>
#include <sys/vbbcreg.h>
#include "user_struct.h"
#include "../inc/perr.h"

#define FACILITY "LIBCPS"

#define BUFFER_SIZE   4096  /* number of bytes per page */


/************************************************************************
 *                                                               	*
 * acp_vbbc_get - Routine that gets data from a given node using the  	*
 *                vbbc device driver                                    *   
 *                                                               	*
 * Sintax:                                                       	*
 *	acp_vbbc_get (src, nbytes, dest_buf, offset)                    *
 *	              where:                                            * 
 *                     - struct acp_vbbc_info *dest_buf -  a pointer to *
 *      the structure that holds the information from where the data    *
 *      should be read                                                  * 
 *                     - char *src  - virtual address from the          *
 *      block where the data should be written to                       *
 *                     - nbyte  - total number of bytes to be           *
 *    transferred                                                       * 
 *                     - offset - block offset                       	*
 *									*
 * Author: Carmem Silva          					*
 * 02/11/92								*
 *									*
 ************************************************************************/

acp_vbbc_get(src,nbyte,dest_buf,offset)
char *src;
int nbyte;
struct acp_vbbc_info *dest_buf;
{
   int  size, rsize,tsize;
   char *buffer; 
   int i, status;
   char errmsg[80];

    
   tsize = nbyte;
 
   rsize = drem(tsize,BUFFER_SIZE);
   if (rsize < 0)
    rsize += BUFFER_SIZE;

   if ( ( rsize+= dest_buf->offset) > BUFFER_SIZE)
        rsize -= BUFFER_SIZE;         
   
   if ( (tsize < BUFFER_SIZE) && (dest_buf->npaddr < 2))
      size = tsize;
   else
      size = BUFFER_SIZE - dest_buf->offset;

   buffer = src;   

   for (i=0; i < dest_buf->npaddr; i++)
    {
      if ( (rsize > 0) && (i == (dest_buf->npaddr-1) ))
           size = rsize;

      status = acp_node_read(buffer, dest_buf->crate, dest_buf->vme_addr, 
                                       dest_buf->paddr[i], size);
      if (status != 0) {
	   sprintf(errmsg,"VBBC read error; error=%d, phyaddr=%x",
		status,dest_buf->paddr[i]);
	   printerr(FACILITY,ERROR_MSG,errmsg,NULL,37);
      }
      buffer += size; 
      size = BUFFER_SIZE;
    }

}


/************************************************************************
 *                                                               	*
 * acp_vbbc_send - Routine that sends data to a given node using the   	*
 *                vbbc device driver                                    * 
 *                                                               	*
 * Sintax:                                                       	*
 *	acp_vbbc_send ( src_buf, nbyte, dest, offset )                  *
 *	              where:                                            * 
 *                     - struct acp_vbbc_info *dest -  a pointer to     *
 *      the structure that holds the information to  where the data     *
 *      should be written                                               * 
 *                     - char *src_buf - virtual address from the       *
 *      block where the data should be read from                        *
 *                     - int nbytes - total number of bytes to be       *
 *    transferred                                                       *
 *                     - int offset  - block offset                     *
 *                                                                  	*
 *									*
 * Author: Carmem Silva          					*
 * 02/11/92								*
 *									*
 ************************************************************************/
acp_vbbc_send(src_buf,nbyte,dest,offset)
char *src_buf;
int nbyte,offset;
struct acp_vbbc_info *dest;
{
   int  size, rsize,tsize;
   char *buffer; 
   int i, status;
   char errmsg[80];

   tsize = nbyte;
   rsize = drem(tsize,BUFFER_SIZE);
   if (rsize < 0)
    rsize = BUFFER_SIZE + rsize;

   if ( ( rsize+= dest->offset) > BUFFER_SIZE)
        rsize -= BUFFER_SIZE;         
   
   if ( (tsize < BUFFER_SIZE) && (dest->npaddr < 2))
      size = tsize;
   else
      size = BUFFER_SIZE - dest->offset;

   buffer = src_buf;  

   for (i=0; i < dest->npaddr; i++)
    {
      if ( (rsize > 0) && ((i == dest->npaddr - 1 )))
           size = rsize;
     status = acp_node_write(buffer, dest->crate, dest->vme_addr,
                                                         dest->paddr[i],size); 
      if (status != 0) {
	   sprintf(errmsg,"VBBC %d write error; error=%d, phyaddr=%x",
	       i,status,dest->paddr[i]);
	   printerr(FACILITY,ERROR_MSG,errmsg,NULL,38);
      }

     buffer += size; 
                               
     size = BUFFER_SIZE;
    }

}



/************************************************************************
 *                                                               	*
 * acp_node_read - Routine that reads the memory of a given node. 	*
 *                                                               	*
 * Sintax:                                                       	*
 *	status = acp_node_read (virtual_add, crate, node, address, size)*
 *		where:                                                 	*
 *			int status - status returned from VBBC device   *
 *                                   driver. 0 means OK. -1            	*
 *									*
 *			void *virtual_add - Virtual address to where	*
 *					    data is transfered.		*
 *			int crate - Crate address			*
 *			int node - Node address (5 node switch foolowed	*
 *				   by 3 zeros)				*
 *			int phys_address - Physical memory address of	*
 *					   data				*
 *			int size - Number of bytes to be transfered.	*
 *									*
 * Author: Mariano Sumrell Miranda					*
 * 05/06/91								*
 *									*
 ************************************************************************/

int acp_node_read (virtual_add, crate, node, phys_address, size)
int *virtual_add;
int crate, node, phys_address, size;
{
   ulong vme_address;

   vme_address = (node << 24) | phys_address ;
   return (acp_vbbc_read (virtual_add, crate, vme_address, size));
}

/************************************************************************
 *                                                               	*
 * acp_vbbc_read - Routine that uses the VBBC device driver to read 	*
 *                 a VME address of a given crate.		 	*
 *                                                               	*
 * Sintax:                                                       	*
 *	status = acp_vbbc_read (virtual_add, crate, vme_address, size)	*
 *		where:                                                 	*
 *			int status - status returned from VBBC device   *
 *                                   driver. 0 means OK. -1           	*
 *									*
 *			void *virtual_add - Virtual address to where	*
 *					    data is transfered.		*
 *			int crate - Crate address			*
 *			int vme_address - VME address of data		*
 *			int size - Number of bytes to be transfered.	*
 *									*
 * Author: Mariano Sumrell Miranda					*
 * 05/06/91								*
 *									*
 ************************************************************************/

int acp_vbbc_read (virtual_add, crate, vme_address, size)
int *virtual_add;
int crate, vme_address, size;
{

   return ( vbbc_transfer(virtual_add, crate, vme_address, size, 0));

}

/************************************************************************
 *                                                               	*
 * acp_node_write - Routine that write to the memory of a given node. 	*
 *                                                               	*
 * Sintax:                                                       	*
 *	status = acp_node_write (virtual_add, crate, node,              *
 *				 address, size)				*
 *		where:                                                 	*
 *			int status - status returned from VBBC device   *
 *                                   driver. 0 means OK. -1             *
 *									*
 *			void *virtual_add - Virtual address from where	*
 *					    data is transfered.		*
 *			int crate - Crate address			*
 *			int node - Node address (5 bits from node	*
 *				   switch fllowed by 3 bits zeros)	*
 *			int phys_address - Physical memory address	*
 *					   where data will be written	*
 *			int size - Number of bytes to be transfered.	*
 *									*
 * Author: Mariano Sumrell Miranda					*
 * 05/06/91								*
 *									*
 ************************************************************************/

int acp_node_write (virtual_add, crate, node, phys_address, size)
int *virtual_add;
int crate, node, phys_address, size;
{
   ulong vme_address;

   vme_address = (node << 24) | phys_address ;
   return (acp_vbbc_write (virtual_add, crate, vme_address, size));
}

/************************************************************************
 *                                                               	*
 * acp_vbbc_write - Routine that uses the VBBC device driver to write 	*
 *                  to a VME address of a given crate.		 	*
 *                                                               	*
 * Sintax:                                                       	*
 *	status = acp_vbbc_write (virtual_add, crate, vme_address, size)	*
 *		where:                                                 	*
 *			int status - status returned from VBBC device   *
 *                                   driver. 0 means OK. -1            	*
 *									*
 *			void *virtual_add - Virtual address from where	*
 *					    data is transfered.		*
 *			int crate - Crate address			*
 *			int vme_address - VME address to write  data	*
 *			int size - Number of bytes to be transfered.	*
 *									*
 * Author: Mariano Sumrell Miranda					*
 * 05/06/91								*
 *									*
 ************************************************************************/

int acp_vbbc_write (virtual_add, crate, vme_address, size)
int *virtual_add;
int crate, vme_address, size;
{
   return ( vbbc_transfer(virtual_add, crate, vme_address, size, 1));
}

/************************************************************************
 *                                                               	*
 * vbbc_transfer - Routine that read or write transfer using the VBBC	*
 *		   device driver.					*
 *                                                               	*
 * Sintax:                                                       	*
 *	status = vbbc_transfer (virtual_add, crate, vme_address, size,	*
 *				operation);				*
 *		where:                                                 	*
 *			int status - status returned from VBBC device   *
 *                                   driver. 0 means OK. -1            	*
 *									*
 *			void *virtual_add - Virtual address from where	*
 *					    data is transfered.		*
 *			int crate - Crate address			*
 *			int vme_address - VME address to write  data	*
 *			int size - Number of bytes to be transfered.	*
 *			int operation  = 0 ==> read operation		*
 *				       = 1 ==> write operation		*
 *									*
 * Author: Mariano Sumrell Miranda					*
 * 05/06/91								*
 *									*
 ************************************************************************/

static int vbbc_transfer (virtual_add, crate, vme_address, size, operation)
int *virtual_add;
int crate, vme_address, size, operation;
{
   static int fd = 0;		/* fd for VBBC device driver. If 0,
				   hasn't been opened yet		*/
   bb_operation bb_op;
   bb_operation_status bb_status;


   if ( !fd )				/* If VBBC device driver is not	*/
   {					/* open, open it		*/

      fd = open ("/dev/vbbc", O_RDWR);
      if (fd < 0 )
      {
	 fd = 0 ;			/* Clear fd			*/
	 return (-1);		/* If can't open, return -1)	*/
      }
   }

/*****	Initialize the bb_operation structure	*****/
   bb_op.byte_count = size;
   bb_op.vme_address = (ulong)vme_address;
   bb_op.virtual_address = (caddr_t) virtual_add ;
   bb_op.write_enable = (char) operation;	/* Define if read or write */
   bb_op.flags = 0;
   bb_op.crate_id = (ushort) crate;
   bb_op.status = &bb_status;

/*****	Do ioctl to perform read	*****/
   ioctl (fd, BB_XFER, &bb_op);

/*****	Return operation status		*****/
   return(bb_status.error_code);
}


#endif
