/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*                                                                      */
/*                                                                      */
/*                       Message Queue Utility Routines                 */
/*                                                                      */
/*                                                                      */
/************************************************************************/



/*****************/
/* Include Files */
/*****************/

#include <stdio.h>
#include "../inc/msgqueue.h"

/**************************/
/* Parameter Declarations */
/**************************/



/*********************/
/* Macro Definitions */
/*********************/



/********************/
/* Global Variables */
/********************/

static struct msg_queue *free_q = NULL;	    /* Queue of free messages */

/************************/
/* External Definitions */
/************************/


/******************************************************************/
/*
    FUNCTION NAME:

      create_queue

  DESCRIPTION:

      Creates a message queue.

								  */
/******************************************************************/

struct msg_queue *create_queue()
{
	struct msg_queue *q;

	q = (struct msg_queue *) hmalloc(sizeof(struct msg_queue));
	q->front = NULL;
	q->back = NULL;
	q->len = 0;
	q->lock = 0;
	return(q);
}

/******************************************************************/
/*
    FUNCTION NAME:

      enter_q

  DESCRIPTION:

      Places an element onto a queue.

  FORMAL PARAMETERS:

      struct msg_queue *q - Queue to enter onto.

      struct queue_element *e - Element to enter.

								  */
/******************************************************************/


enter_q(q,e)
struct msg_queue *q;
struct queue_element *e;
{

	if (q->lock == 1) {
		printf("Recursive queue operation (enter_q)\n");
		exit(1);
	}

	/* Place element in the back of the queue */

	q->lock = 1;	    /* Operation in progress */
	if (q->back == NULL) {
		q->back = e;
		q->front = e;
		q->len = 1;
		e->next = NULL;
	}
	else {
		q->back->next = e;
		q->back = e;
		q->len++;
		e->next = NULL;
	}
	q->lock = 0;

}

/******************************************************************/
/*
    FUNCTION NAME:

      remove_q

  DESCRIPTION:

      Removes an element from a queue.

  FORMAL PARAMETERS:

      struct msg_queue *q - Queue to remove element from.

  FUNCTION VALUE:

      (struct queue_element *) - Element removed from the queue.

								  */
/******************************************************************/

struct queue_element *remove_q(q)
struct msg_queue *q;
{

	struct queue_element *e,*e1;

	if (q->lock == 1) {
		printf("Recursive queue operation (remove_q)\n");
		exit(1);
	}

	q->lock = 1;
	e = q->front;
	if (e == NULL) {
		q->lock = 0;
		return(e);
	}
	e1 = e->next;
	e->next = NULL;
	q->front = e1;
	q->len--;
	if (e1 == NULL) q->back = NULL;
	q->lock = 0;
	return(e);

}

/******************************************************************/
/*
    FUNCTION NAME:

      alloc_queue_elem

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

struct queue_element *alloc_queue_elem()
{
	struct queue_element *e;

	if (free_q == NULL) free_q = create_queue();
	e = remove_q(free_q);
	if (e == NULL)
		e = (struct queue_element *) hmalloc(sizeof(struct queue_element));
	return(e);
}

/******************************************************************/
/*
    FUNCTION NAME:

      dealloc_queue_elem

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

								  */
/******************************************************************/

dealloc_queue_elem(e)
{
	enter_q(free_q,e);
}
