/*
**
** Demo program:
** 
** Various bracket manipulations.
** 
** --- Leo Michelotti
** --- June 17, 1994
**
*/


#include "mxyzptlk.rsc"

main() {
 double u1( 0.32 ), u2( 0.5 ), 
        v1( -3.1 ), v2( 1.5 );

 Jet::Setup( 4, 6, 4 );

 double w, y, z, answer;
 coord  x1( u1 ), x2( u2 ), p1( v1 ), p2( v2 );
 Jet    a, b, c, pb;

 // -- Calculation of Poisson bracket via Jets
 a = (x1*x1) * (x2*x2*x2) * p1 * (p2*p2*p2*p2);
 b = sin( x1 * (p2*p2) * (x2*x2*x2) );
 pb = a^b;
 cout << "Computed by Jet: " << pb.standardPart() << "\n";

 // -- Hand calculations
 w = (u1*u1) * (u2*u2*u2) * v1 * (v2*v2*v2*v2);
 y =      u1 * (v2*v2) * (u2*u2*u2)  ;
 z = cos( y );
 answer = w*y*z*( 6.0/(u2*v2) - 1.0/(u1*v1) - 12.0/(u2*v2) );
 cout << "Exact answer   : " << answer << "\n";

 cout << "And also       : " 
      << ( ( (x1*x1) * (x2*x2*x2) * p1 * (p2*p2*p2*p2) ) ^
           ( sin( x1 * (p2*p2) * (x2*x2*x2) ) ) 
         ).standardPart()
      << "\n\n";

 // -- Test of the Jacobi identity
 c = exp( p1*x1 + p2*x2 );
 cout << "Jacobi identity" << endl;
 ( (a^(b^c)) + (b^(c^a)) + (c^(a^b)) ).printCoeffs();

 // -- Hamiltonian vector fields
 LieOperator V_a ( a  );
 LieOperator V_b ( b  );
 LieOperator V_pb( pb );

 LieOperator W = ( V_a ^ V_b );

 cout << "Hamiltonian test" << endl;
 ( V_pb + W ).printCoeffs();
}
