      
#ifndef CLIEOPERATOR_HXX
#define CLIEOPERATOR_HXX

typedef char (*MX_C_FUNCPTR)(const int*, const complex);

class CLieOperator {
private:
  JetC* v;
  int accuWgt;
  int weight;
  complex* refPoint; // Point about which derivatives are to be expanded.

  // Friends
  friend class  JetC;
  friend struct JLC;
  friend void   JLCFixReferenceAtStart( const CLieOperator& );

public:

  // Constructors and destructors_____________________________________
  CLieOperator();
  CLieOperator( CLieOperator& );
  CLieOperator( LieOperator& );
  CLieOperator( JetC* );     
  CLieOperator( char* );   // Creates the identity function.  
  ~CLieOperator();

  // Member functions (public)
  void reset();     // Resets the private data.
  void fixReference( complex* );
  void fixReference();
  void fixReference( JetC& );
  void fixReferenceAtEnd( const CLieOperator& );
  void fixReferenceAtStart( const CLieOperator& );
  void setVariable( const JetC&, const int );
  void setVariable( const complex, const int );
  void getReference( complex* );
  void peekAt();
  void peekAt( char* );
  dlist* image();
  void printCoeffs();
  void printCoeffs( char* );
  void standardPart( complex* ) const;
  void weightedDerivative( int*, complex* );
  void derivative( int*, complex* );
  CLieOperator filter( int, int );
  CLieOperator filter( MX_C_FUNCPTR [] );
  MatrixC Jacobian() const;

  CLieOperator inverse();
  CLieOperator Inverse() const;
  CLieOperator EpsInverse() const;

  char isNilpotent();

  // Friendly operators ...........................
  CLieOperator operator()( CLieOperator& ) const;  // Self explanatory ...

  friend ostream& operator<<(ostream&, CLieOperator&);
  friend istream& operator>>(istream&, CLieOperator&);
  friend char operator==(const  CLieOperator&, const CLieOperator& );
  friend char operator!=( const CLieOperator&, const CLieOperator& );  
  char operator==( const complex );
  char operator!=( const complex );
  CLieOperator& operator=( const CLieOperator& );
  JetC operator[]( int );
  JetC operator() ( const int i ) const { return v[i];}

  JetC operator^( JetC& );         // Action as a Lie operator
  JetC expMap( const double, const JetC& );    // Performs exponential map on
                                //  the second argument.
  JetC expMap( const JetC&, const JetC& );    // The orbit parameter is here
                                //  a Jet variable.
  CLieOperator expMap( const complex, const CLieOperator& ); // Performs exponential map 
                                               //  componenet wise.
  CLieOperator expMap( const JetC&,    const CLieOperator& );    

  friend CLieOperator operator*( const CLieOperator&, const CLieOperator& );
                                // This is concatenation.

  // Arithmetic
  friend CLieOperator operator+( const CLieOperator&, const CLieOperator& );
  friend CLieOperator operator-( const CLieOperator&, const CLieOperator& );
  friend CLieOperator operator-( const CLieOperator& );
  friend CLieOperator operator*( const complex&, const CLieOperator& );
  friend CLieOperator operator*( const CLieOperator&, const complex& );
  friend CLieOperator operator*( const JetC&, const CLieOperator& );
  friend CLieOperator operator*( const CLieOperator&, const JetC& );

  friend  CLieOperator operator*( MatrixC&, const CLieOperator& );

} ;

void operator*=( CLieOperator& x, const CLieOperator& y );

#endif // CLIEOPERATOR_HXX
