/*
 * Header file for class JetVector
 * 
 * Version 1.0       April, 1995
 * 
 * Leo Michelotti
 * michelot@calvin.fnal.gov
 * 
 */


#ifndef JETVECTOR_H
#define JETVECTOR_H

#include <iostream.h>
#include <math.h>
#include "mxyzptlk.hxx"
#include "Vector.h"

class JetVector 
{
private:
  int     dim;
  Jet*    comp;

public:
  // Constructors and the destructor ...
  JetVector( const int   /* dim */        = 3,
             const Jet*  /* components */ = 0 );
  JetVector( const JetVector& );
  ~JetVector();

  // Assignment ...
  void             Set              ( const Jet* );
  Jet&             operator()       ( int ) const; // return or set
                                              // component
  // Algebraic functions ...
  JetVector&       operator=      ( const JetVector& );

  JetVector        operator+      ( const JetVector& ) const;
  JetVector        operator+=     ( const JetVector& );
  friend JetVector operator-      ( const JetVector& );
  JetVector        operator-      ( const JetVector& ) const;
  JetVector        operator-=     ( const JetVector& );
  JetVector        operator*      ( const Jet&    ) const;
  JetVector        operator*      (     double    ) const;
  friend JetVector operator*      ( const Jet&, const JetVector& );
  friend JetVector operator*      (     double, const JetVector& );
  JetVector        operator*=     ( const Jet&    );
  JetVector        operator*=     (     double    );
  JetVector        operator/      ( const Jet&    ) const;
  JetVector        operator/      (     double    ) const;
  JetVector        operator/=     ( const Jet&    );
  JetVector        operator/=     (     double    );


  Jet              operator*      ( const JetVector& ) const; // dot product
  JetVector        operator^      ( const JetVector& ) const; // cross product:
                                                        // only works if
                                                        // the vector is
                                                        // three-dimensional
  JetVector        operator^      ( Vector ) const;
  friend
  JetVector        operator^      ( Vector, const JetVector& );

  // Boolean functions ...
  char          operator==     ( const JetVector& ) const;
  char          operator!=     ( const JetVector& ) const;
  char          operator<      ( const JetVector& ) const;
  char          operator<=     ( const JetVector& ) const;
  char          operator>      ( const JetVector& ) const;
  char          operator>=     ( const JetVector& ) const;
  char          IsNull         () const;
  char          IsUnit         () const;

  // Queries ...
  int  Dim() const { return dim; }

  // Utilities ..
  void        peekAt           () const;
  void        printCoeffs      () const;
  Jet         Norm             () const;
  JetVector   Unit             () const;           // returns unit vector
  void        Rotate           ( JetVector& v, 
                                 double theta ) const;
  void        Rotate           ( JetVector& v, 
                                 const Jet& theta ) const;
                                                   // rotates v through 
                                                   // an angle theta using
                                                   // *this as the axis

  friend ostream& operator<<( ostream&, const JetVector& );
} ;



#endif  // JETVECTOR_H
