#include "mxyzptlk.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}

// ================================================================

#ifndef MX_SMALL
#define MX_SMALL    1.0e-12 // Used by JL::addTerm to decide 
                            //   removal of a JLterm.
#endif  // MX_SMALL

#ifndef MX_MAXITER
#define MX_MAXITER  100     // Maximum number of iterations allowed
                            //   in iterative routines
#endif  // MX_MAXITER

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::operator+=( const Jet& y ) {
                     // ??? Come up with a more efficient implementation SOON.
                                       // ??? This should be modified so that
                                       // terms beyond the accurate weight of
 JLterm* p;                            // x or y are not computed and carried
 JLterm* q;                            // into the answer.
 JL* xPtr =   jl;
 JL* yPtr = y.jl;

 dlist_iterator gx( *(dlist*) xPtr );
 dlist_iterator gy( *(dlist*) yPtr );
 int i;
 
 // Check for consistency and set reference point of the sum.
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
   if( fabs( xPtr->refPoint[i] - yPtr->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) 
   {
     cout << "\n*** JL ADDITION ERROR: Inconsistent reference points." << endl;
     exit(0);
   }
 
 // If one of the arguments is void, then return the other ..
 if( xPtr->count < 1 ) { (*this) = y; return; }
 if( yPtr->count < 1 ) return;

 // .. otherwise, continue normal operations.
 while( p = (JLterm*) gy() ) {
   q = new JLterm( p );
   xPtr->addTerm( q );
 }
 
 // Determine the maximum weight computed accuratelyPtr->
 if( xPtr->accuWgt > yPtr->accuWgt ) xPtr->accuWgt = yPtr->accuWgt;
}

void operator-=( Jet& x, const Jet& y ) {
 x = x - y;
}

void operator*=( Jet& x, const Jet& y ) {
 x = x * y;
}

void operator/=( Jet& x, const Jet& y ) {
 x = x / y;
}

void operator/=( Jet& x, const double y ) {
 x = x / y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

//     Member functions(public)  |||||||||||||||||||||||||||||||

void Jet::addTerm( JLterm* a) {
 jl->addTerm( a );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::fixReference( const double* r ) {
 jl->fixReference( r );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::fixReference() {
 jl->fixReference();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||


//***************************************************************
//***************************************************************
//***************************************************************
//
//      The overloaded operators for class Jet
//
istream& operator>>(istream& is,  Jet& x) {
 return operator>>( is, *(x.jl) );
}

ostream& operator<<(ostream& os, Jet& x) {
 return operator<<( os, *(x.jl) );
}

char operator==( const Jet& x, const Jet& y ) {
 return *(x.jl) == *(y.jl);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const Jet& x, const double& y ) {
 return *(x.jl) == y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const double& y, const Jet& x ){
 return *(x.jl) == y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const Jet& x, const Jet& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const Jet& x, const double& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const double& x, const Jet& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::operator+=( const double& x ) {   // ??? Untested!!
 if( jl == NULL ) jl = new JL;              // ??? DANGER: what is the 
                                            // ??? reference point?
 jl->operator+=( x );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::operator-=( const double& x ) {
 operator+=( -x );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator+( const Jet& x, const Jet& y ) { 
                                       // ??? This should be modified so that
 Jet z;                                // terms beyond the accurate weight of
 JLterm* p;                            // x or y are not computed and carried
 JLterm* q;                            // into the answer.
 JL* xPtr = x.jl;
 JL* yPtr = y.jl;
 JL* zPtr = z.jl;

 dlist_iterator gx( *(dlist*) xPtr );
 dlist_iterator gy( *(dlist*) yPtr );
 int i;
 
 // Check for consistency and set reference point of the sum.
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( fabs( xPtr->refPoint[i] - yPtr->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << "\n*** JL ADDITION ERROR: Inconsistent reference points." << endl;
     exit(0);
   }
   zPtr->refPoint[i] = xPtr->refPoint[i];
 }
 
 // If one of the arguments is void, then return the other ..
 if( xPtr->count < 1 ) return y;   // ??? This will not work.  What's 
 if( yPtr->count < 1 ) return x;   // ??? returned owns the same data
                                   // ??? as x or y!!  Change throughout!!
 // .. otherwise, continue normal operations.
 if( xPtr->count > yPtr->count ) {
   *zPtr = *xPtr;
   while( p = (JLterm*) gy() ) {
     q = new JLterm( p );
     zPtr->addTerm( q );
   }
 }
 else {
   *zPtr = *yPtr;
   while( p = (JLterm*) gx() ) {
     q = new JLterm( p );
     zPtr->addTerm( q );
   }
 }
 
 // Determine the maximum weight computed accurately.
 if( xPtr->accuWgt < yPtr->accuWgt ) zPtr->accuWgt = xPtr->accuWgt;
 else                                zPtr->accuWgt = yPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator+( const Jet& x, const double& y ) {
 Jet z;    // !!! "Jet z = x;" does not work.  The copy constructor
 z = x;    // ??? is called, which makes x and z own the same data!!
 z += y;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator+( const double& y, const Jet& x ) {
 Jet z;
 z = x;
 z += y;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator-( const Jet& y ) {  
 Jet z;
 JL* yPtr = y.jl;
 JL* zPtr = z.jl;   // ??? This is a dangerous strategy. 
 dlist_iterator getNext( *(dlist*) yPtr );  // ??? It will fail if
 JLterm* p;                 // ??? The statement "z = ...: appears.         
 JLterm* q;
 int i;
 
 // If the argument is void, then return it ..
 if( y->count < 1 ) return y;

 // .. otherwise continue normal operations.
 zPtr->clear();
 
 // Insert the first one
 p = (JLterm*) getNext();
 q = new JLterm;
 *q = *p;
 q->value = - ( p->value );
 zPtr->insert( q );
 
 // and append the rest
 while( p = (JLterm*) getNext() ){
   q = new JLterm;
   *q = *p;
   q->value = - ( p->value );
   zPtr->append( q );
   }
 
 zPtr->weight  = y->weight;
 zPtr->accuWgt = y->accuWgt;
 zPtr->count   = y->count;
 
 // Set the reference point.
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = yPtr->refPoint[i];
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator-( const Jet& x, const Jet& y ) {  
 Jet z;
 JLterm* p;
 JLterm* q;
 dlist_iterator gx( *(dlist*) x.jl );
 dlist_iterator gy( *(dlist*) y.jl );
 int i;
 
 // Check for consistency and set reference point of the difference.
 static char divCheck;
 divCheck = 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( fabs( x->refPoint[i] - y->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     divCheck = 1;
     cout << "\n*** Jet SUBTRACTION ERROR: Inconsistent reference points.\n" 
          << "x->refPoint[ " << i << " ] = " << x->refPoint[i] << "\n"
          << "y->refPoint[ " << i << " ] = " << y->refPoint[i] << "\n"
          << endl;
   }
   z->refPoint[i] = x->refPoint[i];
 }
 if( divCheck ) exit(0);
 
 // If one of the arguments is void, then return the other ..
 if( x->count < 1 ) return -y;
 if( y->count < 1 ) return x;

 // .. otherwise, continue normal operations.
 if( x->count > y->count ) {
   z = x;
   while( p = (JLterm*) gy() ) {
     q = new JLterm( p );
     ( q->value ) = - ( q->value );
     z->addTerm( q );
     }
   }
 
 else {
   z = - y;
   while( p = (JLterm*) gx() ) {
     q = new JLterm( p );
     z->addTerm( q );
     }
   }
 
 // Determine the maximum weight computed accurately.
 if( x->accuWgt < y->accuWgt ) z->accuWgt = x->accuWgt;  // ??? Change this too, as in
 else                          z->accuWgt = y->accuWgt;  // ??? operaotr+
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator*( const Jet& x, const Jet& y ) {        
                                       // ??? This should be modified so that
 Jet z;                                // terms beyond the accurate weight of
 JLterm* p;                            // x or y are not computed and carried
 JLterm* q;                            // into the answer.
 JLterm* r;

 JL* xPtr = x.jl;
 JL* yPtr = y.jl;
 JL* zPtr = z.jl;

 dlist_looper gx( *(dlist*) xPtr );
 dlist_looper gy( *(dlist*) yPtr );
 int i;
 
 // Check for consistency and set reference point of the product.
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( fabs( xPtr->refPoint[i] - yPtr->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << "\n*** JL MULTIPLICATION ERROR: Inconsistent reference points." << endl;
     exit(0);
     }
   zPtr->refPoint[i] = xPtr->refPoint[i];
   }
 
 // If one of the arguments is void, return it ..
 if ( xPtr->count < 1 ) return x;
 if ( yPtr->count < 1 ) return y;

 // .. otherwise continue normal operations.
 while( p = (JLterm*) gy() ) {
 while( q = (JLterm*) gx() ) {
 
   if( ( p->weight + q->weight ) > Jet::oneEnv.MaxWeight ) continue;
                                   // ??? Change this: compute zPtr->accuWgt earlier 
   r = new JLterm;                 // ??? and use it.
   *r = (*p)*(*q);
   zPtr->addTerm( r );
   }}
 
 // Determine the maximum weight computed accurately.
 if( xPtr->accuWgt < yPtr->accuWgt ) zPtr->accuWgt = xPtr->accuWgt;
 else                                zPtr->accuWgt = yPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator*( const Jet& x, const double y ) {  
 Jet z;
 JLterm* p;
 JLterm* q;
 JL* xPtr = x.jl;
 JL* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 if ( xPtr->count < 1 ) return x;
 if( y == 0.0 ) {
   z = 0.0;
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];  // Since zPtr is no longer
   z.stacked = 1;                        // correct, use z.
   return z;
   }
 
 while( p = (JLterm*) gx() ) {
   q = new JLterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator*( const Jet& x, const int j ) { 
 Jet z;
 JLterm* p;
 JLterm* q;
 JL* xPtr = x.jl;
 JL* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 double y;
 
 if ( xPtr->count < 1 ) return x;
 if( ( y = (double) j ) == 0.0 ) {
   z = 0.0;
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];
   z.stacked = 1;
   return z;
   }
 
 while( p = (JLterm*) gx() ) {
   q = new JLterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator*( const double y, const Jet& x ) {  
 Jet z;
 JLterm* p;
 JLterm* q;
 JL* xPtr = x.jl;
 JL* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 if ( xPtr->count < 1 ) return x;
 if( y == 0.0 ) {
   z = 0.0;
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];
   z.stacked = 1;
   return z;
   }
 
 while( p = (JLterm*) gx() ) {
   q = new JLterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator*( const int j, const Jet& x ) { 
 Jet z;
 JLterm* p;
 JLterm* q;
 JL* xPtr = x.jl;
 JL* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 double y;
 
 if ( xPtr->count < 1 ) return x;
 if( ( y = (double) j ) == 0.0 ) {
   z = 0.0;
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];
   z.stacked = 1;
   return z;
   }
 
 while( p = (JLterm*) gx() ) {
   q = new JLterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator/( const Jet& x, const double& y ) { 
 Jet z;
 JLterm* p;
 JLterm* q;
 JL* xPtr = x.jl;
 JL* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 // Check for void operators ..
 if ( xPtr->count < 1 ) return x;

 // Check for division by zero ..
 if( y == 0.0 ) {
   cout << ::form("\n*** JL ERROR *** ");
   cout << ::form("Attempt to divide by a scalar zero.\n");
   cout << endl;
   exit(0);
 }
 
 // .. and continue.
 while( p = (JLterm*) gx() ) {
   q = new JLterm( p );
   q->value /= y;
   zPtr->addTerm( q );
 }
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator/( const Jet& x, const int& j ) { 
 Jet z;
 JLterm* p;
 JLterm* q;
 JL* xPtr = x.jl;
 JL* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 double y;
 
 // Check for void operators ..
 if ( xPtr->count < 1 ) return x;

 // Check for division by zero ..
 if( ( y = (double) j ) == 0.0 ) {
   cout << ::form("\n*** JL ERROR *** ");
   cout << ::form("Attempt to divide by a scalar zero.\n");
   cout << endl;
   exit(0);
 }
 
 // .. and continue.
 while( p = (JLterm*) gx() ) {
   q = new JLterm( p );
   q->value /= y;
   zPtr->addTerm( q );
 }
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator/( double a, const Jet& b ) {
  return Jet(a)/b;
} 

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator|( const Jet& u, const Jet& v ) { 
 /*
 Truncated multiplication: used only by the division operator.
 No checks are made on the reference point.
 */
 Jet z;
 JLterm* p;
 JLterm* q;
 JLterm* r;
 dlist_looper gu( *(dlist*) u.jl );
 dlist_looper gv( *(dlist*) v.jl );
 
 // If one of the arguments is void, return it ..
 if ( u->count < 1 ) return u;
 if ( v->count < 1 ) return v;

 // .. otherwise continue normal operations.
 while( p = (JLterm*) gv() ) {
 while( q = (JLterm*) gu() ) {
   if( ( p->weight + q->weight ) > Jet::oneEnv.weightLimit ) continue;
   r = new JLterm;
   *r = (*p)*(*q);
   z.addTerm( r );
 }}
 
 // Determine the maximum weight computed accurately.
 if( u->accuWgt < v->accuWgt ) z->accuWgt = u->accuWgt;
 else                          z->accuWgt = v->accuWgt;
 
 z.stacked = 1;   // ??? Should the reference point be copied???
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator/( const Jet& wArg, const Jet& uArg ) { 
 dlist_looper gu( *(dlist*) uArg.jl );
 dlist_looper gw( *(dlist*) wArg.jl );
 Jet v;
 Jet vn;
 Jet w, u;
 JLterm* qu;
 JLterm* qw;
 double u0, w0;    // ??? What is w0 for?
 int wgt;
 int i;
 
 // Check for void operators ..
 if ( wArg->count < 1 ) return wArg;
 if ( uArg->count < 1 ) {
   cout << ::form("\n*** JL ERROR *** ");
   cout << ::form("Attempt to divide by a void JL variable.\n");
   cout << endl;
   exit(0);
   }
 
 // Check for consistency and set reference point of the quotient.
 static char divCheck;
 divCheck = 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( fabs( wArg->refPoint[i] - uArg->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     divCheck = 1;
     cout << "\n*** Jet DIVISION ERROR: Inconsistent reference points.\n" 
          << "wArg->refPoint[ " << i << " ] = " << wArg->refPoint[i] << "\n"
          << "uArg->refPoint[ " << i << " ] = " << uArg->refPoint[i] << "\n"
          << endl;
   }
   v->refPoint[i] = wArg->refPoint[i];
 }
 if( divCheck ) exit(0);
 
 // Normalize the denominator
 if( ( qu = (JLterm*) gu() ) == 0 ) {
   cout << ::form( "\n*** JL ERROR ***  Division algorithm called with uninitialized JL.\n" );
   cout << endl;
   exit(0);
   }
 if( ( wgt = qu->weight ) != 0 ) {
   cout << ::form( "\n*** JL ERROR ***  Attempt to divide by nilpotent element.\n" );
   cout << endl;
   exit(0);
   }
 if( ( u0 = qu->value ) == 0.0 ) {
   cout << ::form( "\n*** JL ERROR ***  Attempt to divide by zero.\n" );
   cout << endl;
   exit(0);
   }
 u = uArg / u0;
 
 // Recursive algorithm
 qw = (JLterm*) gw();
 if( qw->weight == 0 ) v = w0 = ( qw->value );
 else                  v = w0 = 0.0;
 
 Jet::oneEnv.weightLimit = 0;
 while ( Jet::oneEnv.weightLimit < Jet::oneEnv.MaxWeight ) {
   Jet::oneEnv.weightLimit++;
   vn = u|v;
   vn = vn.filter( Jet::oneEnv.weightLimit, Jet::oneEnv.weightLimit );
   w = wArg.filter( Jet::oneEnv.weightLimit, Jet::oneEnv.weightLimit );
   v += ( w - vn );
   }
 
 // Correct normalization of the answer ..
 v /= u0;
 
 // .. set the reference point ..
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
   v->refPoint[i] = wArg->refPoint[i];
 
 // Determine the maximum weight computed accurately.
 if( wArg->accuWgt < uArg->accuWgt ) v->accuWgt = wArg->accuWgt;
 else                                v->accuWgt = uArg->accuWgt;
 
 // .. and return value.
 v.stacked = 1;
 return v;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet operator^( const Jet& x, const Jet& y ) { 
 Jet z;
 int i;
 int* m = new int [ Jet::oneEnv.NumVar ];
 int* n = new int [ Jet::oneEnv.NumVar ];
 
 if( Jet::oneEnv.PBOK ) {
 
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
     m[i] = 0;
     n[i] = 0;
     }
 
   for( i = 0; i < Jet::oneEnv.dof; i++ ) {
     m[ i ] = 1;
     n[ i + Jet::oneEnv.dof ] = 1;
 
     z += ( ( x.D(m) * y.D(n) ) - ( x.D(n) * y.D(m) ) );
 
     m[ i ] = 0;
     n[ i + Jet::oneEnv.dof ] = 0;
     }
 
   delete [] m;
   delete [] n;
   z.stacked = 1;
   return z;
   }
 
 cout << ::form( "\n*** ERROR ***                                \n" );
 cout << ::form(   "*** ERROR *** JL operator^( JL&, JL& )       \n" );
 cout << ::form(   "*** ERROR *** Wrong JLSetup used; cannot     \n" );
 cout << ::form(   "*** ERROR *** perform bracket.               \n" );
 cout << ::form(   "*** ERROR ***                                \n" );
 cout << endl;
 
 if( 2 + 3 == 5 ) exit(1);       // Done to avoid compiler warnings
 delete [] m;                    // about leaving the function.
 delete [] n;
 z.stacked = 1;
 return z;                       
}


//***************************************************************
//***************************************************************
//***************************************************************
//
//      Overloaded analytic functions

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet acos( const Jet&  ) {   // ??? Write this damned thing!!!  It's trivial!!
 Jet z;
 cout << ::form( "Jet acos( Jet& x ) to be written.\n" );
 cout << endl;
 exit(0);
 z.stacked = 1;
 return z;
}
 
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet asin( const Jet& x ) { 
 Jet z;
 Jet dz;
 z = x;
 dz = ( sin(z) - x ) / cos(z);
 int iter = 0;
 while( ( dz != 0.0 ) && ( iter++ < MX_MAXITER ) ) {
  z -= dz;
  dz = ( sin(z) - x ) / cos(z);
 }
 if( iter >= MX_MAXITER ) {
  cout << "*** WARNING ***                             \n";
  cout << "*** WARNING *** JL& asin( JL& x ) {         \n";
  cout << "*** WARNING *** Over " 
       << MX_MAXITER 
       << " iterations used;    \n";
  cout << "*** WARNING *** result may be incorrect.    \n";
  cout << "*** WARNING ***                             \n";
  cout << "*** WARNING *** Initial value:              \n";
  cout << "*** WARNING ***                             \n";
  cout << endl;
  x.printCoeffs();
  cout << "*** WARNING *** A few iterations:             " << endl;
  for( iter = 0; iter < 4; iter++ ) {
   cout << "*** WARNING *** Iteration No. " << iter+1 << endl;
   z -= dz;
   dz = ( sin(z) - x ) / cos(z);
   cout << "*** WARNING *** dz = " << endl;
   dz.printCoeffs();
   cout << "*** WARNING *** z = " << endl;
   z.printCoeffs();
  }
 }
 z.stacked = 1;
 return z;
}
 
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet atan( const Jet& x ) { 
 Jet z;
 Jet c, dz;
 z = x;
 c = cos( z );
 dz = c*( sin(z) - x*c );
 int iter = 0;
 while( ( dz != 0.0 ) && ( iter++ < MX_MAXITER ) ) {
  z -= dz;
  c = cos( z );
  dz = c*( sin(z) - x*c );
 }
 if( iter >= MX_MAXITER ) {
  cout << ::form( "*** WARNING ***                             \n" );
  cout << ::form( "*** WARNING *** JL& atan( JL& x ) {         \n" );
  cout << ::form( "*** WARNING *** Over %d iterations used;    \n", MX_MAXITER );
  cout << ::form( "*** WARNING *** result may be incorrect.    \n" );
  cout << ::form( "*** WARNING ***                             \n" );
  cout << endl;
 }

 z.stacked = 1;
 return z;
}
 
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet cos( const Jet& x ) { 
 
 Jet epsSin( const Jet& );
 Jet epsCos( const Jet& );
 Jet epsilon;
 double cs, sn;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  exp called with unevaluated JL.\n");
   cout << endl;
   return x;
   }
 
 p = (JLterm*) getNext();
 if( ( p -> weight ) == 0 ) {         // x has non-zero standard part
   cs = cos( p -> value );
   sn = sin( p -> value );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 // pops the standard part off epsilon
   delete p;
   return cs*epsCos( epsilon ) - sn*epsSin( epsilon );
   }
 else {                               // x is pure infinitesimal
   return epsCos( x );
   }
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet cosh( const Jet& x ) { 
 Jet z;
 z = exp(x);
 z = ( z + ( 1.0 / z ) ) / 2.0;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet exp( const Jet& x ) { 
 
 Jet epsExp( const Jet& );
 Jet epsilon;
 double factor;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  exp called with unevaluated JL.\n");
   cout << endl;
   return x;
   }
 
 p = (JLterm*) getNext();
 if( ( p -> weight ) == 0 ) {         // x has non-zero standard part
   factor = exp( p -> value );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 // pops the standard part off epsilon
   delete p;
   return factor*epsExp( epsilon );
   }
 else {                               // x is pure infinitesimal
   return epsExp( x );
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet log ( const Jet& x ) { 
 Jet             epsilon;
 Jet             u, w;
 double          std, n;
 JLterm*         p;
 dlist_iterator  getNext( *(dlist*) x.jl );
 
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  log called with unevaluated JL.\n");
   cout << endl;
   return x;
   }
 
 p = (JLterm*) getNext();
 if( ( ( p -> weight      ) ==  0   ) &&
     ( ( std = p -> value ) !=  0.0 )
   )                                  // x has non-zero standard part
   {
   if( x->count == 1 ) {             // x may have no derivatives
     epsilon = log( std );
     return epsilon;
   }
   epsilon = x;                      // x must not be altered by the routine
   p = epsilon.get();                //   pops the standard part off epsilon
   delete p;
   epsilon.scaleBy( - 1.0/std );
   w = epsilon;
   u = epsilon;                      // We use here the formulae
   n = 1.0;                          // ln( s + e )   = ln s + ln( 1 + e/s ), and
   while( u->count > 0 ) {           // ln( 1 + e/s ) = ln( 1 - ( -e/s )) 
     u *= epsilon;                   //               = - sum ( - e/s )^n / n
     w += ( u / ++n );
   }
 
   return ( log(std) - w );
 // -------------------
 
   }
 else                                 // x has zero standard part
   {
   cout << ::form( "\n\n*** ERROR *** \n"                                  );
   cout << ::form(     "*** ERROR *** JL log called with argument whose\n" );
   cout << ::form(     "*** ERROR *** standard part vanishes.\n"           );
   cout << ::form(     "*** ERROR *** \n\n"                                );
   cout << endl;
   if( 1 + 1 == 2 ) exit(1);
   return x;
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet log10( const Jet& x ) {
 static const double logE = 0.4342944819032518276511289;
 return  logE*log(x);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet pow( const Jet& x, const double s ) { 
 Jet epsPow( const Jet&, const double );
 Jet epsilon;
 double factor;
 double std;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLterm* p;
 int u;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  pow called with unevaluated JL.\n");
   cout << endl;
   return x;
 }
 
 p = (JLterm*) getNext();
 if( ( ( p -> weight      ) ==  0   ) &&
     ( ( std = p -> value ) !=  0.0 )
   )                                  // x has non-zero standard part
   {
   if( x->count == 1 ) {             // x may have no derivatives
     epsilon = pow( std, s );
     return epsilon;
   }
   factor = pow( std, s );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 //   pops the standard part off epsilon
   delete p;
   epsilon.scaleBy( 1.0/std );
   epsilon = factor*epsPow( epsilon, s );
   epsilon.stacked = 1;
   return epsilon;
   }
 else                                 // x is pure infinitesimal
   {
   u = nearestInteger( s );
   if( s != (double) u ) {
     cout << ::form( "\n*** ERROR::JL pow: attempt to use infinitesimal " );
     cout << ::form( "as base with non-integer exponent.\n" );
     cout << endl;
     exit(1);
     }
   epsilon = 1.0;
   if ( u == 0 ) {
     epsilon.stacked = 1;
     return epsilon;
   }
   if ( u > 0 ) {
     while( u-- > 0 ) epsilon *= x;
     epsilon.stacked = 1;
     return epsilon;
     }
   else {
     while( u++ < 0 ) epsilon *= x;
     epsilon = 1.0 / epsilon;
     epsilon.stacked = 1;
     return epsilon;
     }
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet pow( const Jet& x, int n ) {
  Jet z;
  static int i;
  if( n == 0 ) z = 1.0;
  else if( n > 0 ) {
    z = x;
    for( i = 2; i <= n; i++ ) z *= x;
  }
  else {
    Jet xr;
    z = 1.0 / x;
    xr = z;
    for( i = -2; i >= n; i-- ) z *= xr;
  }
  z.stacked = 1;
  return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet sin( const Jet& x ) { 
 
 Jet epsSin( const Jet& );
 Jet epsCos( const Jet& );
 Jet epsilon;
 double cs, sn;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  sin called with unevaluated JL.\n");
   cout << endl;
   return x;
   }
 
 p = (JLterm*) getNext();
 if( ( p -> weight ) == 0 ) {         // x has non-zero standard part
   cs = cos( p -> value );
   sn = sin( p -> value );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 // pops the standard part off epsilon
   delete p;
   return sn*epsCos( epsilon ) + cs*epsSin( epsilon );
   }
 else {                               // x is pure infinitesimal
   return epsSin( x );
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet sinh( const Jet& x ) {
 Jet z;
 z = exp(x);
 return ( z - (1.0/z))/2.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet sqrt( const Jet& x ) {
 Jet epsSqrt( const Jet& );
 Jet epsilon;
 double factor;
 double std;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  sqrt called with unevaluated JL.\n");
   cout << endl;
   return x;
   }
 
 p = (JLterm*) getNext();
 if( ( ( p -> weight      ) ==  0   ) &&
     ( ( std = p -> value ) !=  0.0 )
   )                                  // x has non-zero standard part
   {
   factor = sqrt( std );
   epsilon = x;                       // x must not be altered by the routine
   if( epsilon->count == 1 )          // x may have no derivatives
     {
     epsilon = factor;
     return epsilon;
     }
   else                               // normal case
     {
     p = epsilon.get();               // pops the standard part off epsilon
     delete p;
     epsilon.scaleBy( 1.0/std );
     return factor*epsSqrt( epsilon );
     }
   }
 else                                 // x is pure infinitesimal
   {
   cout << ::form( "\n*** ERROR::JL sqrt: attempt to use infinitesimal argument." );
   cout << endl;
   if( 1 + 1 == 2 ) exit(1);
   return x;
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet tan( Jet& x ) { 
 return sin(x) / cos(x) ;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet tanh( const Jet& x ) { 
 return sinh(x) / cosh(x);
}
 

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet fabs( const Jet& x ) {
 static double u;

 if( x->count == 0 ) {
   cout << "\n*** WARNING ***  fabs called with unevaluated JL.\n" << endl;
   return x;
   }
 
 if( (u = x.standardPart()) != 0.0 ) 
 {
   if( u > 0.0 ) return x;
   else          return -x;
 }
 else
 {
   cout << "\n*** ERROR::Jet fabs: attempt to use infinitesimal argument."
        << endl;
   if( 1 + 1 == 2 ) exit(1);   // This weird stuff done to avoid
   return x;                   // compiler warnings.
 }
}


//***************************************************************
//***************************************************************
//***************************************************************
//
//      Auxiliary functions which operate on infinitesimals.
//      It is assumed that the calling program checks to be sure
//      that the argument has zero standard part.
 
Jet epsCos( const Jet& epsilon ) { 
 
 Jet z( 1.0 );
 Jet epsq, term;
 double n;
 int i;
 
 epsq = - epsilon*epsilon;
 
 term = epsq / 2.0;
 n = 2.0;
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
   z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term = ( ( term*epsq ) / ++n ) / ++n;
   }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet epsExp( const Jet& epsilon ) { 
 Jet z( 1.0 );
 Jet term;
 double n;
 int i;
 
 term = epsilon;
 n = 1.0;
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
   z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term = ( term*epsilon ) / ++n;
 }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet epsPow( const Jet& epsilon, const double s ) { 
 Jet z( 1.0 );
 Jet term;
 double f, n;
 int i;
 
 f = s;
 n = 1.0;
 term = s*epsilon;
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
          z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term = ( (--f) * term * epsilon ) / ++n;
   }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet epsSin( const Jet& epsilon ) { 
 
 Jet z;
 Jet epsq, term;
 double n;
 
 epsq = - epsilon*epsilon;
 z = epsilon;
 term = ( epsilon*epsq ) / 6.0;
 n = 3.0;
 
 while( term->count > 0 ) {
   z += term;
   term = ( ( term*epsq ) / ++n ) / ++n;
 }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet epsSqrt( const Jet& epsilon ) {  // This function is identical to epsPow
                                     // with the substitution  s = 1/2
 Jet     z( 1.0 );
 Jet     term;
 double  f, n;
 int     i;
 
 f    = 1.0 / 2.0;
 n    = 1.0;
 term = f*epsilon;
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
          z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term *= ( (--f) * epsilon ) / ++n;
 }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}
 

//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of Class Jet

void Jet::fixReference( const Jet& x ) {
 jl->fixReference( *(x.jl) );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char Jet::isNilpotent() {
 return jl->isNilpotent();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::peekAt() const {
 jl->peekAt();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::printCoeffs() const {
 jl->printCoeffs();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

dlist Jet::image() {      // ??? Problem: nothing is recorded about the
                          // ??? reference point.
 return jl->image();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::writeToFile( char* fileName ) {
 jl->writeToFile( fileName );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::writeToFile( FILE* filePtr ) {
 jl->writeToFile( filePtr );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::scaleBy( double y ) {
 jl->scaleBy( y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLterm* Jet::get() {
 return jl->get();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double Jet::standardPart() const {
 return jl->standardPart();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::clear() {
 jl->clear();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double Jet::weightedDerivative( int* ind ) {
 return jl->weightedDerivative( ind );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double Jet::derivative( int* ind ) {
 return jl->derivative( ind );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::filter( const int wgtLo, const int wgtHi ) const { 
 Jet z;
 JL* zPtr = z.jl;
 JL* thisPtr = jl;
 dlist_iterator getNext( *(dlist*) thisPtr );
 JLterm* p;
 JLterm* q;
 int numTerms;
 int wgt, upperWgt;
 
 if( ( wgtLo <= 0 ) && 
     ( wgtHi >= thisPtr->weight )   
   ) {
   z = *this;
   z.stacked = 1;
   return z;
 } 

 numTerms = 0;
 upperWgt = 0;
 
 while( p = (JLterm*) getNext() ) {
   wgt = p->weight;
   if( ( wgt >= wgtLo ) && ( wgt <= wgtHi ) ) {
     q = new JLterm( p );
     zPtr->append( q );
     if( wgt > upperWgt ) upperWgt = wgt;
     numTerms++;
     }
   }
 
 zPtr->count = numTerms;   // ??? Is this necessary?
 zPtr->weight = upperWgt;  // ??? Where is the copy of the re
 zPtr->accuWgt = thisPtr->accuWgt;   // ??? reference point???

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::filter( char (*f) ( const int*, const double ) ) { 
 Jet z;
 JL* zPtr =  z.jl;
 JL* thisPtr = jl;
 dlist_iterator getNext( *(dlist*) thisPtr );
 JLterm* p;
 JLterm* q;

 int i;
 int nv = oneEnv.NumVar;
 int* oldIndex = new int[ nv ];

 while( p = (JLterm*) getNext() ) {
  for( i = 0; i < nv; i++ ) oldIndex[i] = p->index[i];

  if( (*f)( p->index, p->value ) ) {
   q = new JLterm( p );
   zPtr->append( q );
  }

  for( i = 0; i < nv; i++ ) 
   if( oldIndex[i] != p->index[i] ) {
    cout << ::form( "\n" );
    cout << ::form( "*** ERROR ***                                      \n" );
    cout << ::form( "*** ERROR *** Jet Jet::filter( char (*f) ( int*, double ) \n" );
    cout << ::form( "*** ERROR *** The test routine is not allowed      \n" );
    cout << ::form( "*** ERROR *** to change the values in the index    \n" );
    cout << ::form( "*** ERROR *** array!!                              \n" );
    cout << ::form( "*** ERROR ***                                      \n" );
    cout << endl;
    exit(0);
   }
 }

 delete [] oldIndex;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet concat( const Jet& x ) {
 Jet v;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLterm* p;
 int i;
 
 v = 0.0;
 v.fixReference( Jet::oneEnv.JLmonomial[0] );
 
 while( p = (JLterm*) getNext() ) {
   if( p->weight > x->accuWgt ) break;
	for( i = 0; i < Jet::oneEnv.NumVar; i++ )
     Jet::oneEnv.exponent[i] = (p->index)[i];
   Jet::monoCode();
   v += ( p->value ) * ( Jet::oneEnv.JLmonomial[ Jet::monoRank() ] );
 }
 
 v.stacked = 1;
 return v;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::operator() ( Jet* y ) const { 
 Jet*     u = new Jet [ oneEnv.NumVar ];
 Jet      z, term;
 int      i, j, w;
 JLterm*  p;

 // Check consistency of reference points.
 for( i = 0; i < oneEnv.NumVar; i++ ) {
   for( j = 0; j < oneEnv.NumVar; j++ )
     if( fabs( y[i]->refPoint[j] - y[0]->refPoint[j] ) > MX_SMALL*Jet::oneEnv.scale[j] ) {
       cout << ::form( "\n*** Jet Jet::operator( Jet* ) ERROR: Inconsistent reference points.\n" );
       cout << ::form("y[%d].refPoint[%d] != y[0].refPoint[%d]\n", i, j, j);
       cout << ::form("%lf != %lf\n", y[i]->refPoint[j], y[0]->refPoint[j]);
       cout << endl;
       exit(1);
     }
   if( fabs( (*this)->refPoint[i] - y[i].weightedDerivative( oneEnv.AllZeroes ) ) 
       > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << ::form( "\n*** Jet Jet::operator( Jet* ) ERROR: Inconsistent reference points.\n" );
     cout << ::form("(*this)-> refPoint[%d] != y[%d].weightedDerivative( oneEnv.AllZeroes )\n",
             i, i);
     cout << ::form("%lf != %lf\n",
             (*this)-> refPoint[i], y[i].weightedDerivative( oneEnv.AllZeroes ));
     cout << endl;
     exit(1);
   }
 }
 
 // Subtract reference point prior to concatenation.
 for( i = 0; i < oneEnv.NumVar; i++ ) {
   u[i] = y[i];
   p    = ( u[i] ).get();
   if( p->weight != 0 ) {
     cout << ::form( "\n*** JL JL::operator() ERROR: Argument has no nonstandard part!!\n" );
     cout << endl;
     exit(0);
   }
   delete p;
 }
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 oneEnv.JLmonomial[0] = 1.0;
 oneEnv.JLmonomial[0].fixReference( y[0] );
 
 // For all higher weights ...
 for( w = 1; w <= (*this)->accuWgt; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, oneEnv.NumVar, oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = oneEnv.JLmonomial[ Jet::monoRank() ];
     ( oneEnv.exponent[i] )++;
     Jet::monoCode();
     oneEnv.JLmonomial[ Jet::monoRank() ] = term * u[i];   // ??? Is this OK???
 
     } // End while loop.
 
 // Monomials have been stored.
 // Now traverse the JL variable and evaluate.
 
 z = concat( *this );  // ??? START HERE, LEO!!!
 
 // Set the correct reference point in z.
 for( i = 0; i < oneEnv.NumVar; i++ ) 
   z->refPoint[i] = y[0]->refPoint[i];
 
 delete [] u;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::operator() ( LieOperator& y ) const { 
 static countErrors = 0;
 Jet z;
 LieOperator u;
 Jet term;
 int i;
 int w;
 dlist_iterator getNext( *(dlist*) jl );
 JLterm* p;
 
 u = y;
 u.reset();

 // Check consistency of reference points.
 if( countErrors < 20 ) {
   countErrors++;
   for( i = 0; i < oneEnv.NumVar; i++ )
     if( fabs( jl->refPoint[i] - ( y.v[i] ).weightedDerivative( oneEnv.AllZeroes ) ) 
 	 > MX_SMALL*Jet::oneEnv.scale[i] ) {
       cout << "\n*** WARNING *** Jet Jet::operator( LieOperator& ): Inconsistent reference points.\n";
       cout << "*** WARNING *** jl->refPoint[" << i << "] != ( y.v[" 
 			       << i << "] ).weightedDerivative( oneEnv.AllZeroes )\n";
       cout << "*** WARNING *** " << jl->refPoint[i] << " != " << ( y.v[i] ).weightedDerivative( oneEnv.AllZeroes )
 	    << endl;
     }
 }

 // In the event that the argument y is null
 if( u.weight <= 0 ) {
  if( ( p = (JLterm*) getNext() ) == 0 ) {   // ??? Should this be an error
   z.stacked = 1;                            // ??? condition instead?
   return z;                                 
  }
  if( p->weight == 0 ) z = 0.0;
  else                 z = p->value;
  z.stacked = 1;
  return z;
 }

 // Subtract reference point prior to concatenation.
 for( i = 0; i < oneEnv.NumVar; i++ ) {
  p = ( u.v[i] ).get();
  if( p->weight != 0 ) {
    cout << ::form( "\n*** JL JL::operator() ERROR: Argument has no nonstandard part!!\n" );
    cout << endl;
    exit(0);
  }
  delete p;
 }
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 oneEnv.JLmonomial[0] = 1.0;
 oneEnv.JLmonomial[0].fixReferenceAtStart( y );
 
 // For all higher weights ...
 for( w = 1; w <= jl->accuWgt; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, oneEnv.NumVar, oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = oneEnv.JLmonomial[ Jet::monoRank() ];
     ( oneEnv.exponent[i] )++;
     Jet::monoCode();
     oneEnv.JLmonomial[ Jet::monoRank() ] = term * u.v[i];   // ??? Is this OK???
 
     } // End while loop.
 
 // Monomials have been stored.
 // Now traverse the JL variable and evaluate.
 
 z = concat( *this );
 
 // Set the correct reference point in z.
 for( i = 0; i < oneEnv.NumVar; i++ ) z->refPoint[i] = y.refPoint[i];
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

double Jet::operator() ( double* x ) const {
 return jl->operator()( x );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::D( int* n ) const { 
 char noTermAdded;
 int f, i, j, k, w;
 Jet z;
 dlist_iterator getNext( *(dlist*) jl );
 JLterm* p;
 JLterm* q;
 
 noTermAdded = 1;
 
 // --- Preliminary check of index set.
 
 w = 0;
 
 for( i = 0; i < oneEnv.NumVar; i++ ) {
   if( n[i] < 0 ) {
     cout << ::form( "\n*** JL ERROR: Cannot differentiate with negative index." );
     cout << ::form( "\n*******************************************************\n" );
     cout << endl;
     exit(0);
     }
   w += n[i];
 }
 
 if( w > (*this)->accuWgt ) {
   cout << ::form( "\n*** JL ERROR: Differentiation request beyond accuracy allowed." );
   cout << ::form( "\n**************************************************************\n" );
   cout << endl;
   exit(0);
 }
 
 if( w == 0 ) {
   z = *this;
   z.stacked = 1;
   return z;
 }

 // --- Construct the derivative one link at a time.
 // --- ( See note Obs.4 )
 
 while( p = (JLterm*) getNext() ) {
 
   q = new JLterm( p );
 
   char doit = 1;
   // -- Reset the index.
   for( i = 0; i < oneEnv.NumVar; i++ ) 
     doit = doit && ( ( q->index[i] -= n[i] ) >= 0 );
 
   if( doit ) {
     // -- Build factorial multiplier.
     f = 1;
     for( k = 0; k < oneEnv.NumVar; k++ ) {
       j = q->index[k];
       for( i = 0; i < n[k]; i++ ) f *= ++j;
       }
     if( f <= 0 ) {
       cout << ::form( "\n*** JL ERROR: Horrible, unexplainable error while differentiating!" );
       cout << ::form( "\n******************************************************************\n" );
       cout << endl;
       exit(0);
       }                           // Super fussbudget!! ( f "must" be positive )
   
     // -- Make final changes in private data of the JLterm and
     //    absorb it into the answer.
     (q->value) *= f;              // OK. So I'm a fussbudget with parentheses again ..
     (q->weight ) -= w;
   
     noTermAdded = 0;
     z->addTerm( q );               // ??? This also may be unnecessarily conservative.
                                    // The append function may be adequate, if the
                                    // ordering is preserved. ???
     } 
   }
 
 if( noTermAdded ) z = 0.0;
 
 
 // --- Finally, adjust accuWgt and return value
 
 z->accuWgt = (*this)->accuWgt - w;       // ??? Is this correct ???

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::inverse() {    // ??? REMOVE: Obsolete

 static char firstTime = 1;
 if( firstTime ) {
  firstTime = 0;
  cout << "\n*** WARNING ***                                  \n"
          "*** WARNING *** You are using the obsolete         \n"
          "*** WARNING *** Jet::inverse routine.  It will     \n"
          "*** WARNING *** soon vanish from the face of       \n"
          "*** WARNING *** the Earth.  Use the new and        \n"
          "*** WARNING *** improved Jet::Inverse instead.     \n"
          "*** WARNING ***                                    \n"
       << endl;
 }

 if( oneEnv.SpaceDim != 1 ) {
  cout << ::form( "\n" );
  cout << ::form( "*** ERROR ***                                     \n" );
  cout << ::form( "*** ERROR ***  Jet Jet::inverse()                 \n" );
  cout << ::form( "*** ERROR ***  This method can be called only     \n" );
  cout << ::form( "*** ERROR ***  if there is one and only one       \n" );
  cout << ::form( "*** ERROR ***  coordinate.  You have %d.          \n", oneEnv.SpaceDim );
  cout << ::form( "*** ERROR ***                                     \n" );
  cout << endl;
  exit(0);
 }

 Jet z;

 Jet x, v;
 Jet g;
 int d = 1;

 x.setVariable(0.0, 0);   // ??? Work on this!!!
 g = D(&d);
 
 int i = 0;
 z = x;
 v = ( operator()(&z) - x ) / g(&z);
 while( ( i++ < MX_MAXITER ) &&  ( v != 0.0 ) ) {
  z -= v;
  v = ( operator()(&z) - x ) / g(&z);
 }

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::Inverse() const {
 static Jet            z;
 static JLterm*        p;
 static double         zero( 0.0 );
 static char           t1, t2;

 if( oneEnv.SpaceDim != 1 ) {
  cout << "\n" 
          "*** ERROR ***                                     \n" 
          "*** ERROR ***  Jet Jet::Inverse()                 \n" 
          "*** ERROR ***  This method can be called only     \n" 
          "*** ERROR ***  if there is one and only one       \n" 
          "*** ERROR ***  coordinate.  You have " << oneEnv.SpaceDim << "\n"
          "*** ERROR ***                                     \n" 
       << endl;
  exit(1);
 }

 if( ( t1 = ( jl->refPoint[0]    == 0.0 ) ) && 
     ( t2 = ( jl->standardPart() == 0.0 ) ) 
   ) return EpsInverse();

 z = *this;   // Deep copy needed here.

 if( !t2 ) { 
  p = z.get();
  z->addTerm( new JLterm( oneEnv.AllZeroes, 0.0 ) );
 }

 if( !t1 ) z.fixReference( &zero );

 z = z.EpsInverse();

 if( t2 ) z.fixReference( &zero );
 else     z.fixReference( &(p->value) );

 delete p;    // ??? OK ???

 if( !t1 ) {
  // A little check
  if( z.standardPart() != 0.0 ) {
   cout << "*** WARNING ***                                              \n"
           "*** WARNING *** Jet::Inverse()                               \n"
           "*** WARNING *** z.standardPart() = " << z.standardPart() << "\n"
           "*** WARNING ***                                              \n"
        << endl;
  }
  z->addTerm( new JLterm( oneEnv.AllZeroes, jl->refPoint[0] ) );
 }

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet Jet::EpsInverse() const {
 static Jet        z, x, v, g;
 static int        d(1);

 x.setVariable(0.0, 0);
 g = D(&d);
 
 int i = 0;
 z = x;
 v = ( operator()(&z) - x ) / g(&z);
 while( ( i++ < MX_MAXITER ) &&  ( v != 0.0 ) ) {
  z -= v;
  v = ( operator()(&z) - x ) / g(&z);
 }

 z.stacked = 1;
 return z;
}
