#include "MathConstants.hxx"
#include <complex.h>

complex w( complex );

//////////////////////////////////////////////////////////////

complex erfSeries( complex z ) 
{
  static complex series;
  static complex oldseries;
  static complex arg;
  static double  den;
  static complex term;

  series        = 1.0;
  oldseries     = 0.0;
  arg           = 2.0*z*z;
  den           = 1.0;
  term          = 1.0;

  while( series != oldseries ) {
    oldseries = series;
    den      += 2.0;
    term     *= ( arg/den );
    series   += term;
  }  

  return z*series;
}

//////////////////////////////////////////////////////////////

complex erf( complex z ) 
{
  if( ( fabs(imag(z)) > 3.9 ) || ( fabs(real(z)) > 3.0 ) ) {
    complex u( - imag(z), real(z) );
    return ( 1.0 - exp(u*u)*w(u) );
  }

  static complex series;
  static complex oldseries;
  static complex arg;
  static complex term;
  static double  den;
  static double  fctr_x;

  series        = 1.0;
  oldseries     = 0.0;
  arg           = - z*z;
  den           = 1.0;
  term          = 1.0;
  fctr_x        = 0.0;

  while( series != oldseries ) {
    oldseries = series;
    den      += 2.0;
    fctr_x   += 1.0;
    term     *= arg/fctr_x;
    series   += term/den;
  }  

  return (2.0/MATH_SQRTPI)*z*series;
}

//////////////////////////////////////////////////////////////

complex w( complex z ) 
{
  static const complex mi( 0., -1. );
  static double x;
  static double y;
  
  x = real(z);
  y = imag(z);

  if( y < 0.0 )
    return 2.0*exp( -z*z ) - w( -z );

  if( x < 0.0 ) 
    return conj( w( complex( - x, y ) ) );

  if( ( x > 6.0 ) || ( y > 6.0 ) ) 
    return ( - mi * z * (
                        ( 0.5124242  /( z*z - 0.2752551 )) + 
                        ( 0.05176536 /( z*z - 2.724745  ))
                        ) 
           );

  if( ( x > 3.9 ) || ( y > 3.0 ) ) 
    return ( - mi * z * (
                        ( 0.4613135   /( z*z - 0.1901635 )) + 
                        ( 0.09999216  /( z*z - 1.7844927 )) + 
                        ( 0.002883894 /( z*z - 5.5253437 ))
                        ) 
           );

  return exp( -z*z )*( 1.0 - erf( mi*z ) );

}

