/*
**
** Utility function.
**
** Calculates the nonresonant normal form
** of a one-turn map.
**
** --- Leo Michelotti
** --- First version: November 4, 1993
** --- Final version: March 3, 1994
**
*/

#include "mxyzptlk.hxx"

#define MLT1  1.0e-10

char sh0( const int* index, const complex /* value */ ) {
 if( index[0] != index[3] + 1 ) return 0;
 if( index[1] != index[4] )     return 0;
 if( index[2] != index[5] )     return 0;
 return 1;
}

char sh1( const int* index, const complex /* value */ ) {
 if( index[0] != index[3] )     return 0;
 if( index[1] != index[4] + 1 ) return 0;
 if( index[2] != index[5] )     return 0;
 return 1;
}

char sh2( const int* index, const complex /* value */ ) {
 if( index[0] != index[3] )     return 0;
 if( index[1] != index[4] )     return 0;
 if( index[2] != index[5] + 1 ) return 0;
 return 1;
}

char sh3( const int* index, const complex /* value */ ) {
 if( index[0] != index[3] - 1 ) return 0;
 if( index[1] != index[4] )     return 0;
 if( index[2] != index[5] )     return 0;
 return 1;
}

char sh4( const int* index, const complex /* value */ ) {
 if( index[0] != index[3] )     return 0;
 if( index[1] != index[4] - 1 ) return 0;
 if( index[2] != index[5] )     return 0;
 return 1;
}

char sh5( const int* index, const complex /* value */ ) {
 if( index[0] != index[3] )     return 0;
 if( index[1] != index[4] )     return 0;
 if( index[2] != index[5] - 1 ) return 0;
 return 1;
}

// ===============================================================

void normalForm( LieOperator&   theMap,     /* input */
                 int            maxOrder,   /* input */
                 MatrixC* Bptr,
                 CLieOperator*  N,
                 CLieOperator*  T ) {

 static const complex FORIRIX_one( 1.0, 0.0 );    
                                            // This is, I hope, temporary.
                                            // It is done only to make
                                            // this code compile on the
                                            // SGI.

 typedef char (*CFUNCPTR)(const int*, const complex);
 static CFUNCPTR shear[] = { sh0, sh1, sh2, sh3, sh4, sh5 };
 int i, j;

      /* CAUTION */  // A little test
      /* CAUTION */  if( !(theMap.isNilpotent()) ) {
      /* CAUTION */   printf( "\n" );
      /* CAUTION */   printf( "***                                        \n" );
      /* CAUTION */   printf( "*** SORRY:                                 \n" );
      /* CAUTION */   printf( "*** SORRY: void normalForm()               \n" );
      /* CAUTION */   printf( "*** SORRY: This version only supports      \n" );
      /* CAUTION */   printf( "*** SORRY: nilpotent maps.                 \n" );
      /* CAUTION */   printf( "*** SORRY:                                 \n" );
      /* CAUTION */   exit(0);
      /* CAUTION */  }

 // Establishing linear normal form coordinates
 MatrixD  A;
 A = theMap.Jacobian();
 MatrixC B;
 B = A.eigenVectors();
 /* DGN  cout << "DGN> Before normalization: B = \n" << B;*/
 /* DGN  cout << endl;*/
 *Bptr           = B;

 MatrixC lambda;
 lambda = A.eigenValues();  // ??? How inefficient is this?
                                          // ??? I've already invoked A.eigenVectors()
      /* CAUTION */  for( i = 0; i < 6; i++ ) {
      /* CAUTION */   if( fabs( abs(lambda(i)) - 1.0 ) > MLT1 ) {
      /* CAUTION */    printf( "\n" );
      /* CAUTION */    printf( "***                                            \n" );
      /* CAUTION */    printf( "*** SORRY:                                     \n" );
      /* CAUTION */    printf( "*** SORRY: void normalForm()                   \n" );
      /* CAUTION */    printf( "*** SORRY: For now, I am only allowing         \n" );
      /* CAUTION */    printf( "*** SORRY: maps with an elliptic fixed point   \n" );
      /* CAUTION */    printf( "*** SORRY: at the origin.  This will be        \n" );
      /* CAUTION */    printf( "*** SORRY: loosened as soon as I'm sure        \n" );
      /* CAUTION */    printf( "*** SORRY: everything is working properly.     \n" );
      /* CAUTION */    printf( "*** SORRY:                                     \n" );
      /* CAUTION */    printf( "*** SORRY: lambda( %d ) has magnitude = %le    \n", 
      /* CAUTION */                                 i,            abs(lambda(i)) );
      /* CAUTION */    printf( "*** SORRY:                                     \n" );
      /* CAUTION */    exit(0);
      /* CAUTION */   }
      /* CAUTION */  }

 // Normalizing the linear normal form coordinates
 complex       mi( 0., -1. );
 MatrixD  J( "J", 6 );
 MatrixC Nx;
 Nx = ( B.transpose() * J * B * J ) * mi;

 /* DGN  cout << "DGN> The normalizer: Nx = \n" << Nx;*/
 /* DGN  cout << endl;*/
 for( i = 0; i < 6; i++ ) {
  Nx( i, i ) = 1.0 / sqrt( abs( Nx(i,i) ) );  
                                       // ??? "abs" should not be necessary,
                                       // ??? but Holt is finding some
                                       // ??? counterexamples.
                                       // ???
                                       // ??? In principle, could get divide
                                       // ??? by zero here.

  if( abs( ( (complex) 1.0 ) - Nx(i,i) ) < 1.0e-10 ) Nx(i,i) = 1.0;

      /* CAUTION */   for( j = 0; j < 6; j++ ) {
      /* CAUTION */    if( j == i ) continue;
      /* CAUTION */    else if( abs( Nx(i,j) ) > MLT1) {
      /* CAUTION */          printf( "\n" );
      /* CAUTION */          printf( "***                                        \n" );
      /* CAUTION */          printf( "*** ERROR:                                 \n" );
      /* CAUTION */          printf( "*** ERROR: void normalForm()               \n" );
      /* CAUTION */          printf( "*** ERROR: Nondiagonal element in BJB^TJ   \n" );
      /* CAUTION */          printf( "*** ERROR: abs( Nx( %d, %d ) ) = %le       \n", i, j, abs(Nx(i,j)) );
      /* CAUTION */          printf( "*** ERROR:                                 \n" );
      /* CAUTION */          exit(0);
      /* CAUTION */         }
      /* CAUTION */        else Nx(i,j) = complex_0;
      /* CAUTION */   }
 }

 B = B*Nx;

 /* DGN  cout << "DGN> The normalizer: Nx = \n" << Nx;*/
 /* DGN  cout << endl;*/
 /* DGN  cout << "DGN> After normalization: B = \n" << B;*/
 /* DGN  cout << endl;*/

 // Some useful matrices
 MatrixC Binv;
 Binv = B.inverse();
 MatrixC D;
 D = Binv * A * B;
 MatrixC Dinv;
 Dinv = D.inverse();

      /* CAUTION */  // A little checking and cleaning.
      /* CAUTION */  for( i = 0; i < 6; i++ ) {
      /* CAUTION */   if( fabs( abs(Dinv(i,i)) - 1.0 ) > MLT1 ) {
      /* CAUTION */    printf( "\n" );
      /* CAUTION */    printf( "***                                            \n" );
      /* CAUTION */    printf( "*** SORRY:                                     \n" );
      /* CAUTION */    printf( "*** SORRY: void normalForm()                   \n" );
      /* CAUTION */    printf( "*** SORRY: For now, I am only allowing         \n" );
      /* CAUTION */    printf( "*** SORRY: maps with an elliptic fixed point   \n" );
      /* CAUTION */    printf( "*** SORRY: at the origin.  This will be        \n" );
      /* CAUTION */    printf( "*** SORRY: loosened as soon as I'm sure        \n" );
      /* CAUTION */    printf( "*** SORRY: everything is working properly.     \n" );
      /* CAUTION */    printf( "*** SORRY:                                     \n" );
      /* CAUTION */    printf( "*** SORRY: Dinv( %d, %d ) has magnitude = %le  \n", 
      /* CAUTION */                               i,  i,            abs(Dinv(i,i)) );
      /* CAUTION */    printf( "*** SORRY:                                     \n" );
      /* CAUTION */    exit(0);
      /* CAUTION */   }
      /* CAUTION */   for( j = 0; j < 6; j++ ) {
      /* CAUTION */    if( j == i ) continue;
      /* CAUTION */    else if( abs( Dinv(i,j) ) > MLT1) {
      /* CAUTION */          printf( "\n" );
      /* CAUTION */          printf( "***                                        \n" );
      /* CAUTION */          printf( "*** ERROR:                                 \n" );
      /* CAUTION */          printf( "*** ERROR: void normalForm()               \n" );
      /* CAUTION */          printf( "*** ERROR: A horrible, inexplicable error  \n" );
      /* CAUTION */          printf( "*** ERROR: has occurred.                   \n" );
      /* CAUTION */          printf( "*** ERROR:                                 \n" );
      /* CAUTION */          exit(0);
      /* CAUTION */         }
      /* CAUTION */        else Dinv(i,j) = complex_0;
      /* CAUTION */   }
      /* CAUTION */  }
      /* CAUTION */ 
      /* CAUTION */  for( i = 0; i < 6; i++ ) {
      /* CAUTION */   if( fabs( abs(D(i,i)) - 1.0 ) > MLT1 ) {
      /* CAUTION */    printf( "\n" );
      /* CAUTION */    printf( "***                                          \n" );
      /* CAUTION */    printf( "*** SORRY:                                   \n" );
      /* CAUTION */    printf( "*** SORRY: void normalForm()                 \n" );
      /* CAUTION */    printf( "*** SORRY: For now, I am only allowing       \n" );
      /* CAUTION */    printf( "*** SORRY: maps with an elliptic fixed point \n" );
      /* CAUTION */    printf( "*** SORRY: at the origin.  This will be      \n" );
      /* CAUTION */    printf( "*** SORRY: loosened as soon as I'm sure      \n" );
      /* CAUTION */    printf( "*** SORRY: everything is working properly.   \n" );
      /* CAUTION */    printf( "*** SORRY:                                   \n" );
      /* CAUTION */    printf( "*** SORRY: D( %d, %d ) has magnitude = %le   \n", 
      /* CAUTION */                            i,  i,                abs(D(i,i)) );
      /* CAUTION */    printf( "*** SORRY:                                   \n" );
      /* CAUTION */    exit(0);
      /* CAUTION */   }
      /* CAUTION */ 
      /* CAUTION */   for( j = 0; j < 6; j++ ) {
      /* CAUTION */    if( j == i ) continue;
      /* CAUTION */    else if( abs( D(i,j) ) > MLT1) {
      /* CAUTION */          printf( "\n" );
      /* CAUTION */          printf( "***                                        \n" );
      /* CAUTION */          printf( "*** ERROR:                                 \n" );
      /* CAUTION */          printf( "*** ERROR: void normalForm()               \n" );
      /* CAUTION */          printf( "*** ERROR: A horrible, inexplicable error  \n" );
      /* CAUTION */          printf( "*** ERROR: has occurred.                   \n" );
      /* CAUTION */          printf( "*** ERROR:                                 \n" );
      /* CAUTION */          exit(0);
      /* CAUTION */         }
      /* CAUTION */        else D(i,j) = complex_0;
      /* CAUTION */   }
      /* CAUTION */  }


 // The original near-identity transformation
 CLieOperator     CL1( theMap );
 CLieOperator     id( "ident" );
 CLieOperator     calN = Dinv*(Binv*CL1( B*id ));
 CLieOperator     mapT;


 // And the rest ...
 complex          factor;
 int              ll;

 CLieOperator     reg;
 JLCtermData*     q;
 JetC             scr;

 dlist*           doc;
 dlist_iterator*  g;

 for( int k = 0; k <= maxOrder - 2; k++ ) {
  reg = id;
  ll = 0;
  while( ll < k ) reg = N[ll++].expMap( -1.0, reg );
  reg = calN( reg );
  // -------------
  reg = reg.filter( k+2, k+2 );
  N[k] = reg.filter( shear );

  doc = ( N[k] - reg ).image();
  for( i = 0; i < 6; i++ ) {
   scr.clear();
   g = new dlist_iterator( doc[i] );
   while( q = (JLCtermData*) (*g)() ) {
    factor = 1.0;
    for( j = 0; j < 6; j++ ) {
         factor *= pow( FORIRIX_one / lambda(j), q->index[j] );
      // factor *= pow( 1.0 / lambda(j), q->index[j] );
    }
    factor *= lambda(i);
    q->value /= ( factor - (complex) 1.0 );
    scr.addTerm( new JLCterm( *q ) );
   }
   T[k].setVariable( scr, i );
   delete g;
  }
  T[k].reset();    // ??? Probably unnecessary

  // Prepare for the next order
  reg = Dinv*id;
  mapT = T[k].expMap( 1.0, id );
  reg = mapT( reg );
  reg = D*reg;
  reg = calN( reg );
  mapT = T[k].expMap( -1.0, id );
  calN = mapT( reg );
  // In one line:
  // calN = T[k].expMap( -1.0, calN( D*( T[k].expMap( 1.0, Dinv*id ) ) ) );
 }

}
